package sedplugin.characterization;

import java.io.File;

import java.net.MalformedURLException;
import java.net.URL;

import sedplugin.errors.CharacterizationException;

import sedplugin.plugin.AladinImage;

/**
 * <p>Singleton Class</p>
 * <p>
 * 	Lets managing Characterization/FITS mapping URL/files,
 * 	and particularly Aladin URLs (see {@link #getAladinCharac(URL)}).
 * </p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 */
public final class CharacManager {

	/**
	 * Empty private constructor => Singleton Class !
	 */
	private CharacManager(){ ; }

	/**
	 * Replaces the extension of the given file path with the given one.
	 * 
	 * @param filePath		Path of a file whose the extension must be changed.
	 * @param newExtension	The new extension of the file file.
	 * 
	 * @return				The given file path, but with the given extension.
	 */
	public static final String changeExtension(final String filePath, final String newExtension){
		if (filePath == null)
			return null;

		int indDot = filePath.lastIndexOf(".");
		if (indDot <= 0)
			indDot = filePath.length()-1;
		return filePath.substring(0, indDot)+"."+newExtension;
	}

	/**
	 * <p>Converts the given image URL into a File instance pointing on a local Characterization file
	 * ONLY IF the image is local (so the URL starts with a "file://").</p>
	 * <p>The file path is extracted from the URL but the extension is changed for ".XML"
	 * (extension of a Characterization file).</p>
	 * 
	 * @param imgUrl	URL pointing on a local image (so, using the FILE protocol).
	 * 
	 * @return			A file corresponding to the given URL or <i>null</i> if the URL is not "file://".
	 * 
	 * @see #changeExtension(String, String)
	 */
	public static final File getDefaultXML(final URL imgUrl){
		if (imgUrl != null && imgUrl.getProtocol().equalsIgnoreCase("file"))
			return new File(changeExtension(imgUrl.getPath(), "xml"));
		else
			return null;
	}

	/**
	 * <p>Converts the given image URL into a File instance pointing on a local FITS mapping file
	 * ONLY IF the image is local (so the URL starts with a "file://").</p>
	 * <p>The file path is extracted from the URL but the extension is changed for ".MAP"
	 * (extension of a FITS mapping file).</p>
	 * 
	 * @param imgUrl	URL pointing on a local image (so, using the FILE protocol).
	 * 
	 * @return			A file corresponding to the given URL or <i>null</i> if the URL is not "file://".
	 * 
	 * @see #changeExtension(String, String)
	 */
	public static final File getDefaultMap(final URL imgUrl){
		if (imgUrl != null && imgUrl.getProtocol().equalsIgnoreCase("file"))
			return new File(changeExtension(imgUrl.getPath(), "map"));
		else
			return null;
	}

	/**
	 * <p>Converts the given image URL into a File instance pointing on a local generic FITS mapping file
	 * ONLY IF the image is local (so the URL starts with a "file://").</p>
	 * <p>The directory path is extracted from the URL but the file name is replaced by "generic.map".</p>
	 * 
	 * @param imgUrl	URL pointing on a local image (so, using the FILE protocol).
	 * 
	 * @return			A file corresponding to the given URL or <i>null</i> if the URL is not "file://".
	 */
	public static final File getGenericCharac(final URL imgUrl){
		if (imgUrl != null && imgUrl.getProtocol().equalsIgnoreCase("file")){
			File f = new File(imgUrl.getPath());
			return new File(f.getParentFile(), "generic.map");
		}else
			return null;
	}

	/**
	 * <p>Extracts and reads the Characterization associated with the given image.</p>
	 * <p>The URL of the image is used to build the Characterization URL (even if it is a local file).
	 * The Characterization URL may point toward either a full Characterization file or a FITS mapping file.</p>
	 * 
	 * @param io	The image whose the Characterization is wanted.
	 * 
	 * @return		All the information extracted from the Characterization file and useful to the SEDPlugin.
	 * 
	 * @throws CharacterizationException	If there is an error while building the Characterization URL
	 * 										or while reading the Characterization/FITS mapping file.
	 * 
	 * @see {@link #getCharacURL(URL, boolean)}
	 * @see Characterization#Characterization(AladinImage, FitsMapping)
	 * @see Characterization#Characterization(AladinImage, URL)
	 */
	public final static Characterization getCharacterization(final AladinImage io) throws CharacterizationException {
		if (io == null)
			return null;

		// Build the Characterization file URL:
		URL characUrl = getCharacURL(io.getImageUrl(), io.isAllSky());

		// Create the Characterization object:
		Characterization charac = null;
		if (characUrl != null){
			if (characUrl.getFile().endsWith(".map"))
				charac = new Characterization(io, new FitsMapping(new File(characUrl.getPath().toString())));
			else
				charac = new Characterization(io, characUrl);
		}

		return charac;
	}

	/**
	 * Constructs the URL which points toward the Characterization file associated with the given URL image.
	 *
	 * @param imageURL 						URL of the image whose the Characterization is wanted.
	 * @param allSky						<i>true</i> if the corresponding image is AllSky, <i>false</i> otherwise.
	 *
	 * @return 								Characterization file URL.
	 *
	 * @throws CharacterizationException	If it's impossible to construct the URL.
	 * 
	 * @see #getLocalCharac(URL)
	 * @see #getDistantCharac(URL, boolean)
	 */
	@SuppressWarnings("deprecation")
	public final static URL getCharacURL(final URL imageURL, final boolean allSky) throws CharacterizationException {
		if (imageURL == null)
			return null;

		URL characURL = null;

		// CASE: LOCAL IMAGE:
		if (imageURL.getProtocol().equalsIgnoreCase("file")){
			File f = getLocalCharac(imageURL);
			if (f != null){
				try {
					characURL = f.toURL();
				} catch (MalformedURLException e) {
					throw new CharacterizationException(imageURL.getFile(), f.getAbsolutePath(), true, CharacterizationException.FETCHING_ERROR, "Malformed URL");
				}
			}

		}// CASE: DISTANT IMAGE:
		else
			characURL = getDistantCharac(imageURL, allSky);

		return characURL;
	}

	/**
	 * <p>Try to find a URL which points toward the Characterization/FITS mapping file associated with the
	 * specified local image.</p>
	 * <p>First, this function searches a Characterization file ".xml" with the same name as the image file.
	 * If no match, it searches for a FITS mapping file ".map" with the same name as the image file.
	 * If still no match, it tries to find a file a global/generic FITS mapping file "generic.map".</p>
	 * 
	 * @param imgUrl	URL of the local image whose the Characterization is wanted.
	 * 
	 * @return			The URL of the Characterization or FITS mapping file corresponding to the specified image,
	 * 					or <i>null</i> if no such file can be find.
	 * 
	 * @see #getDefaultXML(URL)
	 * @see #getDefaultMap(URL)
	 * @see #getGenericCharac(URL)
	 */
	public final static File getLocalCharac(final URL imgUrl) {
		if (imgUrl == null)
			return null;

		File f = null;

		// Look for its own XML Characterization:
		f = getDefaultXML(imgUrl);
		if (f != null && (!f.exists() || !f.isFile() || !f.canRead()))
			f = null;

		// Look for its own MAP Characterization:
		if (f == null){
			f = getDefaultMap(imgUrl);
			if (f != null && (!f.exists() || !f.isFile() || !f.canRead()))
				f = null;
		}

		// Look for a generic MAP Characterization:
		if (f == null){
			f = getGenericCharac(imgUrl);
			if (f != null && (!f.exists() || !f.isFile() || !f.canRead()))
				f = null;
		}

		return f;
	}

	/**
	 * <p>Builds the URL toward the Characterization (if it exists) of the specified distant image.</p>
	 * <p>If the specified image is an AllSky one, the Characterization should be in the same root directory
	 * of the image with the name "charac.xml". If the image comes from the Aladin server (URL starting
	 * with "http://aladin.u-strasbg.fr/cgi-bin/nph-Aladin++"), the image URL is modified to get the Characterization
	 * file (see {@link #getAladinCharac(URL)}). Any other image is ignored and so, can not be used by the
	 * SEDPlugin.</p>
	 * 
	 * @param imgUrl	URL of the image whose the Characterization is wanted.
	 * @param allSky	<i>true</i> if the specified image is an AllSky one, <i>false</i> otherwise.
	 * 
	 * @return			The URL of the corresponding Characterization.
	 * 
	 * @throws CharacterizationException	If the URL can not be built.
	 * 
	 * @see #isImageFromAladinServer(URL)
	 * @see #getAladinCharac(URL)
	 */
	public final static URL getDistantCharac(final URL imgUrl, final boolean allSky) throws CharacterizationException {
		if (imgUrl == null)
			return null;

		URL characUrl = null;

		// CASE: Aladin server for an AllSky image:
		if (allSky){
			try{
				characUrl = new URL(imgUrl.getProtocol(), imgUrl.getHost(), imgUrl.getPort(), imgUrl.getPath()+"/charac.xml");
			} catch (MalformedURLException e) {
				throw new CharacterizationException(imgUrl.getPath()+"/charac.xml", null, true, CharacterizationException.FETCHING_ERROR, "Impossible to well-form the Characterization URL.");
			}

		}// CASE: Aladin server:
		else if (isImageFromAladinServer(imgUrl))
			characUrl = getAladinCharac(imgUrl);

		// OTHERWISE: let's supposing that the Characterization URL is the same than the image URL (except the file extension):
		if (characUrl == null){
			String path = imgUrl.getPath();
			int ind = path.indexOf(".fits");
			if (ind < 0)
				ind = path.indexOf(".jpeg");
			if (ind < 0)
				ind = path.indexOf(".jpg");
			if (ind < 0)
				return null;
			path = path.substring(0, ind)+".xml";
			try{
				characUrl = new URL(path);
			}catch(MalformedURLException ex){
				throw new CharacterizationException(imgUrl.getFile(), path, true, CharacterizationException.FETCHING_ERROR, "Malformed URL.");
			}
		}

		return characUrl;
	}
	
	/**
	 * Tells whether the given URL points toward the Aladin image server.
	 * 
	 * @param imgUrl	An image URL.
	 * 
	 * @return			<i>true</i> if the image comes from the Aladin server, <i>false</i> otherwise.
	 */
	public final static boolean isImageFromAladinServer(final URL imgUrl){
		return imgUrl.toString().startsWith("http://aladin.u-strasbg.fr/cgi-bin/nph-Aladin++");
	}

	/**
	 * <p>Gets the URL of the Characterization file corresponding to the given image and provided by the Aladin
	 * server.</p>
	 * <p>The given URL is copied, and the GET parameters of the copy are modified in the following way:
	 * "out=qualifier" and "mode=charxml". The other parameters are kept.</p>
	 * 
	 * @param imageURL		The URL of the image whose the Characterization is wanted.
	 * 
	 * @return				The corresponding Characterization URL or <i>null</i> if the image is not coming from the
	 * 						Aladin server or is <i>null</i>.
	 * 
	 * @throws CharacterizationException	If there is an error while building the URL.
	 * 
	 * @see #isImageFromAladinServer(URL)
	 */
	public final static URL getAladinCharac(final URL imageURL) throws CharacterizationException {
		if (imageURL == null)
			return null;

		if (imageURL.getQuery() == null)
			throw new CharacterizationException(imageURL.getFile(), null, true, CharacterizationException.FETCHING_ERROR, "Impossible to get the associated Characterization file because the image URL format is unknown.");

		// Check that the image URL is an Aladin server URL:
		if (!isImageFromAladinServer(imageURL)){
			System.err.println("### ERROR[getAladinCharac(URL)]: \""+imageURL+"\" is not an Aladin server URL !");
			return null;
		}else if (!imageURL.toString().startsWith("http://aladin.u-strasbg.fr/cgi-bin/nph-Aladin++.cgi?"))
			System.out.println("# INFO [SEDPlugin]: Aladin Test Server #");

		// Get the query part of the URL and split it so that separating all parameters:
		String[] queryParts = imageURL.getQuery().split("&");
		String query = "";

		// Modify the 'out' and 'mode' parameters and concatenate all them:
		for(int i=0; i<queryParts.length; i++){
			if (queryParts[i].startsWith("out="))
				queryParts[i] = "out=qualifier";
			else if (queryParts[i].startsWith("mode="))
				queryParts[i] = "mode=charxml";

			query = query + ((i==0)?"":"&") + queryParts[i];
		}

		// Create and return the new URL:
		try {
			return new URL(imageURL.getProtocol(), imageURL.getHost(), imageURL.getPort(), imageURL.getPath()+"?"+query);
		} catch (MalformedURLException e) {
			throw new CharacterizationException(imageURL.getFile(), null, true, CharacterizationException.FETCHING_ERROR, "Impossible to well-form the Characterization URL.");
		}
	}
}
