package sedplugin.characterization;

import java.io.IOException;
import java.io.InputStream;

import java.net.URL;

import org.kxml2.io.KXmlParser;

import org.xmlpull.v1.XmlPullParserException;

import sedplugin.errors.CharacterizationException;
import sedplugin.errors.SEDException;

import sedplugin.plugin.SEDPlugin;


/**
 * A parser of Characterisation file, used by SEDPlugin to extract the average wavelength of an aladin image.
 *
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 * 
 * @see KXmlParser
 */
public class CharacParser {

	protected final static String SpatialTAG = "spatial";
	protected final static String TimeTAG = "time";
	protected final static String SpectralTAG = "spectral";
	protected final static String FluxTAG = "flux";

	protected final static String AxisNameSTACK = "characterization characterizationAxis axisName";
	protected final static String CalibrationStatusSTACK = "characterization characterizationAxis calibrationStatus";
	protected final static String DateSTACK = "characterization characterizationAxis coverage location coord Time TimeInstant MJDTime";
	protected final static String UnitSTACK = "characterization characterizationAxis unit";
	protected final static String WavelengthSTACK = "characterization characterizationAxis coverage location coord Spectral Value";

	protected boolean characterizationFile;
	protected boolean spatialAxis;
	protected boolean timeAxis;
	protected boolean spectralAxis;
	protected boolean fluxAxis;

	protected String spatialCalibration;
	protected String spectralCalibration;
	protected String fluxCalibration;

	protected double date;
	protected String spectralUnit;
	protected double wavelength;
	protected String fluxUnit;

	private KXmlParser parser;
	protected String tagStack;

	protected CharacErrors errors;

	public CharacParser(){
		reset();
	}

	/**
	 * Resets all fields designed to store the data required by the SEDPlugin and read by this parser.
	 * The errors are also deleted.
	 */
	protected void reset(){
		characterizationFile = false;
		spatialAxis = false;
		timeAxis = false;
		spectralAxis = false;
		fluxAxis = false;

		spatialCalibration = null;
		spectralCalibration = null;
		fluxCalibration = null;

		date = 0.0;
		spectralUnit = null;
		wavelength = Double.NaN;
		fluxUnit = null;

		parser = null;
		tagStack = "";

		errors = new CharacErrors();
	}

	/**
	 * <p>Parses the specified Characterization file.</p>
	 * <p>If this function returns <i>false</i>, all data required by the SEDPlugin to use the image have not
	 * been found in the XML document or their value is incorrect. To have more details about these problems/errors
	 * use the function {@link #getErrors()}.</p>
	 *
	 * @param imgName		The label of the image.
	 * @param characUrl		The URL of the Characterization file.
	 * 
	 * @return				<i>true</i> if all fetched data are correct,
	 * 						<i>false</i> otherwise (see {@link #getErrors()} for more details).
	 * 
	 * @throws SEDException	If there was an error while reading or parsing the file.
	 * 
	 * @see #getErrors()
	 */
	public boolean parse(String imgName, URL characUrl) throws CharacterizationException {
		if (characUrl == null)
			return false;
		else if (!characUrl.getProtocol().equalsIgnoreCase("file") && !SEDPlugin.hasInternetConnection())
			throw new CharacterizationException(imgName, characUrl.toString(), true, CharacterizationException.READING_ERROR, "No Internet connection ! => Impossible to read the Characterization file.");

		// Reset the whole parser:
		reset();

		// Build the input stream:
		try{
			InputStream input = characUrl.openStream();
			parser = new KXmlParser();
			parser.setFeature(KXmlParser.FEATURE_PROCESS_NAMESPACES, true);
			parser.setInput(input, null);
		}catch(IOException ioe){
			throw new CharacterizationException(imgName, characUrl.toString(), true, CharacterizationException.READING_ERROR, ioe);
		}catch(XmlPullParserException xppe){
			throw new CharacterizationException(imgName, characUrl.toString(), true, CharacterizationException.PARSING_ERROR, xppe);
		}catch(NullPointerException npe){
			throw new CharacterizationException(imgName, characUrl.toString(), true, CharacterizationException.READING_ERROR, npe);
		}

		// Parse the given file:
		try{
			int eventType = parser.getEventType();
			while (eventType != KXmlParser.END_DOCUMENT) {
				/*if(eventType == KXmlParser.START_DOCUMENT) {
	        		System.out.println("Start document");
	        	}else if(eventType == KXmlParser.END_DOCUMENT) {
	        		System.out.println("End document");
	        	}else */if(eventType == KXmlParser.START_TAG) {
	        		startTag();
	        	}else if(eventType == KXmlParser.END_TAG) {
	        		endTag();
	        	}else if(eventType == KXmlParser.TEXT) {
	        		text(parser.getText());
	        	}
	        	eventType = parser.next();
			}
		}catch(IOException ioe){
			throw new CharacterizationException(imgName, characUrl.toString(), true, CharacterizationException.READING_ERROR, ioe);
		}catch(XmlPullParserException e){
			throw new CharacterizationException(imgName, characUrl.toString(), true, CharacterizationException.PARSING_ERROR, e);
		}

		// Check the fetched data:
		if (!characterizationFile)
			errors.otherErrors.add("\""+characUrl+"\" isn't a characterization file (the root node must be \"characterization\") !");
		else if (wavelength == Double.NaN)
			errors.wlValueError = "No wavelength value has been found.";
		errors.wlCalibrated = spectralCalibration.equalsIgnoreCase(FitsMapping.CALIBRATED);
		errors.fluxCalibrated = fluxCalibration.equalsIgnoreCase(FitsMapping.CALIBRATED);

		return errors.isCharacCorrect();
	}

	public final CharacErrors getErrors(){
		return errors;
	}

	private void startTag() throws XmlPullParserException {
		String tag = parser.getName();
		if (tag.equalsIgnoreCase("characterization"))
			characterizationFile = true;
		if (!parser.isEmptyElementTag())
			tagStack = (tagStack+" "+tag).trim();
	}

	private void endTag(){
		String tag = parser.getName();
		if (tag.equalsIgnoreCase("characterization characterizationAxis") && spectralAxis)
			spectralAxis = false;
		int ind = tagStack.indexOf(tag);
		if (ind > -1)
			tagStack = tagStack.substring(0, ind).trim();
	}

	private void text(String txt){
		txt = txt.trim();
		if (tagStack.equals(AxisNameSTACK)){
			fluxAxis = txt.equalsIgnoreCase(FluxTAG);
			spectralAxis = txt.equalsIgnoreCase(SpectralTAG);
			timeAxis = txt.equalsIgnoreCase(TimeTAG);
			spatialAxis = txt.equalsIgnoreCase(SpatialTAG);
		}else if (fluxAxis){
			if (tagStack.equals(CalibrationStatusSTACK))
				fluxCalibration = txt;
			else if (tagStack.equals(UnitSTACK)){
				fluxUnit = txt;
			}
		}else if (spectralAxis){
			if (tagStack.equals(CalibrationStatusSTACK))
				spectralCalibration = txt;
			else if (tagStack.equals(UnitSTACK))
				spectralUnit = txt;
			else if (tagStack.equals(WavelengthSTACK)){
				try{
					wavelength = Double.parseDouble(txt);
				}catch(NumberFormatException nfe){
					wavelength = Double.NaN;
				}
			}
		}else if (timeAxis){
			if (tagStack.equals(DateSTACK)){
				try{
					date = Double.parseDouble(txt);
				}catch(NumberFormatException nfe){
					date = 0.0;
				}
			}
		}else if (spatialAxis){
			if (tagStack.equals(CalibrationStatusSTACK))
				spatialCalibration = txt;
		}
	}

	@Override
	public String toString(){
		return "Calibration spatiale: "+((spatialCalibration.equalsIgnoreCase(FitsMapping.CALIBRATED))?"Oui":"Non")+
		"\n"+"Date: "+date+
		"\n"+"Calibration spectrale: "+(errors.isSpectralCalibrated()?"Oui":"Non")+
		"\n"+"Unité (spectral): "+spectralUnit+
		"\n"+"Calibration des flux: "+(errors.isFluxCalibrated()?"Oui":"Non")+
		"\n"+"Unité (flux): "+fluxUnit+
		"\n"+"Longueur d'onde = "+wavelength+" "+spectralUnit;
	}

	public double getMJDDate(){
		return date;
	}

	public String getSpectralUnit(){
		return spectralUnit;
	}

	public String getSpectralCalibration(){
		return spectralCalibration;
	}

	public double getWavelength(){
		return wavelength;
	}

	public String getFluxCalibration(){
		return fluxCalibration;
	}

	public String getFluxUnit(){
		return fluxUnit;
	}

}
