package sedplugin.characterization;

import java.io.File;

import java.net.URL;

import sedplugin.errors.CharacterizationException;

import sedplugin.plugin.AladinImage;

import sedplugin.sed.ToolBox;

/**
 * <p>This class gather all data required by the SEDPlugin about an image to use it in SEDs.</p>
 * <p>These data are read either from a Characterization file (by {@link CharacParser}), or from
 * a FITS mapping file (by {@link FitsMapping}). The class {@link CharacManager} is the only one to
 * know how to get the Characterization associated with an image:
 * see {@link CharacManager#getCharacterization(AladinImage)}.</p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 */
public class Characterization {

	/** Indicates whether the spectral axis is calibrated. */
	protected String spectralCalibration;
	/** The central wavelength of this image. */
	protected double wavelength;
	/** The unit of the abscissa values. */
	protected String wavelengthUnit;
	/** Indicates whether the flux axis is calibrated. */
	protected String fluxCalibration;
	/** The unit of the ordinate values. */
	protected String fluxUnit;
	/** The date of the image. */
	protected String date;

	/** The associated image. */
	protected final AladinImage image;
	/** The URL of the Characterization file associated with this image. */
	protected URL characUrl;
	/** The file which maps FITS keywords with Characterization data. */
	protected FitsMapping mappingFile;
	/** The date of the last modification of the Characterization file. */
	protected long fileLastModified = 0L;

	/** A list of errors explaining why this Characterization is not usable. */
	protected CharacErrors errors;


	/**
	 * Loads a full Characterization from the given URL for the given image loaded in Aladin.
	 * 
	 * @param image					Image loaded in Aladin and whose this object will represent the Characterization.
	 * @param characterizationURL	URL of the Characterization (associated with the given file) to load.
	 * 
	 * @throws CharacterizationException	If the Characterization can not be loaded.
	 * 
	 * @see #loadCharacterization(URL)
	 */
	public Characterization(AladinImage image, URL characterizationURL) throws CharacterizationException {
		this.image = image;
		errors = new CharacErrors();
		loadCharacterization(characterizationURL);
	}

	/**
	 * Loads a partial Characterization from the given FITS mapping for the given image loaded in Aladin.
	 * 
	 * @param image			Image loaded in Aladin and whose this object will represent the Characterization.
	 * @param mapping		FITS mapping to use to build this Characterization.
	 * 
	 * @see #loadCharacterization(FitsMapping)
	 */
	public Characterization(AladinImage image, FitsMapping mapping){
		this.image = image;
		errors = new CharacErrors();
		loadCharacterization(mapping);
	}

	/**
	 * Reads the specified full Characterization, extracts all data required by the SEDPlugin about the
	 * associated image and stores them in this object. Once loaded, the Characterization is checked
	 * thanks to {@link #checkCharacterization(boolean)}.
	 * 
	 * @param characterizationURL		URL of the Characterization to load.
	 * 
	 * @return							<i>true</i> if the Characterization has been successfully loaded,
	 * 									<i>false</i> otherwise.
	 * 
	 * @throws CharacterizationException	If the specified Characterization can not be loaded.
	 * 
	 * @see CharacParser
	 * @see #checkCharacterization(boolean)
	 */
	public boolean loadCharacterization(URL characterizationURL) throws CharacterizationException {
		if (characterizationURL == null)
			return false;

		characUrl = characterizationURL;
		mappingFile = null;

		// Parse the Characterization file:
		CharacParser parser = new CharacParser();
		boolean success = parser.parse(image.getName(), characUrl);
		if (!success)
			errors = new CharacErrors(parser.getErrors());
		else
			errors.reset();

		// Set the last modification date:
		fileLastModified = 0L;
		if (characterizationURL.getProtocol().equalsIgnoreCase("file")){
			File f = new File(characterizationURL.toString());
			fileLastModified = f.lastModified();
		}

		// Load the parsing results:
		spectralCalibration = parser.getSpectralCalibration();
		wavelength = parser.getWavelength();
		wavelengthUnit = parser.getSpectralUnit();
		fluxCalibration = parser.getFluxCalibration();
		fluxUnit = parser.getFluxUnit();
		date = parser.getMJDDate()+"";

		// Check the current Characterization:
		checkCharacterization(/*false*/);

		return true;
	}

	/**
	 * <p>Loads the given FITS mapping in order to keep all data required by the SEDPlugin
	 * about the associated image. Once loaded, the Characterization is checked
	 * thanks to {@link #checkCharacterization(boolean)}.</p>
	 * 
	 * <p><i><u>note:</u> The header FITS of the associated image is read and given to the given
	 * FITS mapping.</i></p>
	 * 
	 * @param mapping	FITS mapping.
	 * 
	 * @return			<i>true</i> if the Characterization has been successfully loaded,
	 * 					<i>false</i> otherwise.
	 * 
	 * @see AladinImage#getHeaderFits()
	 * @see #checkCharacterization(boolean)
	 */
	public boolean loadCharacterization(FitsMapping mapping){
		if (image == null || mapping == null)
			return false;

		characUrl = null;
		mappingFile = mapping;

		// Set the last modification date:
		fileLastModified = 0L;
		if (mapping.getMappingFile() != null)
			fileLastModified = mapping.getMappingFile().lastModified();

		// Get the header FITS of the image if any:
		HeaderFits header = image.getHeaderFits();

		// Fill this Characterization thanks to the given mapping:
		errors.reset();

		// Spectral calibration:
		if (header != null && FitsMapping.isFitsKeyword(mapping.getSpectralCalibration())){
			HeaderFits.HeaderItem headerItem = header.getKeywordItem(FitsMapping.extractFitsKeywordName(mapping.getSpectralCalibration()));
			if (headerItem != null){
				spectralCalibration = headerItem.value;
				errors.wlCalibrated = spectralCalibration.equalsIgnoreCase(FitsMapping.CALIBRATED);
			}else{
				spectralCalibration = "?";
				errors.wlCalibrationError = "The FITS keyword \""+mapping.getSpectralCalibration()+"\" doesn't exist !";
			}
		}else{
			spectralCalibration = FitsMapping.extractValue(mapping.getSpectralCalibration());
			errors.wlCalibrated = spectralCalibration.equalsIgnoreCase(FitsMapping.CALIBRATED);
		}

		// Wavelength value:
		if (header != null && FitsMapping.isFitsKeyword(mapping.getWavelength())){
			HeaderFits.HeaderItem headerItem = header.getKeywordItem(FitsMapping.extractFitsKeywordName(mapping.getWavelength()));
			if (headerItem != null){
				try{
					wavelength = Double.parseDouble(headerItem.value);
				}catch(NumberFormatException nfe){
					wavelength = Double.NaN;
					errors.wlValueError = "The wavelength must be numeric and the FITS keyword \""+headerItem.keyword+"\" (="+headerItem.value+") is not numeric.";
				}
			}else{
				wavelength = Double.NaN;
				errors.wlValueError = "The FITS keyword \""+FitsMapping.extractFitsKeywordName(mapping.getWavelength())+"\" doesn't exist !";
			}
		}else{
			try{
				wavelength = Double.parseDouble(FitsMapping.extractValue(mapping.getWavelength()));
			}catch(NumberFormatException nfe){
				wavelength = Double.NaN;
				errors.wlValueError = "The wavelength must be numeric and \""+mapping.getWavelength()+"\" is not numeric.";
			}
		}

		// Wavelength unit:
		if (header != null && FitsMapping.isFitsKeyword(mapping.getWavelengthUnit())){
			HeaderFits.HeaderItem headerItem = header.getKeywordItem(FitsMapping.extractFitsKeywordName(mapping.getWavelengthUnit()));
			if (headerItem != null)
				wavelengthUnit = headerItem.value;
			else
				errors.wlUnitError = "The FITS keyword \""+mapping.getWavelengthUnit()+"\" doesn't exist !";
		}else
			wavelengthUnit = FitsMapping.extractValue(mapping.getWavelengthUnit());

		// Flux calibration:
		if (header != null && FitsMapping.isFitsKeyword(mapping.getFluxCalibration())){
			HeaderFits.HeaderItem headerItem = header.getKeywordItem(FitsMapping.extractFitsKeywordName(mapping.getFluxCalibration()));
			if (headerItem != null){
				fluxCalibration = headerItem.value;
				errors.fluxCalibrated = fluxCalibration.equalsIgnoreCase(FitsMapping.CALIBRATED);
			}else{
				fluxCalibration = "?";
				errors.fluxCalibrationError = "The FITS keyword \""+mapping.getFluxCalibration()+"\" doesn't exist !";
			}
		}else{
			fluxCalibration = FitsMapping.extractValue(mapping.getFluxCalibration());
			errors.fluxCalibrated = fluxCalibration.equalsIgnoreCase(FitsMapping.CALIBRATED);
		}

		// Flux unit:
		if (header != null && FitsMapping.isFitsKeyword(mapping.getFluxUnit())){
			HeaderFits.HeaderItem headerItem = header.getKeywordItem(FitsMapping.extractFitsKeywordName(mapping.getFluxUnit()));
			if (headerItem != null)
				fluxUnit = headerItem.value;
			else
				errors.fluxUnitError = "The FITS keyword \""+mapping.getFluxUnit()+"\" doesn't exist !";
		}else
			fluxUnit = FitsMapping.extractValue(mapping.getFluxUnit());

		// Date:
		if (header != null && FitsMapping.isFitsKeyword(mapping.getDate())){
			HeaderFits.HeaderItem headerItem = header.getKeywordItem(FitsMapping.extractFitsKeywordName(mapping.getDate()));
			if (headerItem != null)
				date = headerItem.value;
			else
				errors.dateError = "The FITS keyword \""+mapping.getDate()+"\" doesn't exist !";
		}else
			date = FitsMapping.extractValue(mapping.getDate());

		// Check the current Characterization:
		checkCharacterization(/*false*/);

		return true;
	}

	/**
	 * Reloads the associated FITS mapping file (if any).
	 * 
	 * @return	<i>true</i> if the Characterization has been successfully loaded,
	 * 			<i>false</i> otherwise.
	 * 
	 * @see #loadCharacterization(FitsMapping)
	 */
	public boolean reloadMapping(){
		return loadCharacterization(mappingFile);
	}

	/**
	 * <p>Checks the Characterization currently stored in this object.</p>
	 * <p>Here are the applied checks:</p>
	 * <ul>
	 * 	<li>spectralCalibration != null</li>
	 * 	<li>wavelength >= 0</li>
	 * 	<li>wavelength unit != null and known by the SEDPlugin or can be converted into a known unit</li>
	 * 	<li>fluxCalibration != null</li>
	 * 	<li>flux unit != null and known by the SEDPlugin or can be converted into a known unit</li>
	 * </ul>
	 * 
	 * <p><i><u>note:</u> the calibration status (for the spectral and the flux axis) is checked ONLY
	 * by {@link CharacParser} or {@link FitsMapping}. That's why it is better to never clear the errors.</i></p>
	 * 
	 * @return						<i>true</i> if this Characterization is correct,
	 * 								<i>false</i> if there is at least one error.
	 * 
	 * @see #isCorrect()
	 */
	protected boolean checkCharacterization(/*boolean clearPreviousErrors*/){
		/*// Clear all former errors:
		if (clearPreviousErrors)
			errors.reset();*/

		// Spectral calibration:
		if ((spectralCalibration == null || spectralCalibration.trim().length() == 0) && errors.wlCalibrationError == null)
			errors.wlCalibrationError = "The spectral axis must be calibrated !";

		// Wavelength value:
		if (wavelength < 0 && errors.wlValueError == null)
			errors.wlValueError = "The wavelength must be positive or null.";

		// Wavelength unit:
		if (errors.wlUnitError == null){
			if (wavelengthUnit == null)
				errors.wlUnitError = "The wavelength unit can not be null.";
			else if (ToolBox.getXConverter(wavelengthUnit) == null)
				errors.wlUnitError = "The wavelength unit \""+wavelengthUnit+"\" is unknown.";
		}

		// Flux calibration:
		if ((fluxCalibration == null || fluxCalibration.trim().length() == 0) && errors.fluxCalibrationError == null)
			errors.fluxCalibrationError = "The flux axis must be calibrated !";

		// Flux unit:
		if (errors.fluxUnitError == null){
			if (fluxUnit == null)
				errors.fluxUnitError = "The flux unit can not be null.";
			else if (!ToolBox.isYUnit(fluxUnit))
				errors.fluxUnitError = "The flux unit \""+fluxUnit+"\" is unknown.";
			fluxUnit = ToolBox.normalizeYUnit(fluxUnit);
		}

		// Date:
		// TODO [OPTIONAL] It would be better if the date is ensured to be in MJD and so a double variable !

		return isCorrect();
	}

	/**
	 * Tells whether this Characterization is correct or not.
	 * That's to say, whether there is no error.
	 * 
	 * @return	<i>true</i> if there is no error, <i>false</i> otherwise.
	 */
	public boolean isCorrect(){
		return errors == null || errors.isCharacCorrect();
	}

	/**
	 * Tells whether this Characterization has been fetched from a full Characterization file (.xml)
	 * or from a FITS mapping file (.map).
	 * 
	 * @return	<i>true</i> if this Characterization has been fetched from a full Characterization file,
	 * 			<i>false</i> otherwise.
	 */
	public boolean isFullCharacterization(){
		return mappingFile == null;
	}

	/**
	 * <p>Gets the date of the last modification of the Characterization/FITS mapping file used by this
	 * {@link Characterization} instance.</p>
	 * 
	 * <p><i><u>note:</u> the value is 0L if the Characterization/FITS mapping file is not local.</i></p>
	 * 
	 * @return		Date of the last modification of used local file, or 0L if the used file is distant.
	 */
	public long getFileLastModified(){
		return fileLastModified;
	}


	/* ******************* */
	/* GETTERS AND SETTERS */
	/* ******************* */
	/**
	 * @return The spectral calibration.
	 */
	public final String getSpectralCalibration() {
		return spectralCalibration;
	}

	/**
	 * @return The wavelength.
	 */
	public final double getWavelength() {
		return wavelength;
	}

	/**
	 * @return The wavelengthUnit.
	 */
	public final String getWavelengthUnit() {
		return wavelengthUnit;
	}

	/**
	 * @return The flux calibration.
	 */
	public final String getFluxCalibration() {
		return fluxCalibration;
	}

	/**
	 * @return The fluxUnit.
	 */
	public final String getFluxUnit() {
		return fluxUnit;
	}

	/**
	 * @return The date.
	 */
	public final String getDate() {
		return date;
	}

	/**
	 * @return The characUrl.
	 */
	public final URL getCharacUrl() {
		return characUrl;
	}

	/**
	 * @return The mappingFile.
	 */
	public final FitsMapping getMappingFile() {
		return mappingFile;
	}

	/**
	 * @return The image.
	 */
	public final AladinImage getImage() {
		return image;
	}

	/**
	 * @return All the Characterization errors.
	 */
	public final CharacErrors getErrors(){
		return errors;
	}
}
