package sedplugin.graphe;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.util.Vector;

import sedplugin.sed.PointComparator;
import sedplugin.sed.SED;
import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;

/**
 * Generic frame of reference.<BR />
 * <ul>
 * 	<li>Manages all graphic preferences (title, grid, legend, their visibility, font, mouse position as cursor, ...).</li>
 * 	<li>Allows extended class to custom drawing of all these graph preferences.</li>
 * 	<li>Graphic options can be added by extension of the {@link AGraphFrame#drawOthers(Graphics)} method.</li>
 * 	<li>Computes scaling parameters and allows customization of gaps with {@link AGraphFrame#computeGaps(FontMetrics)},
 * 		{@link AGraphFrame#computeOriginGaps(FontMetrics)} and {@link AGraphFrame#computeShifts(FontMetrics)} methods.</li>
 * </ul>
 * {@link FrameOfReference} is a good example of implementation of AGraphFrame.
 *
 * @see FrameOfReference
 *
 * @author Gr&eacute;gory Mantelet
 * @version 04/2012
 */
public abstract class AGraphFrame {
	/* CONSTANTS */
	/** Constant to specify the type of axis scale. Linear for both axis. */
	public static final int XY_LINEAR = 0;
	/** Constant to specify the type of axis scale. Logarithmic scale only for abscissa axis. */
	public static final int X_LOG = 1;
	/** Constant to specify the type of axis scale. Logarithmic scale only for ordinate axis. */
	public static final int Y_LOG = 2;
	/** Constant to specify the type of axis scale. Logarithmic scale for both axis. */
	public static final int XY_LOG = 3;

	/* UNITS */
	/** Abscissa unit to use. */
	protected XAxis xUnit;

	/** Ordinate unit to use. */
	protected YAxis yUnit;

	/* GRAPH PREFERENCES */
	/** Color of the frame. <i>Color.BLACK</i> by default. */
	protected Color color = Color.BLACK;

	/** The stroke to use for the whole frame of reference. */
	public Stroke stroke = new BasicStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);

	/** Number of values to display on the abscissa axis. */
	protected int nbXIntervals = 10;

	/** Number of values to display on the ordinate axis. */
	protected int nbYIntervals = 10;

	/** Indicates whether the grid must be visible or not. */
	public boolean gridVisible = false;

	/** The font to use in the whole graph. */
	public Font font = null;

	/** The title of the graph. */
	public String title = "";

	/** The font of the graph title. */
	public Font titleFont;

	/** Indicates whether the title must be visible or not. */
	public boolean titleVisible = false;

	/** Indicates whether the legend box must be visible or not. */
	public boolean legendVisible = false;

	/** Indicates whether the axis title must be visible or not. */
	public boolean axisTitleVisible = true;

	/** Indicates whether the ticks value must be visible or not. */
	public boolean axisValuesVisible = true;

	/** Indicates whether the mouse position must be visible or not. */
	public boolean mousePositionVisible = false;

	/* MOUSE MANAGEMENT */
	/** Coordinates of the mouse into the graph panel. */
	protected double[] mousePosition = {0,0};

	/** The used listener for all draggable objects like the title and the legend box. */
	protected DragFrameMouseListener mouseListener = null;

	/* SCALE PARAMETERS */
	/** The type of scale to use. By default, {@link AGraphFrame#XY_LINEAR}. */
	protected int scaleType = XY_LINEAR;

	/** Height of the graph panel. */
	private double height = 0;

	/** Width of the graph panel. */
	private double width = 0;

	/** Extremes of all abscissa values to draw. */
	public double[] xExtremas;

	/** Extremes of all ordinate values to draw. */
	public double[] yExtremas;

	/** Abscissa scale factor. */
	private double xScale;

	/** Ordinate scale factor. */
	private double yScale;

	/** Value of one tick of the abscissa. */
	private double xTick;

	/** Value of one tick of the ordinate. */
	private double yTick;

	/** Horizontal gap between the border of the graph panel and the border of the graph frame. */
	private double xGap;

	/** Vertical gap between the border of the graph panel and the border of the graph frame. */
	private double yGap;

	/** Horizontal gap between the border of the graph frame and the abscissa axis. */
	private double xOriginGap;

	/** Vertical gap between the border of the graph frame and the ordinate axis. */
	private double yOriginGap;

	/** Horizontal gap between the abscissa axis and the minimum abscissa value. */
	private double xShift;

	/** Vertical gap between the ordinate axis and the minimum ordinate value.  */
	private double yShift;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Construct a frame of reference thanks to the specified units.
	 * @param abscissaUnit	The abscissa unit to use.
	 * @param ordinateUnit	The ordinate unit to use.
	 */
	public AGraphFrame(XAxis abscissaUnit, YAxis ordinateUnit){
		xUnit = abscissaUnit;
		yUnit = ordinateUnit;
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	public void setAxes(final XAxis x, final YAxis y){
		xUnit = x;
		yUnit = y;
	}

	/**
	 * Sets the drag listener this graph must use.
	 * @param listener	A DragFrameMouseListener instance.
	 */
	public void setDragListener(DragFrameMouseListener listener){
		mouseListener = listener;
	}

	/**
	 * Width available to draw the graph.
	 * @return	Width of the graph area.
	 */
	public final double getWidth(){
		return width;
	}

	/**
	 * Height available to draw the graph.
	 * @return	Height of the graph area.
	 */
	public final double getHeight(){
		return height;
	}

	/**
	 * Value of one tick of the abscissa axis.
	 * @return The xTick.
	 */
	public final double getXTick() {
		return xTick;
	}

	/**
	 * Value of one tick of the ordinate axis.
	 * @return The yTick.
	 */
	public final double getYTick() {
		return yTick;
	}

	/**
	 * Gets the abscissa gap between the graph and the JPanel's border.
	 * @return	Abscissa gap.
	 */
	public final double getXGap(){
		return xGap;
	}

	/**
	 * Gets the ordinate gap between the graph and the JPanel's border.
	 * @return	Ordinate gap.
	 */
	public final double getYGap(){
		return yGap;
	}

	/**
	 * Gets the gap between the beginning of the abscissa axis and the origin.
	 * @return	Gap before origin.
	 */
	public final double getXOriginGap(){
		return xOriginGap;
	}

	/**
	 * Gets the gap between the beginning of the ordinate axis and the origin.
	 * @return	Gap before origin.
	 */
	public final double getYOriginGap(){
		return yOriginGap;
	}

	/**
	 * Gets the abscissa shift/gap between the origin and the first drawn abscissa value.
	 * @return	Abscissa shift.
	 */
	public final double getXShift(){
		return xShift;
	}

	/**
	 * Gets the ordinate shift/gap between the origin and the first drawn ordinate value.
	 * @return	Ordinate shift.
	 */
	public final double getYShift(){
		return yShift;
	}

	/**
	 * Gets the drawn mouse position.
	 * @return	The mouse position.
	 */
	public final double[] getMousePosition(){
		return mousePosition;
	}

	/**
	 * Sets the mouse position to draw.
	 * @param pos	The position to draw.
	 */
	public final void setMousePosition(double[] pos){
		mousePosition = pos;
	}

	/**
	 * Gets the color used by the whole graph.
	 * @return	Frame of reference color.
	 */
	public final Color getColor(){
		return color;
	}

	/**
	 * Sets the color used by the whole graph.
	 * @param c	The new color of frame of reference.
	 * <i><b>Note:</b> If the parameter is 'null', no modification will be done.</i>
	 */
	public final void setColor(Color c){
		if (c != null)
			color = c;
	}

	/**
	 * Gets the number of abscissa intervals (the grid will be horizontally divided thanks to this number).
	 * @return	The number of abscissa intervals.
	 */
	public final int getNbXIntervals(){
		return nbXIntervals;
	}

	/**
	 * Sets the number of abscissa intervals (the grid will be horizontally divided thanks to this number).
	 * @param nbIntervals	The new number of abscissa intervals.
	 */
	public final void setNbXIntervals(int nbIntervals){
		nbXIntervals = nbIntervals;
	}

	/**
	 * Gets the number of ordinate intervals (the grid will be vertically divided thanks to this number).
	 * @return	The number of ordinate intervals.
	 */
	public final int getNbYIntervals(){
		return nbYIntervals;
	}

	/**
	 * Sets the number of ordinate intervals (the grid will be vertically divided thanks to this number).
	 * @param nbIntervals	The new number of ordinate intervals.
	 */
	public final void setNbYIntervals(int nbIntervals){
		nbYIntervals = nbIntervals;
	}

	public final double getXScale(){
		return xScale;
	}

	public final double getYScale(){
		return yScale;
	}

	/**
	 * Gets the type of scale: linear or logarithmic.
	 * @return	XY_LINEAR, X_LOG, Y_LOG or XY_LOG
	 * @see AGraphFrame#XY_LINEAR
	 * @see AGraphFrame#X_LOG
	 * @see AGraphFrame#Y_LOG
	 * @see AGraphFrame#XY_LOG
	 */
	public final int getScaleType(){
		return scaleType;
	}

	/**
	 * Sets the type of scale: linear or logarithmic.
	 * @param type	The scale type to apply.
	 * <i><b>Note:</b> Nothing will be done if 'type' doesn't correspond to predefined type.</i>
	 * @see AGraphFrame#XY_LINEAR
	 * @see AGraphFrame#X_LOG
	 * @see AGraphFrame#Y_LOG
	 * @see AGraphFrame#XY_LOG
	 */
	public final void setScaleType(int type){
		if (type == XY_LINEAR || type == X_LOG || type == Y_LOG || type == XY_LOG)
			scaleType = type;
	}

	/* *************** */
	/* COMPUTE METHODS */
	/* *************** */
	/**
	 * Convert [xValue,yValue] into [xCoord,yCoord] coordinates in the Graph.
	 * @param xValue	Abscissa value.
	 * @param yValue	Ordinate value.
	 * @return			The [xCoord,yCoord] coordinates.
	 */
	public final double[] getCoordinates(double xValue, double yValue){
		// Linear scale case:
		double xCoord = (xValue-xExtremas[0])*xScale + xGap + xOriginGap + xShift;
		double yCoord = (height-yGap-yOriginGap-yShift)-(yValue-yExtremas[0])*yScale;

		// Logarithmic scale case:
		if (scaleType == X_LOG || scaleType == XY_LOG){
			double minLog10 = (xExtremas[0]<=0)?Double.NaN:log10(xExtremas[0]);
			xCoord = xGap + xOriginGap + xShift + ((xValue <= 0)?Double.NaN:((log10(xValue)-minLog10)*xScale));
		}
		if (scaleType == Y_LOG || scaleType == XY_LOG){
			double minLog10 = (yExtremas[0]<=0)?Double.NaN:log10(yExtremas[0]);
			yCoord = (height-yGap-yOriginGap-yShift) - ((yValue <= 0)?Double.NaN:((log10(yValue)-minLog10)*yScale));
		}

		return new double[]{xCoord,yCoord};
	}

	/**
	 * Convert [xCoord,yCoord] coordinates in the Graph into [xValue,yValue] values.
	 * @param xCoord	Abscissa.
	 * @param yCoord	Ordinate.
	 * @return			The [xValue,yValue] values.
	 */
	public final double[] getValues(double xCoord, double yCoord){
		// Linear scale case:
		double xValue = (xScale == 0)?xExtremas[0]:(((xCoord-xGap-xOriginGap-xShift)/xScale)+xExtremas[0]);
		double yValue = (yScale == 0)?yExtremas[0]:(((height-yGap-yOriginGap-yShift)-yCoord)/yScale)+yExtremas[0];

		// Logarithmic scale case:
		if (scaleType == X_LOG || scaleType == XY_LOG){
			if (xScale == 0)
				xValue = (xExtremas[1] <= 0)?Double.NaN:xExtremas[0];
			else
				xValue = Math.pow(10, (log10(xExtremas[0])+(xCoord-xGap-xOriginGap-xShift)/xScale));
		}
		if (scaleType == Y_LOG || scaleType == XY_LOG){
			if (yScale == 0)
				yValue = (yExtremas[1] <= 0)?Double.NaN:yExtremas[0];
			else
				yValue = Math.pow(10, (log10(yExtremas[0])+(height-yGap-yOriginGap-yShift-yCoord)/yScale));
		}

		return new double[]{xValue, yValue};
	}

	/**
	 * Compute scale factors and axis limits.
	 * @param fm	The font metrics used by the graph to draw strings.
	 **/
	protected final void computeScale(FontMetrics fm){
		// COMPUTE GAPS:
		double[] res = computeGaps(fm);
		xGap = res[0]; yGap = res[1];

		// COMPUTE ORIGIN GAPS:
		res = computeOriginGaps(fm);
		xOriginGap = res[0]; yOriginGap = res[1];

		// COMPUTE SHIFTS:
		res = computeShifts(fm);
		xShift = res[0]; yShift = res[1];

		// WIDTH AND HEIGHT REALLY USABLE:
		double graphWidth = width - 2*xGap - xOriginGap - 2*xShift;
		double graphHeight = height - 2*yGap - yOriginGap - 2*yShift;

		// ORDINATE AXIS:
		// Compute the width of one tick:
		yTick = computeTick(yExtremas[0], yExtremas[1]);
		// Compute extremes:
		yExtremas[0] = floorValue(yExtremas[0], yTick);
		yExtremas[1] = ceilValue(yExtremas[1], yTick);
		yExtremas[2] = roundMinLog(yExtremas[2]);
		// logarithmic scale case (log10):
		if (scaleType == Y_LOG || scaleType == XY_LOG){
			if (yExtremas[1] > 0 && yExtremas[0] <= 0) yExtremas[0] = yExtremas[2];
			if (yExtremas[1] <= 0 || yExtremas[0] == yExtremas[1])
				yScale = 0;
			else
				yScale = graphHeight / (log10(yExtremas[1])-log10(yExtremas[0]));
			// linear scale case (by default):
		}else
			yScale = (yExtremas[0]==yExtremas[1])?0:Math.abs(graphHeight / (yExtremas[1]-yExtremas[0]));

		// ABSCISSA AXIS:
		// Compute the width of one tick:
		xTick = computeTick(xExtremas[0], xExtremas[1]);
		// Compute extremes:
		xExtremas[0] = floorValue(xExtremas[0], xTick);
		xExtremas[1] = ceilValue(xExtremas[1], xTick);
		xExtremas[2] = roundMinLog(xExtremas[2]);
		// logarithmic scale case (log10):
		if (scaleType == X_LOG || scaleType == XY_LOG){
			if (xExtremas[1] > 0 && xExtremas[0] <= 0) xExtremas[0] = xExtremas[2];
			if (xExtremas[1] <= 0 || xExtremas[0] == xExtremas[1])
				xScale = 0;
			else
				xScale = graphWidth / (log10(xExtremas[1])-log10(xExtremas[0]));
			// linear scale case (by default):
		}else
			xScale = (xExtremas[0]==xExtremas[1])?0:Math.abs(graphWidth / (xExtremas[1]-xExtremas[0]));
	}

	protected final static double computeTick(double min, double max){
		double minTemp = signum(min)*Math.pow(10, round(log10(Math.abs(min))));
		double maxTemp = signum(max)*Math.pow(10, round(log10(Math.abs(max))));
		double tick = (min<0 && max>0)?Math.abs(maxTemp + minTemp):Math.abs(maxTemp - minTemp);
		if (tick <= 0)
			tick = round(log10(Math.abs(min)))-1;
		else
			tick = Math.ceil(log10(tick))-1;
		return Math.pow(10, tick);
	}

	protected final static double floorValue(double value, double tickSize){
		return (value<=0)?(Math.signum(value)*Math.ceil(Math.abs(value)/tickSize)*tickSize):(Math.floor(value/tickSize)*tickSize);
	}

	protected final static double ceilValue(double value, double tickSize){
		return (value<=0)?(Math.signum(value)*Math.floor(Math.abs(value)/tickSize)*tickSize):(Math.ceil(value/tickSize)*tickSize);
	}

	protected final static double roundValue(double value, double tickSize){
		double power = log10(tickSize);
		if (!Double.isNaN(power) && power < 0)
			power = Math.pow(10, Math.floor(power));
		else
			power = 1;
		return round(value/power)*power;
	}

	protected final static double roundMinLog(double value){
		if (value <= 0)
			return Double.NaN;
		double power = Math.pow(10, Math.floor(log10(value)));
		return Math.floor(value/power)*power;
	}

	/* ************ */
	/* DRAW METHODS */
	/* ************ */
	/**
	 * Draw the whole frame of reference.
	 * @param g			The Graphics object to use.
	 * @param series	Series to draw.
	 * @param panWidth	Maximum width available.
	 * @param panHeight	Maximum height available.
	 */
	public final void draw(Graphics g, Vector<Serie> series, double panWidth, double panHeight){
		// Set graph width and height:
		width = panWidth;
		height = panHeight;

		// Extract minimum and maximum values for each axis:
		xExtremas = new double[]{Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY};
		yExtremas = new double[]{Double.POSITIVE_INFINITY, Double.NEGATIVE_INFINITY, Double.POSITIVE_INFINITY};
		for(Serie s : series){
			// get abscissa extremas:
			SED sed = s.getSED();
			double[] tmp = sed.getXExtremas();
			if (tmp[0] < xExtremas[0]) xExtremas[0] = tmp[0];
			if (tmp[1] > xExtremas[1]) xExtremas[1] = tmp[1];
			if (tmp[2] < xExtremas[2]) xExtremas[2] = tmp[2];
			// get ordinate extremas:
			tmp = sed.getYExtremas();
			if (tmp[0] < yExtremas[0]) yExtremas[0] = tmp[0];
			if (tmp[1] > yExtremas[1]) yExtremas[1] = tmp[1];
			if (tmp[2] < yExtremas[2]) yExtremas[2] = tmp[2];
		}

		// Set color:
		g.setColor(color);

		// Set stroke style:
		((Graphics2D)g).setStroke(stroke);

		// Set the font to use:
		if (font == null)
			font = g.getFont();

		// Compute scale parameters:
		computeScale(g.getFontMetrics());

		// Draw the frame of reference:
		drawFrame(g);

		// Draw the scale (axis ticks/values) and grid:
		drawScale(g);

		// Draw some graph information:
		if (axisTitleVisible) 		drawAxisTitle(g);
		//if (titleVisible) 			drawTitle(g);
		//if (legendVisible)			drawLegend(g, series);
		//if (mousePositionVisible)	drawMousePosition(g, mousePosition, panColor);

		drawOthers(g);
	}

	/**
	 * Draw the scale (ticks, values and grid).
	 * @param g	The Graphics object to use.
	 */
	protected final void drawScale(Graphics g){
		Graphics2D g2 = (Graphics2D)g;
		g2.setFont(g2.getFont().deriveFont((float)10));
		Color gridColor = color;
		int diffColor = Color.LIGHT_GRAY.getRed()-Color.BLACK.getRed();
		int red = gridColor.getRed()+diffColor; if (red > 255) red = gridColor.getRed();
		int green = gridColor.getGreen()+diffColor; if (green > 255) green = gridColor.getGreen();
		int blue = gridColor.getBlue()+diffColor; if (blue > 255) blue = gridColor.getBlue();
		gridColor = new Color(red, green, blue);

		// Abscissa axis:
		// ...logarithmic scale case:
		if (scaleType == X_LOG || scaleType == XY_LOG){
			if (xExtremas[1] > 0){
				double[] coord = new double[]{0.0, 0.0};
				double xValue = xExtremas[0], lastDrawnValue = Double.NaN;
				int i = 1;
				boolean lastDrawn = false;
				// iterate on each 10 power:
				for(int puissance = (int)Math.floor(log10(xValue)); xValue <= xExtremas[1]; puissance++){
					xValue = roundValue(xValue, xValue);
					// for each 10 power multiple:
					while(xValue <= xExtremas[1] && xValue < Math.pow(10, puissance+1)){
						// get the [x,y] coordinates:
						coord = getCoordinates(xValue, yExtremas[0]); coord[1] = height - yGap - yOriginGap;
						// draw a part of the grid:
						if (gridVisible){
							g2.setColor(gridColor);
							g2.draw(new Line2D.Double(coord[0], coord[1], coord[0], yGap));
							g2.setColor(color);
						}
						// draw the tick:
						drawXTick(g, coord);
						// if asked print their value:
						if (axisValuesVisible){
							lastDrawnValue = xValue;
							lastDrawn = drawXValue(g, coord, xValue, i, false);
						}
						i++;
						xValue += Math.pow(10, puissance);
						xValue = roundValue(xValue, xValue);
					}
				}
				// draw the last tick if it hasn't been done:
				if (!lastDrawn && !Double.isNaN(lastDrawnValue))
					drawXValue(g, coord, lastDrawnValue, i, true);
			}

			// ...linear scale case:
		}else{
			double xValue = xExtremas[0];
			// nb ticks = nb intervals => draw a tick for each interval multiple:
			for(int i=1; xValue<=xExtremas[1]; i++, xValue += xTick, xValue = roundValue(xValue, xTick)){
				// get the [x,y] coordinates:
				double[] coord = getCoordinates(xValue, yExtremas[0]); coord[1] = height - yGap - yOriginGap;
				// draw the grid:
				if (gridVisible){
					g2.setColor(gridColor);
					g2.draw(new Line2D.Double(coord[0], coord[1], coord[0], yGap));
					g2.setColor(color);
				}
				// draw the tick:
				drawXTick(g, coord);
				// if asked print their value:
				if (axisValuesVisible)
					drawXValue(g, coord, xValue, i, false);
			}
		}

		// Ordinate axis:
		// ...logarithmic scale case:
		if (scaleType == Y_LOG || scaleType == XY_LOG){
			if (yExtremas[1] > 0){
				double[] coord = new double[]{0.0, 0.0};
				double yValue = yExtremas[0], lastDrawnValue = Double.NaN;
				int i=1;
				boolean lastDrawn = false;
				// draw all values until yExtremas[1] and by power 10:
				for(int puissance = (int)Math.floor(log10(yValue)); yValue <= yExtremas[1]; puissance++){
					yValue = roundValue(yValue, yValue);
					while(yValue <= yExtremas[1] && yValue < Math.pow(10, puissance+1)){
						// get the [x,y] coordinates:
						coord = getCoordinates(xExtremas[0], yValue); coord[0] = xGap + xOriginGap;
						// draw the grid:
						if (gridVisible){
							g2.setColor(gridColor);
							g2.draw(new Line2D.Double(coord[0], coord[1], width-xGap, coord[1]));
							g2.setColor(color);
						}
						// draw the tick:
						drawYTick(g, coord);
						// if asked print their value:
						if (axisValuesVisible){
							lastDrawnValue = yValue;
							lastDrawn = drawYValue(g, coord, yValue, i, false);
						}
						i++;
						yValue += Math.pow(10, puissance);
						yValue = roundValue(yValue, yValue);
					}
				}
				// draw the last tick if it hasn't been done:
				if (!lastDrawn && !Double.isNaN(lastDrawnValue))
					drawYValue(g, coord, lastDrawnValue, i, true);
			}

			// ...linear scale case:
		}else{
			double yValue = yExtremas[0];
			// nb ticks = nb intervals => draw a tick for each interval multiple:
			for(int i=1; yValue<=yExtremas[1]; i++, yValue += yTick, yValue = roundValue(yValue, yTick)){
				// get the [x,y] coordinates:
				double[] coord = getCoordinates(xExtremas[0], yValue); coord[0] = xGap + xOriginGap;
				// draw the grid:
				if (gridVisible){
					g2.setColor(gridColor);
					g2.draw(new Line2D.Double(coord[0], coord[1], width-xGap, coord[1]));
					g2.setColor(color);
				}
				// draw the tick:
				drawYTick(g, coord);
				// if asked print their value:
				if (axisValuesVisible)
					drawYValue(g, coord, yValue, i, false);
			}
		}

		g2.setFont(g2.getFont().deriveFont((float)12));
	}

	public final static double round(double val){
		double floor = Math.abs(val-Math.floor(val));
		double ceil = Math.abs(Math.ceil(val)-val);
		if (PointComparator.compare(floor, ceil) < 0)
			return Math.floor(val);
		else
			return Math.ceil(val);
	}


	public final static double log10(double val){
		return Math.log(val)/Math.log(10);
	}

	public final static double signum(double val){
		return (val < 0)?-1:1;
	}

	/* **************** */
	/* ABSTRACT METHODS */
	/* **************** */
	/**
	 * Computes xGap (the horizontal gap between the graph and the JPanel's border) et yGap (the vertical gap between the graph and the JPanel's border).
	 * @return	{xGap,yGap}
	 */
	protected abstract double[] computeGaps(FontMetrics fm);

	/**
	 * Computes gaps between the beginning of the axis and the origin.
	 * @return	Gaps before origin: [xOriginGap, yOriginGap].
	 */
	protected abstract double[] computeOriginGaps(FontMetrics fm);

	/**
	 * Computes xShift (the horizontal gap between the origin and the first drawn abscissa value) et yShift (the ordinate gap between the origin and the first drawn ordinate value).
	 * @return	{xShift,yShift}
	 */
	protected abstract double[] computeShifts(FontMetrics fm);

	/**
	 * Draw axis and frame of reference of the graph.
	 * @param g	The Graphics object to use.
	 */
	protected abstract void drawFrame(Graphics g);

	/**
	 * Draw one abscissa tick.
	 * @param g			The Graphics object to use.
	 * @param coord		JPanel's position of the tick.
	 */
	protected abstract void drawXTick(Graphics g, double[] coord);

	/**
	 * Draw one ordinate tick.
	 * @param g			The Graphics object to use.
	 * @param coord		JPanel's position of the tick.
	 */
	protected abstract void drawYTick(Graphics g, double[] coord);

	/**
	 * Draw one abscissa value.
	 * @param g			The Graphics object to use.
	 * @param coord		JPanel's position of the tick.
	 * @param xValue	The value to print.
	 * @param numVal	xValue is the numVal-th abscissa value to draw.
	 * @param force		Indicates whether the value must be drawn whatever are the value of the other parameters.
	 * @return 			<i>true</i> if the value has been drawn, <i>false</i> otherwise.
	 */
	protected abstract boolean drawXValue(Graphics g, double[] coord, double xValue, int numVal, boolean force);
	/**
	 * Draw one ordinate value.
	 * @param g			The Graphics object to use.
	 * @param coord		JPanel's position of the tick.
	 * @param yValue	The value to print.
	 * @param numVal	xValue is the numVal-th ordinate value to draw.
	 * @param force		Indicates whether the value must be drawn whatever are the value of the other parameters.
	 * @return 			<i>true</i> if the value has been drawn, <i>false</i> otherwise.
	 */
	protected abstract boolean drawYValue(Graphics g, double[] coord, double yValue, int numVal, boolean force);

	/**
	 * Draw title of all axis.
	 * @param g	The Graphics object to use.
	 */
	protected abstract void drawAxisTitle(Graphics g);

	/**
	 * Draw the graph title.
	 * @param g	The Graphics object to use.
	 */
	public abstract void drawTitle(Graphics g);

	/**
	 * Draw a legend box (which specifies to which serie correspond each spectrum).
	 * @param g			The Graphics object to use.
	 * @param series	Series to draw.
	 */
	public abstract void drawLegend(Graphics g, Vector<Serie> series);

	/**
	 * Draw the given mouse position on the graph.
	 * @param g					The graphics to use.
	 * @param backgroundColor	Color of the graph background.
	 */
	public abstract void drawMousePosition(Graphics g, Color backgroundColor);

	/**
	 * Draw additionnal graphics object.
	 * @param g	The Graphics object to use.
	 */
	public abstract void drawOthers(Graphics g);
}
