package sedplugin.graphe;


import java.awt.Rectangle;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.awt.geom.RectangularShape;
import java.util.Vector;

/**
 * This listener will allow the user to drag all given RectangleShapes.
 * <ul>
 * 	<li>It must be set as MouseListener and as MouseMotionListener in a Graph instance.</li>
 * 	<li>To set draggable a {@link RectangularShape} use the {@link DragFrameMouseListener#addDraggableShape(Rectangle)}. </li>
 * </ul>
 * 
 * @author Gregory Mantelet
 * @version 03/2012
 */
public class DragFrameMouseListener implements MouseListener, MouseMotionListener {

	/** Associated instance of Graph. */
	protected final Graph myGraph;

	/** List of all draggable boxes.
	 * @see RectangularShape */
	protected final Vector<Rectangle> vRectangles;

	/** Indice of the current selected rectangle. */
	protected int selectedRect = -1;

	/** Last abscissa coordinate of the selected rectangle. */
	protected int lastX = 0;

	/** Last ordinate coordinate of the selected rectangle. */
	protected int lastY = 0;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Initialize a DragFrameMouseListener with the instance of Graph that will use this listener.
	 * @param g	The Graph instance which needs this listener.
	 */
	public DragFrameMouseListener(Graph g){
		myGraph = g;
		vRectangles = new Vector<Rectangle>();
	}

	/**
	 * Initialize a DragFrameMouseListener with the instance of Graph that will use this listener and a list of rectangle shapes to manage.
	 * @param g				The associated Graph instance.
	 * @param rectangles	List of Rectangle objects to manage.
	 */
	public DragFrameMouseListener(Graph g, Vector<Rectangle> rectangles){
		myGraph = g;
		vRectangles = rectangles;
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Selects the first rectangle which contains the given position.
	 * @param x	An abcissa.
	 * @param y	An ordinate.
	 * @return	The first rectangle which contains the given position.
	 */
	private final int getPressedRectangle(int x, int y){
		selectedRect = -1;

		for(int i=0; selectedRect == -1 && i<vRectangles.size(); i++)
			if (vRectangles.get(i).contains(x, y))
				selectedRect = i;

		return selectedRect;
	}

	/**
	 * Adds another Rectangle object which must be draggable.
	 * @param rect	A Rectangle object to make draggable.
	 */
	public final void addDraggableShape(Rectangle rect){
		vRectangles.add(rect);
	}

	/* *************** */
	/* DRAG MANAGEMENT */
	/* *************** */
	/**
	 * Updates the selected Rectangle object position in function of the mouse coordinates.
	 * @param e	A mouse event (which contains current mouse coordinates).
	 */
	private final void updatePosition(MouseEvent e){
		if (selectedRect >= 0 && selectedRect < vRectangles.size()){
			Rectangle rect = vRectangles.get(selectedRect);
			rect.setLocation(lastX+e.getX(), lastY+e.getY());
			myGraph.repaint();
		}
	}

	/**
	 * Selects the first Rectangle object under the mouse position,
	 * saves the rectangle position and updates it in function of the mouse position.
	 * 
	 * @see java.awt.event.MouseListener#mousePressed(java.awt.event.MouseEvent)
	 */
	public void mousePressed(MouseEvent e) {
		if (getPressedRectangle(e.getX(), e.getY()) != -1){
			Rectangle rect = vRectangles.get(selectedRect);
			lastX = rect.x - e.getX();
			lastY = rect.y - e.getY();
			updatePosition(e);
		}
	}

	/**
	 * During a clicked mouse movement update the position of the selected rectangle (if any).
	 * 
	 * @see java.awt.event.MouseMotionListener#mouseDragged(java.awt.event.MouseEvent)
	 */
	public void mouseDragged(MouseEvent e) {
		if (selectedRect != -1)
			updatePosition(e);
	}

	/**
	 * When release a mouse click update the selected rectangle position
	 * and de-select it.
	 * 
	 * @see java.awt.event.MouseListener#mouseReleased(java.awt.event.MouseEvent)
	 */
	public void mouseReleased(MouseEvent e) {
		if (selectedRect != -1){
			updatePosition(e);
			selectedRect = -1;
		}
	}

	public void mouseClicked(MouseEvent arg0) {;}
	public void mouseEntered(MouseEvent arg0) {;}
	public void mouseExited(MouseEvent arg0) {;}
	public void mouseMoved(MouseEvent arg0) {;}

}
