package sedplugin.graphe;


import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Polygon;
import java.awt.Rectangle;
import java.awt.Stroke;
import java.awt.font.TextAttribute;
import java.awt.geom.AffineTransform;
import java.awt.geom.Line2D;
import java.text.AttributedCharacterIterator;
import java.text.DecimalFormat;
import java.util.HashMap;
import java.util.Map;
import java.util.Vector;

import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;

/**
 * A frame of reference.<BR/>
 * It's possible to make other frames of reference by extension of {@link AGraphFrame} class as this implementation.
 *
 * @see AGraphFrame
 *
 * @author Gregory Mantelet
 * @version 04/2012
 */
public class FrameOfReference extends AGraphFrame {

	/** The format to use when numbers aren't so huge. */
	public final static DecimalFormat classicFormat = new DecimalFormat("##0.##");

	/** The number format to use for big number. */
	public final static DecimalFormat scientificFormat = new DecimalFormat("0.##E0");

	/** Gap between the origin of the frame and the first abscissa value. */
	public final static double X_SHIFT = 22.0;

	/** Gap between the origin of the frame and the first ordinate value. */
	public final static double Y_SHIFT = 22.0;

	/** Arrows width. */
	public final static int ARROW_WIDTH = 12;

	/** Arrows height. */
	public final static int ARROW_HEIGHT = 8;

	/** Height (for the abscissa axis, but width for the ordinate axis) of the little line which represents one tick on an axis. */
	public final static int TICK_HEIGHT = 6;

	/** Click area of the title. It allows the user to change the position of the title. */
	protected RepositionnableRectangle rectTitle;

	/** Click area of the legend box. It allows the user to change the position of the legend box. */
	protected RepositionnableRectangle rectLegend;

	/** Gap between the border of the legend and its content. */
	public final static int LEGEND_GAP = 6;

	protected int nbXTicksBeforeValue = 2;
	protected int nbYTicksBeforeValue = 2;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Construct a frame of reference thanks to the specified units.
	 * @param abscissaUnit	The abscissa unit to use.
	 * @param ordinateUnit	The ordinate unit to use.
	 * @see AGraphFrame#AGraphFrame(AUnit, AUnit)
	 */
	public FrameOfReference(XAxis abscissaUnit, YAxis ordinateUnit) {
		super(abscissaUnit, ordinateUnit);

		Map<AttributedCharacterIterator.Attribute, Object> attributesMap = new HashMap<AttributedCharacterIterator.Attribute, Object>();
		attributesMap.put(TextAttribute.FAMILY, "Serif");
		attributesMap.put(TextAttribute.POSTURE, TextAttribute.POSTURE_OBLIQUE);
		attributesMap.put(TextAttribute.UNDERLINE, TextAttribute.UNDERLINE_ON);
		attributesMap.put(TextAttribute.WEIGHT, TextAttribute.WEIGHT_BOLD);
		attributesMap.put(TextAttribute.SIZE, new Float(16));
		titleFont = new Font(attributesMap);

		rectTitle = new RepositionnableRectangle(0,0,0,0);
		rectLegend = new RepositionnableRectangle(0,0,0,0);
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Convert the abscissa value into the best formatted string.
	 * @param xValue	Abscissa value.
	 * @return			The corresponding string with the most adapted number format.
	 */
	public final static String getXValueString(double xValue){
		return (xValue==0)?"0":whichFormat(xValue).format(xValue);
	}

	/**
	 * Convert the ordinate value into the best formatted string.
	 * @param yValue	Ordinate value.
	 * @return			The corresponding string with the most adapted number format.
	 */
	public final static String getYValueString(double yValue){
		return (yValue==0)?"0":whichFormat(yValue).format(yValue);
	}

	/**
	 * Gets the format to use considering the given value.
	 * @param value	An example of value.
	 * @return		The format to use.
	 */
	public final static DecimalFormat whichFormat(double value){
		DecimalFormat df = classicFormat;
		if (((int)value*100) == 0 || ((int)value/1000) != 0)
			df = scientificFormat;

		return df;
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	/**
	 * Makes draggable title and legend boxes.
	 * @see sedplugin.graphe.AGraphFrame#setDragListener(sedplugin.graphe.DragFrameMouseListener)
	 */
	@Override
	public void setDragListener(DragFrameMouseListener listener) {
		super.setDragListener(listener);
		mouseListener.addDraggableShape(rectTitle);
		mouseListener.addDraggableShape(rectLegend);
	}

	@Override
	protected double[] computeGaps(FontMetrics fm) {
		return new double[]{fm.getHeight()*5/3, fm.getHeight()*5/3};
	}

	@Override
	protected double[] computeOriginGaps(FontMetrics fm){
		DecimalFormat[] dfs = {whichFormat(yExtremas[0]), whichFormat(yExtremas[1])};
		int x = 0;
		if (dfs[0].equals(classicFormat) || dfs[1].equals(classicFormat))
			x = 2+fm.stringWidth(classicFormat.toLocalizedPattern())-2*classicFormat.toLocalizedPattern().length();
		else
			x = 2+fm.stringWidth(scientificFormat.toLocalizedPattern())-2*scientificFormat.toLocalizedPattern().length();
		return new double[]{x, fm.getHeight()};
	}

	@Override
	protected double[] computeShifts(FontMetrics fm) {
		return new double[]{fm.stringWidth(scientificFormat.toLocalizedPattern())/2, fm.getHeight()*4/3};
	}

	@Override
	protected void drawFrame(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;

		double[] origin = {getXGap()+getXOriginGap(), getHeight()-getYGap()-getYOriginGap()};

		// Abscissa axis:
		// ...axis
		g2.draw(new Line2D.Double(origin[0]-getXOriginGap(), origin[1], getWidth()-getXGap(), origin[1]));
		// ...arrow
		Polygon xArrow = new Polygon(new int[] {(int)(getWidth()-getXGap()), (int)(getWidth()-getXGap()+ARROW_WIDTH), (int)(getWidth()-getXGap())},
				new int[] {(int)(origin[1]+(ARROW_HEIGHT/2)), (int)origin[1], (int)(origin[1]-(ARROW_HEIGHT/2))}, 3);
		g2.fillPolygon(xArrow);

		// Ordinate axis:
		// ...axis
		g2.draw(new Line2D.Double(origin[0], origin[1]+getYOriginGap(), origin[0], getYGap()));
		// ...arrow
		Polygon yArrow = new Polygon(new int[] {(int)(origin[0]-(ARROW_HEIGHT/2)), (int)origin[0], (int)(origin[0]+(ARROW_HEIGHT/2))},
				new int[] {(int)getYGap(), (int)(getYGap()-ARROW_WIDTH), (int)getYGap()}, 3);
		g2.fillPolygon(yArrow);

		// Compute the frequency of ticks to be valued:
		if (getScaleType() == X_LOG || getScaleType() == XY_LOG)
			nbXTicksBeforeValue = 2;
		else{
			double temp = getCoordinates(xExtremas[0]+getXTick(), 0)[0]-getCoordinates(xExtremas[0], 0)[0];
			if (temp > 0)
				temp = Math.ceil(g.getFontMetrics().stringWidth(whichFormat(xExtremas[1]).format(xExtremas[1])) / temp);
			if (temp <= 0)
				nbXTicksBeforeValue = 0;
			else
				nbXTicksBeforeValue = (int)temp;
		}

		if (getScaleType() == Y_LOG || getScaleType() == XY_LOG)
			nbYTicksBeforeValue = 2;
		else{
			double temp = getCoordinates(0, yExtremas[0])[1]-getCoordinates(0, yExtremas[0]+getYTick())[1];
			if (temp > 0)
				temp = Math.ceil(g.getFontMetrics().getHeight() / temp);
			if (temp <= 0)
				nbYTicksBeforeValue = 0;
			else
				nbYTicksBeforeValue = (int)temp;
		}
	}

	@Override
	protected void drawXTick(Graphics g, double[] coord) {
		((Graphics2D)g).draw(new Line2D.Double(coord[0], coord[1]-TICK_HEIGHT/2, coord[0], coord[1]+TICK_HEIGHT/2));
	}

	@Override
	protected boolean drawXValue(Graphics g, double[] coord, double xValue, int numVal, boolean force) {
		boolean draw = true;
		FontMetrics fm = g.getFontMetrics();
		String value = getXValueString(xValue);

		if (getScaleType() == X_LOG || getScaleType() == XY_LOG)
			draw = force || (numVal == 1) || (xValue == Math.pow(10, Math.round(AGraphFrame.log10(Math.abs(xValue)))));
		else
			draw = force || (nbXTicksBeforeValue == 1 || (nbXTicksBeforeValue != 0 && numVal%nbXTicksBeforeValue != 0));

		if (draw)
			((Graphics2D)g).drawString(value, (int)(coord[0]-fm.stringWidth(value)/2), (int)(coord[1]+fm.getHeight()));
		return draw;
	}

	@Override
	protected void drawYTick(Graphics g, double[] coord) {
		((Graphics2D)g).draw(new Line2D.Double(coord[0]-TICK_HEIGHT/2, coord[1], coord[0]+TICK_HEIGHT/2, coord[1]));
	}

	@Override
	protected boolean drawYValue(Graphics g, double[] coord, double yValue, int numVal, boolean force) {
		boolean draw = true;
		FontMetrics fm = g.getFontMetrics();

		if (getScaleType() == Y_LOG || getScaleType() == XY_LOG)
			draw = force || (numVal == 1) || (yValue == Math.pow(10, Math.round(AGraphFrame.log10(Math.abs(yValue)))));
		else
			draw = force || (nbYTicksBeforeValue == 1 || (nbYTicksBeforeValue != 0 && numVal%nbYTicksBeforeValue != 0));

		if (draw){
			String value = getYValueString(yValue);
			((Graphics2D)g).drawString(value, (int)(coord[0]-fm.stringWidth(value)-4), (int)(coord[1]+fm.getHeight()/2));
		}
		return draw;
	}

	@Override
	protected void drawAxisTitle(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;
		FontMetrics fm = g.getFontMetrics();

		// change the stroke style:
		Font oldFont = fm.getFont();
		g2.setFont(new Font(oldFont.getFontName(), Font.BOLD, oldFont.getSize()+2));

		// Abscissa axis:
		drawXAxis(g2, fm);

		// Ordinate axis:
		AffineTransform oldAt = g2.getTransform(), at = new AffineTransform(oldAt);
		at.rotate(-Math.PI/2, 10+(fm.getHeight())/2, (int)((getHeight()/2)-getYGap()));
		g2.setTransform(at);
		drawYAxis(g2, fm);
		g2.setTransform(oldAt);

		// set the former stroke style:
		g2.setFont(oldFont);
	}

	protected void drawXAxis(Graphics2D g2, FontMetrics fm){
		String unitStr = xUnit.getQuantity().getUnicodeSymbol()+" ["+xUnit.getUnit().unicodeSymbol+"]";
		g2.drawString(unitStr, (int)(getWidth()-fm.stringWidth(unitStr))/2, (int)(getHeight()-10));
	}

	protected void drawYAxis(Graphics2D g2, FontMetrics fm){
		String unitStr = yUnit.getQuantity().getUnicodeSymbol()+" ["+yUnit.getUnit().unicodeSymbol+"]";
		int x = (20+fm.getHeight()-fm.stringWidth(unitStr))/2;
		int y = (int)((getHeight()/2)-getYGap());

		boolean intensity = false;
		if (yUnit.isQuantity(YAxis.NU_Inu) || (intensity = (yUnit.isQuantity(YAxis.INTENSITY)))){
			if (yUnit.isQuantity(YAxis.NU_Inu))
				unitStr = "\u03BDI\u03BD ["+yUnit.getUnit().unicodeSymbol+"]";
			else
				unitStr = "Intensity (I\u03BD) ["+yUnit.getUnit().unicodeSymbol+"]";
			x = (20+fm.getHeight()-fm.stringWidth(unitStr))/2;

			unitStr = (intensity) ? "Intensity (I" : "\u03BDI";
			Font oldFont = g2.getFont();
			g2.drawString(unitStr, x, y);

			x += fm.stringWidth(unitStr);
			HashMap<AttributedCharacterIterator.Attribute, Integer> subAttribute = new HashMap<AttributedCharacterIterator.Attribute, Integer>();
			subAttribute.put(TextAttribute.SUPERSCRIPT, TextAttribute.SUPERSCRIPT_SUB);
			g2.setFont(g2.getFont().deriveFont(subAttribute));
			g2.drawString("\u03BD", x, y);

			x += g2.getFontMetrics().stringWidth("\u03BD");
			g2.setFont(oldFont);
			g2.drawString((intensity ? ") " : " ")+"["+yUnit.getUnit().unicodeSymbol+"]", x, y);
		}else
			g2.drawString(unitStr, x, y);
	}

	@Override
	public void drawTitle(Graphics g) {
		if (!titleVisible)
			return;

		if (title != null){
			if (title.length() > 0){
				Graphics2D g2 = (Graphics2D)g;
				Font f = g2.getFont();
				g2.setFont(titleFont);
				FontMetrics tempFontMetrics = g2.getFontMetrics();
				if (rectTitle.getHeight() == 0.0 && rectTitle.getWidth() == 0.0)
					rectTitle.setLocation((int)(getWidth()-tempFontMetrics.stringWidth(title))/2, tempFontMetrics.getHeight()/3, getWidth(), getHeight());
				else
					rectTitle.updatePosition(getWidth(), getHeight());
				rectTitle.setSize(tempFontMetrics.stringWidth(title), tempFontMetrics.getHeight());
				(g2).drawString(title, (int)rectTitle.getX(), (int)(rectTitle.getY()+rectTitle.getHeight()));
				g2.setFont(f);
			}else
				rectTitle.setSize(0,0);
		}else
			rectTitle.setSize(0,0);
	}

	@Override
	public void drawLegend(Graphics g, Vector<Serie> series) {
		if (!legendVisible)
			return;

		Graphics2D g2 = (Graphics2D)g;
		FontMetrics fm = g.getFontMetrics();
		int nbSeries = series.size();
		int tailleMax = 0;
		Point p1, p2;
		String label;
		int tailleTemp;

		// Set the legend box position:
		if (rectLegend.getHeight() == 0.0 && rectLegend.getWidth() == 0.0)
			rectLegend.setLocation((int)getWidth()*2/3, (int)getYGap(), getWidth(), getHeight());
		else
			rectLegend.updatePosition(getWidth(), getHeight());

		/* Gets the first item position in the legend box
		 * considering LEGEND_GAP between the border of the box and items: */
		int posX = (int)rectLegend.getX() + LEGEND_GAP;
		int posY = (int)rectLegend.getY() + LEGEND_GAP + fm.getHeight();

		// Draw a legend item for each drawn serie:
		for(Serie s : series){
			// create two points to represent the current serie:
			p1=Point.createPredefinedPoint(posX, posY-fm.getHeight()/2, s, s.getPointStyle());
			p2=Point.createPredefinedPoint(posX+p1.getSide()*3, posY-fm.getHeight()/2, s, s.getPointStyle());
			// set the item color to the current serie color:
			g.setColor(s.getColor());
			p1.draw(g); p2.draw(g);
			// connect points if done for the current serie:
			if (s.arePointsConnected())
				g2.draw(new Line2D.Double(p1.getX(), p1.getY(), p2.getX(), p2.getY()));
			// draw its name:
			g2.drawString(label = s.getSED().getName(), (int)(p2.getX()+p2.getSide()+LEGEND_GAP), posY);

			// keep the item width if larger than preceding one:
			tailleTemp = (int)(p2.getSide()*4+fm.stringWidth(label)+3*LEGEND_GAP);
			if (tailleTemp > tailleMax)
				tailleMax = tailleTemp;

			// compute the ordinate of the next item:
			posY += fm.getHeight() + LEGEND_GAP;
		}

		// Adapt the legend box size to its content:
		rectLegend.setSize(tailleMax, nbSeries*fm.getHeight()+LEGEND_GAP*(nbSeries+1));
		// Set the legend box color:
		g.setColor(color);
		// Set its stroke style:
		Stroke st = g2.getStroke();
		g2.setStroke(new BasicStroke(1, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 10.0f, new float[]{2.0f}, 0.0f));
		// Draw the legend box:
		g2.draw(rectLegend);
		g2.setStroke(st);
	}

	@Override
	public void drawMousePosition(Graphics g, Color backgroundColor){
		if (!mousePositionVisible)
			return;

		Graphics2D g2 = (Graphics2D)g;

		Stroke st = g2.getStroke();
		Color oldColor = g2.getColor();

		// set the lines stroke:
		g2.setStroke(new BasicStroke(1, BasicStroke.CAP_SQUARE, BasicStroke.JOIN_ROUND));

		// set the color so that mouse position is visible on any background color:
		g2.setColor(new Color(255-backgroundColor.getRed(), 0, 0));

		// draw an horizontal line at the given ordinate:
		g2.draw(new Line2D.Double(mousePosition[0], 0, mousePosition[0], getHeight()));

		// draw a vertical line at the given abscissa:
		g2.draw(new Line2D.Double(0, mousePosition[1], getWidth(), mousePosition[1]));

		g2.setStroke(st);
		g2.setColor(oldColor);
	}

	@Override
	public void drawOthers(Graphics g) {;}

	/**
	 * Draggable rectangle.
	 *
	 * @see Rectangle
	 *
	 * @author Gregory Mantelet
	 * @version 01/2010
	 */
	protected class RepositionnableRectangle extends Rectangle {

		private static final long serialVersionUID = 1L;

		protected double panelWidth = 1;
		protected double panelHeight = 1;

		public RepositionnableRectangle(int posX, int posY, int rectWidth, int rectHeight) {
			super(posX, posY, rectWidth, rectHeight);
		}

		public void setLocation(int posX, int posY, double panWidth, double panHeight){
			if (panWidth > 0)
				panelWidth = panWidth;
			if (panHeight > 0)
				panelHeight = panHeight;

			super.setLocation(posX, posY);
		}

		public void updatePosition(double panWidth, double panHeight){
			if (panWidth > 0 && panHeight > 0){
				super.setLocation((int)(panWidth*(getX()/panelWidth)), (int)(panHeight*(getY()/panelHeight)));
				panelWidth = panWidth;
				panelHeight = panHeight;
			}
		}

	}
}
