package sedplugin.graphe;


import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Dimension;
import java.awt.Font;
import java.awt.FontMetrics;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Image;
import java.awt.RenderingHints;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.awt.image.BufferedImage;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.Vector;

import javax.swing.JLabel;
import javax.swing.JPanel;

import sedplugin.interfaceGraphique.IGraphManager;
import sedplugin.plugin.UserPreferences;

/**
 * A JPanel to draw a graph represented by some Serie objects.
 *
 * @see Serie
 * @see AGraphFrame
 * @see sedplugin.interfaceGraphique.IGraphManager
 *
 * @author Gregory Mantelet
 * @version 04/2012
 */
public class Graph extends JPanel {

	private static final long serialVersionUID = 1L;

	/** The frame of reference used to compute coordinates of points. */
	protected final AGraphFrame frameOfRef;

	/** The graph container.
	 * @see sedplugin.interfaceGraphique.IGraphManager */
	protected final IGraphManager myManager;

	/** Points to draw. */
	protected Vector<Serie> mySeries;

	/** The used MouseMotionListener. */
	protected CoordinateMouseMotionListener myMouseMotionListener;

	/** The used MouseListener. */
	protected SelectPointMouseListener myMouseListener;

	protected JPanel panWelcome;

	protected boolean yUnitKnown = true;

	protected Image mainLayer = null;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Creates a white board meaning there is no spectrum to draw.
	 * @param manager	The parent of this JPanel (a JFrame which implements IGraphManager interface).
	 * @param ref		The frame of reference to use to compute scale and so draw points of series.
	 * @see sedplugin.interfaceGraphique.IGraphManager
	 */
	public Graph(IGraphManager manager, AGraphFrame ref){
		myManager = manager;

		mySeries = new Vector<Serie>();

		frameOfRef = ref;
		DragFrameMouseListener tempListener = new DragFrameMouseListener(this);
		frameOfRef.setDragListener(tempListener);
		addMouseListener(tempListener);
		addMouseMotionListener(tempListener);

		JLabel lblWelcomeMsg = new JLabel(manager.getWelcomeMessage());
		lblWelcomeMsg.setForeground(new Color(0,128,0));
		lblWelcomeMsg.setFont(lblWelcomeMsg.getFont().deriveFont(14));

		//panWelcome = new AboutDialog.BetaPanel(); // To display an icon to mean that this plugin is in a Beta version !
		panWelcome = new JPanel();
		panWelcome.setLayout(new BorderLayout());
		panWelcome.add(lblWelcomeMsg, BorderLayout.CENTER);

		setLayout(new BorderLayout());
		add(panWelcome, BorderLayout.CENTER);

		setBackground(Color.WHITE);
		setPreferredSize(new Dimension(700, 500));

		addMouseMotionListener(myMouseMotionListener = new CoordinateMouseMotionListener());
		addMouseListener(myMouseListener = new SelectPointMouseListener());
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Indicates whether this Graph object has at least one serie to draw.
	 * @return <i>true</i> if at least one serie has been already drawn, <i>false</i> else.
	 */
	public final boolean isEmpty(){
		if (mySeries == null)
			return true;

		return mySeries.isEmpty();
	}

	/**
	 * Gets the used frame of reference.
	 * @return	The used frame of reference.
	 */
	public final AGraphFrame getFrameOfReference(){
		return frameOfRef;
	}

	/**
	 * Gets the number of managed series.
	 * @return	The number of drawn series.
	 */
	public final int getNbSeries(){
		return (mySeries == null)?0:mySeries.size();
	}

	/**
	 * Gets the serie ind-th serie drawn by this Graph object.
	 * @param ind	The index of the serie to get.
	 * @return		The corresponding Serie object.
	 */
	public final Serie getSerie(int ind){
		if (mySeries == null)
			return null;

		if (ind < 0 || ind >= mySeries.size())
			return null;

		return mySeries.get(ind);
	}

	/**
	 * Gets all drawn series.
	 * @return	All drawn series.
	 */
	public final Vector<Serie> getAllSeries(){
		return mySeries;
	}

	/* ************** */
	/* EXPORT METHODS */
	/* ************** */
	/**
	 * Export the drawn series as an image.
	 * @return The corresponding image.
	 */
	public BufferedImage exportAsImage(){
		BufferedImage image = new BufferedImage(getWidth(), getHeight(), BufferedImage.TYPE_INT_RGB);
		Graphics2D g2 = image.createGraphics();

		g2.setColor(getBackground());
		g2.fillRect(0, 0, getWidth(), getHeight());
		paintGraph(g2);

		return image;
	}

	/* ************ */
	/* DRAW METHODS */
	/* ************ */
	public final Vector<Serie> addSerie(Serie serie){
		if (serie == null || mySeries.contains(serie))
			return mySeries;

		if (panWelcome != null){
			remove(panWelcome);
			panWelcome = null;
		}
		selectPoint(null, -1);

		if (!serie.hasKnownYUnit()){
			if (yUnitKnown)
				mySeries.clear();
			else{
				final String customUnit = serie.getSED().getYAxis().getUnit().symbol;
				Iterator<Serie> it = mySeries.iterator();
				while(it.hasNext()){
					Serie s = it.next();
					if (!s.hasKnownYUnit()){
						if (!s.getSED().getYAxis().getUnit().symbol.equals(customUnit))
							it.remove();
					}else
						it.remove();
				}
			}
		}else if (!yUnitKnown)
			mySeries.clear();

		mySeries.add(serie);
		yUnitKnown = serie.hasKnownYUnit();

		update();
		myManager.drawingSeries(mySeries, yUnitKnown);

		return mySeries;
	}

	public final Vector<Serie> removeSerie(Serie serie){
		if (serie == null)
			return mySeries;

		selectPoint(null, -1);
		mySeries.remove(serie);
		update();
		myManager.drawingSeries(mySeries, yUnitKnown);
		return mySeries;
	}

	/**
	 * After painting the JPanel the frame of reference, points and legends are drawn.
	 * @see javax.swing.JComponent#paintComponent(java.awt.Graphics)
	 */
	@Override
	public void paintComponent(Graphics g) {
		Graphics2D g2 = (Graphics2D)g;

		RenderingHints rh = g2.getRenderingHints();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

		if (mainLayer == null || getWidth() != frameOfRef.getWidth() || getHeight() != frameOfRef.getHeight())
			paintGraph();

		// Draw the main layer (axes, frame, grid, series):
		g2.drawImage(mainLayer, 0, 0, null);

		if (isEnabled()){
			// Draw values of the point(s) under the mouse:
			Iterator<Point> it = myMouseMotionListener.getPoints();
			while(it.hasNext()){
				Point p = it.next();
				p.drawValues(g2, p.equals(myMouseListener.getSelectedPoint()));
			}

			// Highlight the selected point:
			if (myMouseListener.getSelectedPoint() != null)
				myMouseListener.getSelectedPoint().drawSelection(g2);

			// Draw the mouse position if needed:
			frameOfRef.drawMousePosition(g2, getBackground());
			// Draw the legend if needed:
			frameOfRef.drawLegend(g2, mySeries);
			// Draw the title if needed:
			frameOfRef.drawTitle(g2);
		}

		g2.setRenderingHints(rh);

		g.dispose();
	}

	/**
	 * Lets refreshing the drawn graph.
	 * @see JPanel#repaint()
	 */
	public final void update(){
		paintGraph();
		repaint();
	}

	protected final void paintGraph(){
		mainLayer = createImage(getWidth(), getHeight());
		paintGraph((Graphics2D)mainLayer.getGraphics());
	}

	/**
	 * Paint the graph frame and series or white board with instructions.
	 * @param g	The Graphics object to use.
	 */
	protected final void paintGraph(Graphics2D g2){
		g2.setColor(getBackground());
		g2.fillRect(0, 0, getWidth(), getHeight());

		RenderingHints rh = g2.getRenderingHints();
		g2.setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);

		// If there is a spectrum to draw...
		if (!isEmpty()){
			// ...set the units:
			frameOfRef.setAxes(mySeries.get(0).getSED().getXAxis(), mySeries.get(0).getSED().getYAxis());
			// ...draw its frame of reference:
			frameOfRef.draw(g2, mySeries, getWidth(), getHeight());
			// ...and draw its points:
			for(Serie s : mySeries)
				s.draw(g2);
		}else{	// If the graph hasn't been yet initialized, draw a white board with some instructions:
			Font ancienne = g2.getFont();
			g2.setFont(new Font(g2.getFont().getFontName(), Font.ITALIC, g2.getFont().getSize()+4));
			FontMetrics fm = g2.getFontMetrics();

			Color col = g2.getColor();
			g2.setColor(new Color(0,128,0));

			String msg = "[Select a SED in the left list or draw a region to display a SED.]";
			g2.drawString(msg, (getWidth()-fm.stringWidth(msg))/2, (getHeight()-fm.getHeight())/2);

			g2.setFont(ancienne);
			g2.setColor(col);
		}

		g2.setRenderingHints(rh);
	}

	/**
	 * Selects ({@link Point#setSelected(boolean)} method) the ind-th point of the first drawn serie.
	 * @param ind	The indice of the point to select.
	 */
	public final void selectPoint(Serie s, int pointIndex){
		if (isEnabled()){
			if (s == null || pointIndex <= -1 || pointIndex >= s.getNbPoints())
				myMouseListener.setSelectedSerie(null);
			else{
				s.setSelectedPoint(pointIndex);
				myMouseListener.setSelectedSerie(s);
			}
			repaint();
		}
	}

	/* ************************** */
	/* MOUSE MANAGEMENT LISTENERS */
	/* ************************** */
	/**
	 * When a click occurs on the JPanel, selects (if existing) the point which is at the clicked position.
	 * @author Gregory Mantelet
	 * @version 04/2012
	 */
	public class SelectPointMouseListener extends MouseAdapter {
		private Serie selectedSerie = null;
		@Override
		public void mouseClicked(MouseEvent e) {
			if (mySeries != null){
				int pointIndex = -1;
				for(int i=0; i<mySeries.size(); i++){
					Serie s = mySeries.get(i);
					if (s.selectPointAt(e.getX(), e.getY())){
						setSelectedSerie(s);
						pointIndex = s.getSelectedPointIndex();
						repaint();
						break;
					}
				}
				if (pointIndex > -1)
					myManager.highlightSeriePoint(selectedSerie, pointIndex);
				else if (selectedSerie != null){
					myManager.highlightSeriePoint(selectedSerie, -1);
					selectedSerie = null;
				}
			}
		}
		public Point getSelectedPoint(){ return (selectedSerie == null) ? null : selectedSerie.getSelectedPoint(); }
		public void setSelectedSerie(Serie s){
			if (selectedSerie != null && (s == null || !s.equals(selectedSerie)))
				selectedSerie.setSelectedPoint(-1);
			selectedSerie = s;
		};
	}

	/**
	 * When the mouse moves over the JPanel, give its new position to the parent (thanks to ISEDFrame interface) so that it can prints it.<BR />
	 * If asked, x-line and y-line are drawn.
	 * @see sedplugin.interfaceGraphique.IGraphManager#changeMousePosition(double, double, double, double)
	 * @author Gregory Mantelet
	 * @version 04/2012
	 */
	public class CoordinateMouseMotionListener extends MouseMotionAdapter {
		private ArrayList<Point> hoverPoints = new ArrayList<Point>();
		private String pointsID = "";
		public final Iterator<Point> getPoints(){
			return hoverPoints.iterator();
		}
		public final void clearPoints(){
			hoverPoints.clear();
		}
		@Override
		public void mouseMoved(MouseEvent e) {
			boolean modifications = false;
			if (!isEmpty()){
				if (frameOfRef.mousePositionVisible){
					double[] tmp = frameOfRef.getValues(e.getX(), e.getY());

					frameOfRef.setMousePosition(new double[]{e.getX(), e.getY()});
					myManager.changeMousePosition(e.getX(), e.getY(), tmp[0], tmp[1]);
					modifications = true;
				}

				if (UserPreferences.getCurrentInstance().printValuesOnOver() && mySeries != null){
					String newPointsID = "";
					hoverPoints.clear();
					for(Serie s : mySeries){
						Point p = s.getPointAt(e.getX(), e.getY());
						if (p != null){
							hoverPoints.add(p);
							newPointsID += p.hashCode();
						}
					}
					modifications = (!pointsID.equals(newPointsID)) || modifications;
					pointsID = newPointsID;
				}else{
					hoverPoints.clear();
					pointsID = "";
				}


				if (modifications)
					repaint();
			}
		}
	}

}
