package sedplugin.graphe;

import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Shape;

import sedplugin.sed.SEDPoint;

/**
 * Graphic representation of a spectrum's point ({@link spectrum.SpectrumPoint}).<BR />
 * Once initialized you can draw this point thanks to the function {@link sedplugin.graphe.Point#draw(Graphics)}.
 * 
 * @see CrossPoint
 * @see StarPoint
 * @see SquarePoint
 * @see CirclePoint
 * 
 * @author Gregory Mantelet
 * @version 03/2012
 */
public abstract class Point {

	public final static int DEFAULT_POINT = 0;
	public final static int SQUARE_POINT = 1;
	public final static int CIRCLE_POINT = 2;
	public final static int CROSS_POINT = 3;
	public final static int STAR_POINT = 4;

	/** Spectrum's point to draw.
	 * @see spectrum.SpectrumPoint */
	protected SEDPoint point = null;

	/** Frame of reference to use to draw the point.
	 * @see AGraphFrame */
	protected AGraphFrame monRepere = null;

	/** Serie which manages this point. */
	protected final Serie serie;

	/** Abscissa coordinate of the graphical representation of myPoint. */
	protected double xCoord = Double.NaN;

	/** Ordinate coordinate of the graphical representation of myPoint. */
	protected double yCoord = Double.NaN;

	public double side = 6.0;

	/** Graphical representation of myPoint. */
	protected Shape pointArea = null;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Constructs the graphical representation of the spectrum's point sp, on the given frame of reference rep.
	 * @param sp	The spectrum's point to draw.
	 * @param ref	The frame of reference to use to draw sp.
	 * @param s		The serie which manages this point.
	 * @see spectrum.SpectrumPoint
	 * @see AGraphFrame
	 */
	public Point(SEDPoint sp, AGraphFrame ref, Serie s){
		point = sp;
		monRepere = ref;
		serie = s;
	}

	/**
	 * Constructs the graphical representation of the spectrum's point sp, with the given coordinates.
	 * @param x	Abscissa.
	 * @param y	Ordinate.
	 * @param s The serie which manages this point.
	 * @see spectrum.SpectrumPoint
	 * @see AGraphFrame
	 */
	public Point(double x, double y, Serie s){
		xCoord = x;
		yCoord = y;
		serie = s;
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Gets the side of the area which delimits the point.
	 * @return Side of this area.
	 */
	public final double getSide(){
		return side;
	}

	/**
	 * Lets set the side of the area which delimits the point.
	 * @param s	Side of this area.
	 */
	public final void setSide(double s){
		side = s;
	}

	/**
	 * Return the abscissa coordinate of this point.
	 * @return	Abscissa.
	 */
	public final double getX(){
		return xCoord;
	}

	/**
	 * Sets the abscissa coordinate of this point.
	 * @param xValue	The new abscissa.
	 */
	public final void setX(double xValue){
		xCoord = xValue;
	}

	/**
	 * Return the ordinate coordinate of this point.
	 * @return	Ordinate.
	 */
	public final double getY(){
		return yCoord;
	}

	/**
	 * Sets the ordinate coordinate of this point.
	 * @param yValue	The new ordinate.
	 */
	public final void setY(double yValue){
		yCoord = yValue;
	}

	/**
	 * Gets the spectrum point which is graphically represented.
	 * @return	The source spectrum point.
	 */
	public final SEDPoint getSpectrumPoint(){
		return point;
	}

	/**
	 * Tells wether the graphical representation of the point contains a point with given coordinates (x,y).
	 * @param x	Abscissa of the tested point.
	 * @param y	Ordinate of the tested point.
	 * @return	true if the point with coordinates (x,y) is within the graphical representation of myPoint, false else.
	 * @see java.awt.Shape#contains(double, double)
	 */
	public final boolean contains(double x, double y){
		if (pointArea == null)
			return false;
		else
			return pointArea.contains(x,y);
	}

	/* ***************** */
	/* PREDEFINED POINTS */
	/* ***************** */
	/**
	 * Create a Point object with the given predefined style.
	 * @param sp	Source spectrum point.
	 * @param fr	Frame of reference to use.
	 * @param style	Point style (DEFAULT_POINT, SQUARE_POINT, CIRCLE_POINT, STAR_POINT)
	 * @return		The created Point object.<BR />
	 * <i><b>Note:</b> By default a point is created with DEFAULT_POINT style if 'style' is unknown.</i>
	 */
	public static final Point createPredefinedPoint(SEDPoint sp, AGraphFrame fr, Serie serie, int style){
		switch(style){
		case SQUARE_POINT:
			return new SquarePoint(sp, fr, serie);
		case CIRCLE_POINT:
			return new CirclePoint(sp, fr, serie);
		case CROSS_POINT:
			return new CrossPoint(sp, fr, serie);
		case STAR_POINT:
			return new StarPoint(sp, fr, serie);
		case DEFAULT_POINT:
		default:
			return new DefaultPoint(sp, fr, serie);
		}
	}

	/**
	 * Create a Point object with the given predefined style.
	 * @param x		Abscissa coordinate.
	 * @param y		Ordinate coordinate.
	 * @param style	Point style (DEFAULT_POINT, SQUARE_POINT, CIRCLE_POINT, STAR_POINT)
	 * @return		The created Point object.<BR />
	 * <i><b>Note:</b> By default a point is created with DEFAULT_POINT style if 'style' is unknown.</i>
	 */
	public static final Point createPredefinedPoint(double x, double y, Serie serie, int style){
		switch(style){
		case SQUARE_POINT:
			return new SquarePoint(x, y, serie);
		case CIRCLE_POINT:
			return new CirclePoint(x, y, serie);
		case CROSS_POINT:
			return new CrossPoint(x, y, serie);
		case STAR_POINT:
			return new StarPoint(x, y, serie);
		case DEFAULT_POINT:
		default:
			return new DefaultPoint(x, y, serie);
		}
	}

	/* ************ */
	/* DRAW METHODS */
	/* ************ */
	/**
	 * Draws the point and its label, if selected.
	 * @param g	The graphics to use to draw the point.
	 */
	public final void draw(Graphics g){
		if (point != null){
			// Get its attributes (wavelength and flux):
			double wl = point.getX(), f = point.getY();

			// Compute its coordinates in the panel with these attributes:
			double[] coord = monRepere.getCoordinates(wl, f);
			xCoord = coord[0]; yCoord = coord[1];
		}

		// Draw the graphical representation of this point with coordinates (xCoord,yCoord):
		Graphics2D g2 = (Graphics2D)g;
		drawPoint(g2);
	}

	/**
	 * Draws the label of this point (its coordinates in the frame of reference).
	 * @param g2	The Graphics object to use.
	 */
	public void drawValues(Graphics2D g2, boolean selected){
		Color col = g2.getColor();
		if (selected)
			g2.setColor(Color.RED);
		else if (serie != null)
			g2.setColor(serie.getColor());

		String pointStr = getStringValue();
		int stringWidth = g2.getFontMetrics().stringWidth(pointStr);
		if (monRepere.getWidth()-xCoord <= stringWidth)
			g2.drawString(pointStr, (int)(xCoord-stringWidth-2), (int)(yCoord-2));
		else
			g2.drawString(pointStr, (int)(xCoord+2), (int)(yCoord-2));

		if (selected) g2.setColor(col);
	}

	protected final String getStringValue(){
		if (point == null)
			return xCoord+";"+yCoord;
		else
			return FrameOfReference.getXValueString(point.getX())+point.getSed().getXAxis().getUnit().unicodeSymbol+" ; "+FrameOfReference.getYValueString(point.getY())+point.getSed().getYAxis().getUnit().unicodeSymbol;
	}

	/* **************** */
	/* ABSTRACT METHODS */
	/* **************** */
	/**
	 * Draws a graphical representation of the spectrum's point.
	 * @param g2	The Graphics object to use.
	 */
	protected abstract void drawPoint(Graphics2D g2);

	/**
	 * Draws something that highlights the point.
	 * 
	 * @param g2	The Graphics object to use.
	 */
	public abstract void drawSelection(Graphics2D g2);

	/**
	 * Gets the style of the point.
	 * @see Point#DEFAULT_POINT
	 * @see Point#SQUARE_POINT
	 * @see Point#CIRCLE_POINT
	 * @see Point#CROSS_POINT
	 * @see Point#STAR_POINT
	 * @return	The style of the drawn point.
	 */
	public abstract int getPointStyle();
}
