package sedplugin.graphe;

import java.awt.BasicStroke;
import java.awt.Color;
import java.awt.Graphics;
import java.awt.Graphics2D;
import java.awt.Stroke;
import java.awt.geom.Line2D;
import java.util.Iterator;

import sedplugin.sed.SED;
import sedplugin.sed.SEDPoint;
import sedplugin.sed.source.SEDSource;
import sedplugin.sed.source.aladin.PhotometrySource;

/**
 * A Serie object is a Point container. It manages their kind (cross, star, square, circle, ...) and
 * calls their draw method.
 *
 * @see Point
 *
 * @author Gregory Mantelet
 * @version 03/2012
 */
public class Serie {

	/** Default available colors. */
	public static final Color[] SERIE_COLORS = {
		new Color(0,0,0),
		Color.DARK_GRAY,
		Color.GRAY,
		Color.LIGHT_GRAY,
		new Color(255,255,255),

		new Color(255,255,0),
		new Color(204,153,0),
		new Color(255,136,0),
		new Color(255,128,128),
		Color.red,
		new Color(175,24,24),
		new Color(128,0,0),
		new Color(102,51,51),

		new Color(128,255,128),
		new Color(140,255,0),
		Color.green,
		new Color(153,204,0),
		new Color(0,128,0),
		new Color(17,132,15),
		new Color(22,86,21),

		new Color(0,255,255),
		new Color(0,140,255),
		new Color(62, 106, 162),
		Color.blue,
		new Color(0,0,128),
		new Color(15,22,80),
		new Color(128,128,255),
		new Color(255,0,100),
		new Color(255,0,255),
		new Color(140,0,255)
	};

	protected final static int[] autoColors = new int[]{9, 15, 23, 10, 16, 24, 11, 17, 25, 12, 18, 26, 13, 19, 27, 14, 20, 28};
	protected static int nextAutoColor = 0;

	/** The spectrum from which comes the above list of points.
	 * @see spectrum.Spectrum */
	final protected SED sed;

	/** The graph which manages/draws this serie. */
	final protected Graph graph;

	/** The reference graph of frame to use to build points of this serie. */
	final protected AGraphFrame reference;

	/** List of managed points.
	 * @see Point */
	protected Point[] myPoints;

	/** Indicates wether points must be connected or not. */
	protected boolean pointsConnected;

	/** Color of the serie. If null, the color is choosen automatically by the plugin. This attribute is never used if the SED is extracted from an Aladin drawing (circle, polygon). */
	protected Color color = null;

	protected int autoColorIndex = -1;

	/** Points style of this serie. */
	protected int pointStyle;

	protected int selectedPoint = -1;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Create a serie from a spectrum.<BR />
	 * <i>Note: the color of the serie will be randomly choosen and won't be white.</i>
	 *
	 * @param sp	Source spectrum.
	 */
	public Serie(SED sp){
		this(sp, null);
	}

	/**
	 * Create a serie from a spectrum.<BR />
	 * <i>Note: the color of the serie will be randomly chosen and will be different of the graph background color.</i>
	 *
	 * @param sp	Source spectrum.
	 * @param graph	The graph which will manipulate this serie.
	 */
	public Serie(SED sp, Graph graph){
		sed = sp;
		this.graph = graph;
		reference = (graph!=null)?graph.getFrameOfReference():null;

		build();

		pointsConnected = true;
		pointStyle = Point.SQUARE_POINT;

		setColor(null);
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Tells whether points of this serie must be graphically connected.
	 * @return	<i>true</i> points are connected, <i>false</i> else.
	 */
	public final boolean arePointsConnected(){
		return pointsConnected;
	}

	/**
	 * Sets wether points must be connected or not.
	 * @param connected	<i>true</i> points are connected, <i>false</i> else.
	 */
	public final void setPointsConnected(boolean connected){
		pointsConnected = connected;
	}

	/**
	 * Gets the spectrum associated with this serie.
	 * @return	The associated spectrum.
	 */
	public final SED getSED(){
		return sed;
	}

	/**
	 * Tells whether the ordinate unit is known or not.
	 * 
	 * @return	<i>true</i> if the Y unit of the represented SED is known, <i>false</i> if it is custom.
	 */
	public final boolean hasKnownYUnit(){
		return !sed.getYAxis().isCustomUnit();
	}

	/**
	 * Gets the graph which manages/draws this serie.
	 * 
	 * @return	Its graph.
	 */
	public final Graph getGraph(){
		return graph;
	}

	/**
	 * Gets the number of points of this serie.
	 * @return	The number of drawn points.
	 */
	public final int getNbPoints(){
		return myPoints.length;
	}

	/**
	 * Gets the chosen color for this serie.
	 * @return	The used color.
	 */
	public final Color getColor(){
		if (color == null){
			SEDSource sedSrc = sed.getSource();
			if (sedSrc instanceof PhotometrySource)
				return ((PhotometrySource)sedSrc).getColor();
			else
				return color;
		}else
			return color;
	}

	/**
	 * Sets the color with an indice to a color of the SERIE_COLOR constant.
	 * @param c	Indice of a SERIE_COLOR color to use.
	 */
	public final void setColor(int c){
		if (c >= 0 && c < SERIE_COLORS.length)
			setColor(SERIE_COLORS[c]);

	}

	/**
	 * Sets directly a color.
	 * @param c	The color to use.
	 */
	public final void setColor(Color c){
		Color currentColor = getColor();
		if ((c == null && autoColorIndex != -1) || (c != null && currentColor != null && c.getRGB() == currentColor.getRGB()))
			return;

		if (c == null)
			c = chooseColor(reference.getColor());
		else
			releaseAutoColor();

		if (color == null){
			SEDSource sedSrc = sed.getSource();
			if (sedSrc instanceof PhotometrySource)
				((PhotometrySource)sedSrc).setColor(c);
			else
				color = c;
		}else
			color = c;
	}

	/**
	 * Get the graphical representation kind of points.
	 * @return	The used point style (DEFAULT_POINT, SQUARE_POINT, CIRCLE_POINT, CROSS_POINT, STAR_POINT).
	 * @see Point#DEFAULT_POINT
	 * @see Point#SQUARE_POINT
	 * @see Point#CIRCLE_POINT
	 * @see Point#CROSS_POINT
	 * @see Point#STAR_POINT
	 */
	public final int getPointStyle(){
		return pointStyle;
	}

	/**
	 * Lets changing point style of this serie:
	 * @param style	The point style to use (DEFAULT_POINT, SQUARE_POINT, CIRCLE_POINT, CROSS_POINT, STAR_POINT).
	 * @see Point#DEFAULT_POINT
	 * @see Point#SQUARE_POINT
	 * @see Point#CIRCLE_POINT
	 * @see Point#CROSS_POINT
	 * @see Point#STAR_POINT
	 * <i><b>Note:</b> Nothing will be done if point style is unknown. To manage other point style, you must extend Serie and redefine {@link Serie#buildPoint(SpectrumPoint)} and {@link Serie#setPointStyle(int)} functions.</i>
	 */
	public void setPointStyle(int style){
		if (style == Point.DEFAULT_POINT || style == Point.SQUARE_POINT || style == Point.CIRCLE_POINT || style == Point.CROSS_POINT || style == Point.STAR_POINT){
			pointStyle = style;
			int tmp = selectedPoint;
			build();
			selectedPoint = tmp;
		}
	}

	/**
	 * Gets the ind-th point of this serie.
	 * @param ind	The indice of the point to return.
	 * @return		The ind-th point.
	 */
	public final Point getPoint(int ind){
		if (ind>=0 && ind<myPoints.length)
			return myPoints[ind];
		else
			return null;
	}

	/**
	 * Gets the selected point.
	 * 
	 * @return	The selected point or <i>null</i>.
	 */
	public final Point getSelectedPoint(){
		if (selectedPoint < 0 || selectedPoint >= myPoints.length){
			selectedPoint = -1;
			return null;
		}else
			return myPoints[selectedPoint];
	}

	/**
	 * Gets the index of the selected point.
	 * 
	 * @return	The index of the selected point, or -1.
	 */
	public final int getSelectedPointIndex(){
		return selectedPoint;
	}

	/**
	 * Updates the index of the selected point.
	 * 
	 * @param index	The index of the selected point.
	 */
	public final void setSelectedPoint(int index){
		selectedPoint = index;
	}

	/**
	 * Selects the first point found at the given position (if any).
	 * 
	 * @param x	Abscissa.
	 * @param y	Ordinate.
	 * @return <i>true</i> if a point has been found at the given position.
	 * @see Point#contains(double, double)
	 */
	public final boolean selectPointAt(int x, int y){
		for(int i=0; i<myPoints.length; i++){
			if (myPoints[i].contains(x, y)){
				selectedPoint = (selectedPoint==i) ? -1 : i;
				return true;
			}
		}
		selectedPoint = -1;
		return false;
	}

	/**
	 * Scan the list of points so that return (if existing) the first point which graphically represents the given position.
	 * @param x Abscissa.
	 * @param y	Ordinate.
	 * @return	The graphical representation of the point which represents the given position.
	 * @see Point#contains(double, double)
	 */
	public final Point getPointAt(int x, int y){
		boolean found = false; int i;

		for(i=0; i<myPoints.length && !found; i++)
			found = myPoints[i].contains(x, y);

		return (found)?myPoints[i-1]:null;
	}

	/* ****************** */
	/* GENERATION METHODS */
	/* ****************** */
	/**
	 * Build the points array from spectrum points of the setted spectrum.
	 */
	public final void build(){
		myPoints = new Point[sed.size()];

		Iterator<SEDPoint> it = sed.iterator();
		for(int i=0; it.hasNext(); i++){
			SEDPoint spt = it.next();
			myPoints[i] = buildPoint(spt);
		}

		selectedPoint = -1;
	}

	/**
	 * Build a point from a SpectrumPoint thanks to the choosen pointStyle (DEFAULT_POINT, SQUARE_POINT, CIRCLE_POINT, CROSS_POINT, STAR_POINT).
	 * @param spt	The source SpectrumPoint.
	 * @return 		The created Point object.
	 * @see Point#DEFAULT_POINT
	 * @see Point#SQUARE_POINT
	 * @see Point#CIRCLE_POINT
	 * @see Point#CROSS_POINT
	 * @see Point#STAR_POINT
	 * <i><b>Note:</b> A default point will be created if point style is unknown. To manage other point style, you must extend Serie and redefine {@link Serie#buildPoint(SpectrumPoint)} and {@link Serie#setPointStyle(int)} functions.</i>
	 */
	protected Point buildPoint(SEDPoint spt){
		Point newPoint = null;
		if (reference == null)
			newPoint = Point.createPredefinedPoint(spt.getX(), spt.getY(), this, pointStyle);
		else
			newPoint = Point.createPredefinedPoint(spt, reference, this, pointStyle);
		return newPoint;
	}

	/**
	 * Randomly chooses a color in the {@link Serie#SERIE_COLORS} array avoiding the given one.
	 * @param colorToAvoid	The color which mustn't be choosen.	<i>(MUST BE NON-NULL)</i>
	 * @return				The randomly choose color.
	 */
	protected final Color chooseColor(Color colorToAvoid){
		/*int indColor;
		do{
			indColor = (int)(Math.random()*SERIE_COLORS.length);
		}while(SERIE_COLORS[indColor].getRGB() == colorToAvoid.getRGB());
		return SERIE_COLORS[indColor];*/
		if (nextAutoColor < 0 || nextAutoColor >= autoColors.length)
			nextAutoColor = 0;
		autoColorIndex = nextAutoColor;
		return SERIE_COLORS[autoColors[nextAutoColor++]];
	}

	public void releaseAutoColor(){
		if (autoColorIndex >= 0 && autoColorIndex < autoColors.length){
			nextAutoColor = autoColorIndex;
			autoColorIndex = -1;
		}
	}

	/* ************* */
	/* PRINT METHODS */
	/* ************* */
	@Override
	public String toString(){
		return sed.getName();
	}

	/**
	 * Lets draw all the managed points.
	 * @param g The Graphics object to use to draw points.
	 */
	public void draw(Graphics g){
		// set some graphic options:
		Graphics2D g2 = (Graphics2D)g;
		Color oldColor = g2.getColor();
		g2.setColor(getColor());
		Stroke oldStroke = g2.getStroke(), lineStroke = new BasicStroke(1, BasicStroke.CAP_ROUND, BasicStroke.JOIN_ROUND);

		double[] coord1=null, coord2=null;
		for(int i=0; i<myPoints.length; i++){
			// draw current point:
			myPoints[i].draw(g2);
			// draw connection with the previous point, if asked:
			if (arePointsConnected()){
				coord1 = coord2;
				coord2 = new double[]{myPoints[i].getX(), myPoints[i].getY()};
				if (coord1 != null){
					g2.setStroke(lineStroke);
					g2.draw(new Line2D.Double(coord1[0], coord1[1], coord2[0], coord2[1]));
					g2.setStroke(oldStroke);
				}
			}
		}
		g2.setColor(oldColor);
	}
}
