package sedplugin.plugin;

import java.io.IOException;

import java.net.MalformedURLException;
import java.net.URL;

import java.util.Iterator;

import sedplugin.characterization.Characterization;
import sedplugin.characterization.HeaderFits;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;

import sedplugin.sed.source.aladin.Drawing;

import cds.aladin.AladinData;
import cds.aladin.AladinException;

/**
 * Association with an Aladin image plan.
 *
 * @author Gregory Mantelet
 * @version 03/2012
 */
public class AladinImage {

	protected final AladinProxy proxy;

	/** The associated Aladin plan. */
	protected AladinData plane = null;

	/** The label of this plan. */
	protected String label = null;

	/** The URL of the image. */
	protected URL url;

	/** The FITS header of this image. */
	protected HeaderFits headerFits;

	/** The Characterization associated with this image. */
	protected Characterization characterization;

	/** The type of the image (jpeg, fits, ...). */
	protected String fileType = null;

	/** Tells whether this image is AllSky. */
	protected final boolean isAllSky;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Constructs an ImageOrigin object.
	 *
	 * @param ad		The AladinData containing the image from which the following information have been extracted.
	 * @param aladin	Object which manages the interaction with Aladin
	 */
	public AladinImage(AladinData ad, AladinProxy aladin){
		proxy = aladin;
		plane = ad;
		label = (ad != null)?ad.getLabel():"";
		isAllSky = (ad != null && ad.getPlaneType() != null) ? ad.getPlaneType().equals(AladinProxy.IMAGE_ALL_SKY_PLAN) : null;

		if (isAllSky && ad.getOrigin() != null){
			try{
				url = new URL(ad.getOrigin());
			}catch(MalformedURLException mue){
				url = ad.getUrl();
			}
		}else
			url = ad.getUrl();

		loadHeaderFits();
		characterization = null;
	}

	protected void loadHeaderFits(){
		if (plane != null){
			headerFits = null;
			try {
				headerFits = new HeaderFits(plane.getFitsHeader());
			} catch (IOException e) {
				ErrorsLog.getCurrentInstance().addRefreshException(new SEDException(e, "While getting the FITS header of \""+label+"\"."));
			} catch (AladinException e) {
				ErrorsLog.getCurrentInstance().addRefreshException(new SEDException(e, "While getting the FITS header of \""+label+"\"."));
			}
		}
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Gets the Aladin plan.
	 * @return	An AladinData object if still existing in the Aladin stack , <i>null</i> else.
	 */
	public final AladinData getPlane(){
		return plane;
	}

	/**
	 * Tells whether the corresponding Aladin image is AllSky or not.
	 * 
	 * @return	<i>true</i> if the image is AllSky, <i>false</i> otherwise.
	 */
	public final boolean isAllSky(){
		return isAllSky;
	}

	/**
	 * Updates the name of the Aladin image.
	 */
	public final void updateName(){
		if (plane != null)
			label = plane.getLabel();
	}

	/**
	 * Remove the Aladin image association. After calling this method this Origin object will considere
	 * that the Aladin plan doesn't exist any more. (= setImage(null, ..))
	 */
	public final void noMoreImage(){
		if (plane != null)
			label = plane.getLabel();
		plane = null;
		url = null;
	}

	/**
	 * Lets setting the image.
	 * @param img		The associated AladinData.
	 * @param indPlan 	The indice of img in the Aladin stack (<b><u>WARNING</u>: if indPlan < 0 then {@link AladinImage#noMoreImage()} method will be called and the association with img won't be done</b>).
	 */
	public final void setImage(AladinData img, int indPlan){
		if (img == null || indPlan < 0)
			noMoreImage();
		else{
			plane = img;
			label = img.getLabel();
			url = plane.getUrl();
			loadHeaderFits();
			characterization = null;
		}
	}

	/**
	 * The name of the Aladin plan.
	 * @return	Name of the given AladinData.
	 */
	public final String getName(){
		return label;
	}

	/**
	 * Get the URL of the image.
	 * @return	The URL of the image.
	 */
	public final URL getImageUrl(){
		return url;
	}

	/**
	 * Gets the FITS header of this image.
	 * @return	Its header FITS.
	 */
	public final HeaderFits getHeaderFits(){
		return headerFits;
	}

	/**
	 * Indicates whether this image is associated with a Characterization.
	 * @return	<i>true</i> if there is a Characterization, <i>false</i> otherwise.
	 */
	public final boolean hasCharacterization(){
		return characterization != null;
	}

	/**
	 * Get the Characterization of this image.
	 * @return	Its Characterization.
	 */
	public final Characterization getCharacterization(){
		return characterization;
	}

	/**
	 * Set the Characterization of this image.
	 * @param charac		Its new Characterization.
	 */
	public final void setCharacterization(Characterization charac){
		characterization = charac;
		characterizationUpdated();
	}

	/**
	 * Updates all SED source which depends on this image.
	 * 
	 * @see AladinProxy#getDrawings(cds.aladin.AladinData)
	 * @see SEDSource#update(boolean);
	 */
	public final void characterizationUpdated(){
		Iterator<Drawing> it = proxy.getDrawings(plane).iterator();
		while(it.hasNext())
			it.next().update(true);
	}

	/**
	 * Gets the image type (jpeg, fits, ...).
	 * @return	The image type.
	 */
	public final String getImageType(){
		return fileType;
	}

	/**
	 * Sets the image type.
	 * @param type	Type of the associated image (jpeg, fits, ...).
	 */
	public final void setImageType(String type){
		fileType = type;
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	public String toAladinCommand(boolean drawIfRequired){
		/*String command = "setconf Sync=off ; ";

		if (originDataExists())
		command += "cview \""+getName()+"\"";
		else if (drawIfRequired && url != null)
			command += getName()+" = load "+getImageUrl();

		return command+" ; setconf Sync=on";*/

		return "";
	}

	/* ************** */
	/* FORMAT METHODS */
	/* ************** */
	@Override
	public String toString(){
		return label/*+"["+getIndice()+"]"*/;
	}
}
