package sedplugin.plugin;


import java.beans.PropertyChangeListener;

import java.io.InputStream;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;

import javax.swing.SwingWorker.StateValue;

import sedplugin.characterization.CharacManager;
import sedplugin.errors.CharacterizationException;
import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.interfaceGraphique.SEDFrame.SynchronizePropertyChangeListener;
import sedplugin.sed.source.aladin.Drawing;

import cds.aladin.Aladin;
import cds.aladin.AladinData;
import cds.aladin.AladinException;

import cds.tools.VOApp;
import cds.tools.VOObserver;



/**
 * <p>Objects which manages all interactions between the SEDPlugin and Aladin.</p>
 * <p>It particularly keeps updated a simplified view of the Aladin stack.</p>
 * <p>
 * 	It also loads the Characterization of all images loaded in Aladin.
 * 	If an image has no Characterization or if it can not be fetched, this image will not be usable in the SEDPlugin.
 * </p>
 * <p>Drawing planes are also browsed so that identifying SED possible sources (for the moment only Circle and Polygon).</p>
 * 
 * <p>
 * 	If at any moment, the user or the plugin itself do not want any more interacting with Aladin,
 * 	the method {@link #setIsInteracting(boolean)} must be called with <i>false</i> as parameter.
 * 	In this case, all Aladin event will be ignored. Obviously it will have the same effect
 * 	for actions coming from the plugin to Aladin.
 * </p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2011
 */
public class AladinProxy implements VOApp, VOObserver {

	/** Plane type of an image AladinData. To check if an AladinData is an image: <pre>ad.getPlaneType().equals(IMAGE_PLAN) <i>(where </i>ad<i> is an AladinData instance)</i></pre> */
	public static final String IMAGE_PLAN = "Image";
	public static final String IMAGE_ALGO_PLAN = "Image/Algo";
	public static final String IMAGE_RESAMP_PLAN = "Image/Resampled";
	public static final String IMAGE_ALL_SKY_PLAN = "Image/AllskyImage";

	/** Plane type of a measures AladinData. */
	public static final String DRAWING_PLAN = "Overlay/Tool";

	/** Plane type of a catalog AladinData. */
	public static final String CATALOG_PLAN = "Overlay/Catalog";

	/** Plane type of a measures AladinData. To check if an AladinData is an image: <pre>ad.getPlaneType().equals(MEASURES_PLAN) <i>(where </i>ad<i> is an AladinData instance)</i></pre> */
	public static final String MEASURES_PLAN = "Overlay/Tool";

	/** Obj type of a circle (drawn by the user) intern object of a measures AladinData. To check if an Obj is a circle: <pre>obj.getObjType().equals(CIRCLE_OBJECT) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String CIRCLE_OBJECT = "Tag";

	/** Obj type of a photometry object (drawn by the user) intern object of a measures AladinData. To check if an Obj is a photometry: <pre>obj.getObjType().equals(PHOT_OBJECT) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String PHOT_OBJECT = "Phot";

	/** Obj type of a circle (drawn by a command) intern object of a measures AladinData. To check if an Obj is a circle: <pre>obj.getObjType().equals(CIRCLE_OBJECTBis) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String CIRCLE_OBJECTBis = "Circle";

	/** Obj type of a polygon point (begin or middle) intern object of a measures AladinData. To check if an Obj is a polygon point: <pre>obj.getObjType().equals(POLYGON_OBJECT) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String POLYGON_OBJECT = "Line+";

	/** Obj type of a polygon point (end only) intern object of a measures AladinData. To check if an Obj is a polygon point: <pre>obj.getObjType().equals(POLYGON_END) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String POLYGON_END = "Line";

	/**
	 * Possible types of an Aladin plane.
	 * 
	 * @author Gr&eacute;gory Mantelet (CDS)
	 * @version 03/2012
	 */
	public static enum PlaneType {
		DRAWING, IMAGE, CATALOG;

		/**
		 * Tells whether the given Aladin plane type corresponds to the one represented by this object.
		 * 
		 * @param planeType	An Aladin plane type.
		 * 
		 * @return	<i>true</i> if this type is equal to the given one, <i>false</i> otherwise.
		 */
		public boolean hasSameType(final String planeType){
			if (planeType == null)
				return false;

			switch(this){
			case DRAWING:
				return planeType.equalsIgnoreCase(DRAWING_PLAN);
			case CATALOG:
				return planeType.equalsIgnoreCase(CATALOG_PLAN);
			case IMAGE:
				return planeType.equals(IMAGE_PLAN) || planeType.equals(IMAGE_ALGO_PLAN) || planeType.equals(IMAGE_RESAMP_PLAN) || planeType.equals(IMAGE_ALL_SKY_PLAN);
			default:
				return false;
			}
		}
	}

	protected final SEDPlugin plugin;

	protected final Aladin aladin;

	protected final HashMap<Integer, Drawing> drawings;

	protected final HashMap<Integer, AladinImage> images;

	protected final ArrayList<Integer> allPlanes;

	protected boolean interacting = true;

	protected MySwingWorker loader = null;

	protected PropertyChangeListener loaderListener = null;

	/**
	 * Builds an {@link AladinProxy}.
	 * By default, no synchronization ({@link #synchronize()}) is done at creation.
	 * 
	 * @param aladin	The Aladin instance with which this proxy will synchronize.
	 * @param plugin	The plugin which requires this proxy.
	 */
	public AladinProxy(Aladin aladin, SEDPlugin plugin){
		this.plugin = plugin;
		this.aladin = aladin;
		drawings = new HashMap<Integer, Drawing>(2);
		images = new HashMap<Integer, AladinImage>(10);
		allPlanes = new ArrayList<Integer>(10);
	}

	/* ************************** */
	/* SYNCHRONIZATION MANAGEMENT */
	/* ************************** */
	/**
	 * Tells whether this proxy must interact with Aladin.
	 * @return	<i>true</i> if this proxy is allowed to interact with Aladin, <i>false</i> otherwise.
	 */
	public final boolean isInteracting(){
		return interacting;
	}

	/**
	 * Asks to this proxy to interact with Aladin.
	 * @param interact	<i>true</i> if this proxy must interact with Aladin, <i>false</i> otherwise.
	 */
	public final void setIsInteracting(final boolean interact){
		interacting = interact;
	}

	/**
	 * Updates its view of the Aladin stack.
	 * @see SynchronizeWorker
	 */
	public synchronized void synchronize(){
		if (loader != null){
			if (loader.isDone() || loader.cancel(true))
				loader = null;
			else
				return;
		}

		loader = new SynchronizeWorker();
		if (loaderListener != null){
			loader.addPropertyChangeListener(loaderListener);
			((SynchronizePropertyChangeListener)loaderListener).setTask(loader);
		}
		loader.execute();
	}

	/**
	 * Stops the current synchronization, if any.
	 */
	public void stopSynchronizing(){
		if (isSynchronizing())
			loader.cancel(true);
	}

	/**
	 * Tells whether this proxy is currently synchronizing with Aladin.
	 * @return	<i>true</i> if it is synchronizing with Aladin, <i>false</i> otherwise.
	 */
	public boolean isSynchronizing(){
		return (loader != null && loader.getState() == StateValue.STARTED);
	}

	/**
	 * Sets the object which wants to watch all synchronization progressions.
	 * 
	 * @param listener	A property change listener.
	 */
	public void setLoaderListener(final PropertyChangeListener listener){
		if (loader != null && loaderListener != null){
			loader.removePropertyChangeListener(loaderListener);
			loaderListener = null;
		}

		if (loader != null && listener != null)
			loader.addPropertyChangeListener(listener);

		loaderListener = listener;
	}

	/* **************** */
	/* PLANE MANAGEMENT */
	/* **************** */
	/**
	 * Gets the Aladin image corresponding to the given plane.
	 * 
	 * @param plane	The plane corresponding to the {@link AladinImage} to get/
	 * @return		The corresponding Aladin image.
	 */
	public AladinImage getImage(AladinData plane){
		return images.get(new Integer(plane.getPlaneHashCode()));
	}

	/**
	 * Gets all images under the given plane in the Aladin stack.
	 * 
	 * @param limit		Limit of the iteration. When this plane is encountered, the iterator stops. (MAY BE null)
	 * 
	 * @return			The list of images.
	 */
	public ArrayList<AladinImage> getImages(final AladinData limit){
		ArrayList<AladinImage> lstImages = new ArrayList<AladinImage>();
		final Integer hashCodeLimit = (limit == null) ? null : new Integer(limit.getPlaneHashCode());
		for(Integer i : allPlanes){
			if (hashCodeLimit != null && hashCodeLimit.equals(i))
				return lstImages;
			if (images.get(i) != null)
				lstImages.add(images.get(i));
		}
		return lstImages;
	}

	/**
	 * Counts the number of images under the given Aladin plane.
	 * 
	 * @param limit		Limit of the iteration. When this plane is encountered, the iterator stops. (MAY BE null)
	 * 
	 * @return			The number of images under the given plane.
	 */
	public int countImages(final AladinData limit){
		int nbImages = 0;
		final Integer hashCodeLimit = (limit == null) ? null : new Integer(limit.getPlaneHashCode());
		for(Integer i : allPlanes){
			if (hashCodeLimit != null && hashCodeLimit.equals(i))
				return nbImages;
			if (images.get(i) != null)
				nbImages++;
		}
		return nbImages;
	}

	/**
	 * Gets all drawings under the given plane in the Aladin stack.
	 * 
	 * @param limit		Limit of the iteration. When this plane is encountered, the iterator stops. (MAY BE null)
	 * 
	 * @return			The list of drawings.
	 */
	public ArrayList<Drawing> getDrawings(final AladinData limit){
		ArrayList<Drawing> lstDrawings = new ArrayList<Drawing>();
		final Integer hashCodeLimit = (limit == null) ? null : new Integer(limit.getPlaneHashCode());
		boolean found = (hashCodeLimit == null);
		for(Integer i : allPlanes){
			if (!found)
				found = hashCodeLimit.equals(i);
			else if (drawings.containsKey(i))
				lstDrawings.add(drawings.get(i));
		}
		return lstDrawings;
	}

	/* ************************** */
	/* ACTIONS TO APPLY IN ALADIN */
	/* ************************** */
	/**
	 * Centers the Aladin reticle to the given coordinates (in RADEC).
	 * 
	 * @param ra	Right ascension.
	 * @param dec	Declination.
	 */
	public void centerPosition(double ra, double dec){
		if (interacting)
			aladin.execCommand("draw mode(RADEC) ; "+ra+" "+dec);
	}

	/**
	 * Highlights the given image in Aladin.
	 * That's to say the image is selected/shown.
	 * 
	 * @param io	The Aladin image to highlight.
	 */
	public void highlightImage(AladinImage io) {
		if (interacting)
			aladin.execAsyncCommand("setconf Sync=off ; cview \""+io.getName()+"\" ; setconf Sync=on");
	}

	/* ************************** */
	/* ACTIONS COMING FROM ALADIN */
	/* ************************** */
	public String execCommand(String cmd) {
		if (!interacting || UserPreferences.getCurrentInstance().whenAllRefreshing() != UserPreferences.ON_ANY_ALADIN_MODIFICATION)
			return "";

		if (cmd.equals("info stackEvent")){
			if (UserPreferences.getCurrentInstance().whenAllRefreshing() == UserPreferences.ON_ANY_ALADIN_MODIFICATION)
				synchronize();
		}else if (cmd.startsWith("set "))
			synchronize();

		return "";
	}

	public void pixel(double arg0) { ; }
	public void position(double ra, double dec) {
		/*SEDSource src = new Position(ra, dec, this, plugin);
		SED sed = src.getSED();
		System.out.println("DEBUG: (POSITION "+ra+" ; "+dec+") = SED with "+sed.size());
		for(SEDPoint point : sed)
			System.out.println("   + "+point);
		src.removeSED();*/
	}
	public void addObserver(VOObserver arg0, int arg1) { ; }
	public InputStream getFITS(String arg0) { return null; }
	public InputStream getVOTable(String arg0) { return null; }
	public String putFITS(InputStream arg0, String arg1) { return null; }
	public String putVOTable(InputStream arg0, String arg1) { return null; }
	public String putVOTable(VOApp arg0, InputStream arg1, String arg2) { return null; }
	public void selectVOTableObject(String[] arg0) { ; }
	public void setVisible(boolean arg0) { ; }
	public void showVOTableObject(String[] arg0) { ; }


	/**
	 * Task which synchronizes its view of the Aladin stack with Aladin.
	 * 
	 * @author Gr&eacute;gory Mantelet (CDS)
	 * @version 04/2012
	 */
	protected class SynchronizeWorker extends MySwingWorker {
		@Override
		protected Void doInBackground() throws Exception {
			String[] stack = aladin.getAladinStack();
			String progressMessage = "Synchronizing...";

			ArrayList<Integer> planesToRemove = new ArrayList<Integer>(allPlanes);
			allPlanes.clear();

			ErrorsLog.getCurrentInstance().clearLastExceptions();

			String oldPrgMsg = "";
			setProgress(0);
			firePropertyChange("ProgressMessage", oldPrgMsg, progressMessage);

			for(int i=0; i<stack.length; i++){
				try {
					if (isCancelled()) break;
					AladinData ad = aladin.getAladinData(stack[i]);
					if (ad.isReady()){
						if (isCancelled()) break;
						Integer hashCode = new Integer(ad.getPlaneHashCode());
						allPlanes.add(hashCode);
						planesToRemove.remove(hashCode);

						String planeType = ad.getPlaneType();

						oldPrgMsg = progressMessage; progressMessage = stack[i];
						firePropertyChange("ProgressMessage", oldPrgMsg, progressMessage);

						// DRAWING PLAN:
						if (planeType.equals(DRAWING_PLAN)){
							if (isCancelled()) break;
							if (!drawings.containsKey(hashCode)){
								Drawing newDrawing = new Drawing(ad, AladinProxy.this, plugin);
								drawings.put(hashCode, newDrawing);
								plugin.addSEDSource(newDrawing);
							}

						}// IMAGE PLAN:
						else if (planeType.equals(IMAGE_PLAN) || planeType.equals(IMAGE_ALGO_PLAN) || planeType.equals(IMAGE_RESAMP_PLAN) || planeType.equals(IMAGE_ALL_SKY_PLAN)){
							if (isCancelled()) break;
							if (images.get(hashCode) == null){
								AladinImage ai = new AladinImage(ad, AladinProxy.this);
								try{
									ai.setCharacterization(CharacManager.getCharacterization(ai));
								}catch(CharacterizationException ce){
									ErrorsLog.getCurrentInstance().addRefreshException(ce);
								}
								images.put(hashCode, ai);
								plugin.addImage(ai);
							}

							/*}// CATALOG PLAN:
						else if (planeType.equals(CATALOG_PLAN)){
							System.out.println("catalog)");*/

						}// UNKNOWN PLAN:
						/*else
							System.out.println("??? = "+planeType+")");*/
					}
				} catch (AladinException e) {
					System.out.println("((X)) [SEDPlugin] Unable to load the "+(i+1)+"-th plane, named \""+stack[i]+"\" because: "+e.getMessage()+" !");
					ErrorsLog.getCurrentInstance().addRefreshException(new SEDException(e, "Unable to load the "+(i+1)+"-th plane (named \""+stack[i]+"\") of the Aladin stack !"));
				}
				setProgress((int)(((double)i/stack.length)*40));
			}

			// Remove objects (drawings/images) associated with the removed plans:
			oldPrgMsg = progressMessage; progressMessage = "Cleaning...";
			firePropertyChange("ProgressMessage", oldPrgMsg, progressMessage);
			int count = 0;
			for(Integer i : planesToRemove){
				Drawing d = drawings.remove(i);
				plugin.removeSEDSource(d);
				if (d != null)
					d.disconnectFromAladin();

				AladinImage img = images.remove(i);
				if (img != null)
					plugin.removeImage(img);
				setProgress((int)(40+(((double)++count)/planesToRemove.size())*30));
			}

			// Updates all remaining drawings (if not frozen):
			if (isCancelled()) return null;
			oldPrgMsg = progressMessage; progressMessage = "Updating drawings...";
			firePropertyChange("ProgressMessage", oldPrgMsg, progressMessage);
			Iterator<Drawing> it = drawings.values().iterator();
			count = 0;
			while(it.hasNext()){
				if (isCancelled()) return null;
				Drawing d = it.next();
				if (!d.isFrozen())
					d.update();
				setProgress((int)(70+(((double)++count)/drawings.size())*30));
			}

			if (!isCancelled()) setProgress(100);
			firePropertyChange("ProgressMessage", progressMessage, isCancelled()?"Cancelled !":"Synchronized !");

			loader = null;

			return null;
		}
	}

}
