package sedplugin.plugin;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;


/**
 * List all preferences available into the SEDPlugin.<BR />
 * Only the last instance of the UserPreferences class can be accessible thanks to the static method {@link UserPreferences#getCurrentInstance()}.<BR />
 * Options for each preference:
 * <ul>
 * 	<li><b>synchronizationProblem:</b> ASK, APPLY_ALL or IGNORE_ALL</li>
 * 	<li><b>errorOnImageLoading:</b> ASK, IGNORE_ALL or DELETE_IMAGE</li>
 * 	<li><b>refreshAllOn:</b> ON_ACTIVATE or ON_USER_DEMAND</li>
 * 	<li><b>deactivateInteractions:</b> NEVER or AFTER_EACH_ONE</li>
 * 	<li><b>printPointValuesOnOver:</b> true or false</li>
 * 	<li><b>askConfirmationBeforeSendToSamp:</b> true or false</li>
 * </ul>
 *
 * @author Gregory Mantelet
 * @version 04/2012
 */
public class UserPreferences {

	protected static UserPreferences currentInstance;

	public static final int ASK = 0;

	// REFRESH_ALL OPTIONS:
	/*public static final int APPLY_ALL = 1;
	public static final int IGNORE_ALL = 2;

	public static final String KEY_SYNCHRONIZATION_PROBLEM = "synchronization_behaviour";
	protected int synchronizationProblem = APPLY_ALL;	// ASK, APPLY_ALL or IGNORE_ALL*/

	// IMAGE LOADING FAIL:
	/*public static final int DELETE_IMAGE = 3;

	public static final String KEY_ERROR_ON_IMAGE_LOADING = "error_on_image_loading";
	protected int errorOnImageLoading = IGNORE_ALL;	// IGNORE_ALL or DELETE_IMAGE

	public static final String KEY_DISPLAY_LOADING_ERROR = "display_loading_errors";
	protected boolean displayImageLoadingError = false;*/

	// REFRESH_ALL ON...
	public static final int ON_ANY_ALADIN_MODIFICATION = 4;
	public static final int ON_ACTIVATE = 5;
	public static final int ON_USER_DEMAND = 6;

	public static final String KEY_REFRESH_ALL_ON_KEY = "refresh_all_on";
	protected int refreshAllOn = ON_ACTIVATE;	// ON_ACTIVATE or ON_EACH_ALADIN_MODIFICATION or ON_USER_DEMAND

	// WHEN DEACTIVATE INTERACTION:
	public static final int NEVER = 7;
	public static final int AFTER_EACH_ONE = 8;

	// INTERACTION DEACTIVATION:
	public static final String KEY_DEACTIVATE_INTERACTIONS_KEY = "deactivate_interactions";
	protected int deactivateInteractions = NEVER;	// NEVER or AFTER_EACH_ONE

	// DISPLAY ERRORS FOR GRAVITY:
	/*public static final String KEY_NOTIFY_USER_FOR = "notify_user_for";
	protected int notifyUserFor = 0; //SEDException.ERROR;*/

	// SAVE ERRORS LOG:
	public static final String KEY_SAVE_ERRORS_LOG = "save_errors";
	protected boolean saveErrorsLog = false;

	// CREATE A SED WHEN CLICKING ON AN IMAGE:
	/*public static final String KEY_CREATE_SED_ON_CLICK = "create_sed_on_click";
	protected boolean createSEDOnClick = false;*/

	// MOVE THE ALADIN RETICULE WHEN CLICKING ON A SPECTRUM:
	public static final String KEY_MOVE_ON_SPECTRUM_CLICK = "move_on_spectrum_click";
	protected boolean moveToOnSpectrumClick = true;

	// PRINT POINT VALUES:
	public static final String KEY_PRINT_VALUES = "point_values_visible";
	protected boolean printPointValuesOnOver = true;

	// SHOW A CONFIRM DIALOG BEFORE SEND A SPECTRUM VIA SAMP:
	public static final String KEY_ASK_CONFIRMATION_SEND_SAMP = "confirmation_before_send_samp";
	protected boolean askConfirmationBeforeSendToSamp = true;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	protected UserPreferences(){ ; }

	public static final void setCurrentInstance(File configFile){
		boolean loaded = false;
		if (configFile.exists() && configFile.isFile() && configFile.canRead()){
			int numLine = 1;
			try {
				currentInstance = new UserPreferences();
				BufferedReader reader = new BufferedReader(new FileReader(configFile));
				String line = null;
				while((line = reader.readLine()) != null){
					if (!line.startsWith("#") && line.trim().length()!=0){
						String[] keyValues = line.split("=");
						if (keyValues.length == 2){
							keyValues[0] = keyValues[0].trim();
							if (keyValues[1].indexOf("#") > -1)
								keyValues[1] = keyValues[1].substring(0, keyValues[1].indexOf("#"));
							keyValues[1] = keyValues[1].trim();
							if (keyValues[0].equalsIgnoreCase(KEY_DEACTIVATE_INTERACTIONS_KEY))
								currentInstance.deactivateInteractionsOn(keyValues[1]);
							/*else if (keyValues[0].equalsIgnoreCase(KEY_CREATE_SED_ON_CLICK))
								currentInstance.setCreateSEDOnClick(keyValues[1]);*/
							else if (keyValues[0].equalsIgnoreCase(KEY_MOVE_ON_SPECTRUM_CLICK))
								currentInstance.setMoveToOnSpectrumClick(keyValues[1]);
							else if (keyValues[0].equalsIgnoreCase(KEY_PRINT_VALUES))
								currentInstance.setPrintValuesOnOver(keyValues[1]);
							else if (keyValues[0].equalsIgnoreCase(KEY_ASK_CONFIRMATION_SEND_SAMP))
								currentInstance.setAskConfirmationBeforeSendToSamp(keyValues[1]);
							else if (keyValues[0].equalsIgnoreCase(KEY_REFRESH_ALL_ON_KEY))
								currentInstance.setRefreshAllOn(keyValues[1]);
							/*else if (keyValues[0].equalsIgnoreCase(KEY_SYNCHRONIZATION_PROBLEM))
								currentInstance.setOnSynchronizationProblemAction(keyValues[1]);
							else if (keyValues[0].equalsIgnoreCase(KEY_NOTIFY_USER_FOR))
								currentInstance.displayErrorsIfGravityIs(keyValues[1]);*/
							else if (keyValues[0].equalsIgnoreCase(KEY_SAVE_ERRORS_LOG))
								; // TODO Save_errors_log => Nothing to update !
							/*else if (keyValues[0].equalsIgnoreCase(KEY_ERROR_ON_IMAGE_LOADING))
								currentInstance.onImageLoadingError(keyValues[1]);
							else if (keyValues[0].equalsIgnoreCase(KEY_DISPLAY_LOADING_ERROR))
								currentInstance.displayImageLoadingError(keyValues[1]);*/
							else
								System.err.println("### WARNING [ConfigurationFile] : Unknown key \""+keyValues[0]+"\" !");
						}else
							System.err.println("### WARNING [ConfigurationFile] : Syntax error at line "+numLine+" ! The correct syntax is: \"key = value\".");
					}
					numLine++;
				}
				reader.close();
				loaded = true;
			} catch (FileNotFoundException e) {
				ErrorsLog.getCurrentInstance().addException(new SEDException(e, "Opening the configuration file \""+configFile.getAbsolutePath()+"\"."));
				System.err.println("### WARNING [ConfigurationFile] : Impossible to open the configuration file \""+configFile.getAbsolutePath()+"\" because: \n\t=> "+e.getMessage());
			} catch (IOException e) {
				ErrorsLog.getCurrentInstance().addException(new SEDException(e, "Reading the line "+numLine+" of the configuration file \""+configFile.getAbsolutePath()+"\"."));
				System.err.println("### WARNING [ConfigurationFile] : Error while reading the configuration file \""+configFile.getAbsolutePath()+"\" at line "+numLine+" because: \n\t=> "+e.getMessage());
			}
		}else{
			ErrorsLog.getCurrentInstance().addException(new SEDException("Impossible to load the configuration file.", "Checking the readability of the configuration file.", "Impossible to load the configuration file \""+configFile.getAbsolutePath()+"\" because it is not a file, or it does not exist or there is no read authorization."));
			System.err.println("### WARNING [ConfigurationFile] : Impossible to read the configuration file \""+configFile.getAbsolutePath()+"\" ! ###");
		}
		if (!loaded)
			currentInstance = new UserPreferences();
	}

	public static final void saveCurrentInstance(File configFile) {
		if (currentInstance != null && !configFile.exists() || (configFile.isFile() && configFile.canWrite())){
			try {
				BufferedWriter writer = new BufferedWriter(new FileWriter(configFile));

				writer.write("# When the SEDPlugin must not interact anymore with Aladin ? "+NEVER+" for \"never\" or "+AFTER_EACH_ONE+" for \"after each interaction\"\n");
				writer.write(KEY_DEACTIVATE_INTERACTIONS_KEY+" = "+currentInstance.deactivateInteractions+"\n\n");

				/*writer.write("# Create a SED when clicking on an image in Aladin ?\n");
				writer.write(KEY_CREATE_SED_ON_CLICK+" = "+currentInstance.createSEDOnClick+"\n\n");*/

				writer.write("# Show the corresponding Aladin plan (image or figure) when clicking on a SED point or a SED ?\n");
				writer.write(KEY_MOVE_ON_SPECTRUM_CLICK+" = "+currentInstance.moveToOnSpectrumClick+"\n\n");

				writer.write("# Display coordinates of the a SED point when the mouse is over it ?\n");
				writer.write(KEY_PRINT_VALUES+" = "+currentInstance.printPointValuesOnOver+"\n\n");

				writer.write("# Ask confirmation before sending a SED by SAMP ?\n");
				writer.write(KEY_ASK_CONFIRMATION_SEND_SAMP+" = "+currentInstance.askConfirmationBeforeSendToSamp+"\n\n");

				writer.write("# When the SEDPlugin must be synchronized with Aladin ? "+ON_ACTIVATE+" for \"SEDPlugin activation\", "+ON_ANY_ALADIN_MODIFICATION+" for \"aladin modification\" or "+ON_USER_DEMAND+" for \"user demand\"\n");
				writer.write(KEY_REFRESH_ALL_ON_KEY+" = "+currentInstance.refreshAllOn+"\n\n");

				/*writer.write("# What must be done on Aladin modifications (image/figure removed, added, modified) ? "+ASK+" for \"ask\", "+APPLY_ALL+" for \"apply all\" or "+IGNORE_ALL+" for \"ignore all\"\n");
				writer.write(KEY_SYNCHRONIZATION_PROBLEM+" = "+currentInstance.synchronizationProblem+"\n\n");*/

				/*writer.write("# An error must be shown to the user when its gravity is greater or equal ...? "+SEDException.WARNING+" for \"WARNING\", "+SEDException.ERROR+" for \"ERROR\" or "+SEDException.CRITICAL_ERROR+" for \"CRITICAL_ERROR\"\n");
				writer.write(KEY_NOTIFY_USER_FOR+" = "+currentInstance.notifyUserFor+"\n\n");*/

				/*writer.write("# What must be done if an image of the Aladin's image stack can not be loaded ? "+IGNORE_ALL+" for \"ignore image\" or "+DELETE_IMAGE+" for \"delete image\"\n");
				writer.write(KEY_ERROR_ON_IMAGE_LOADING+" = "+currentInstance.errorOnImageLoading+"\n\n");

				writer.write("# Display image loading errors ?\n");
				writer.write(KEY_DISPLAY_LOADING_ERROR+" = "+currentInstance.displayImageLoadingError+"\n");*/

				writer.close();
			} catch (IOException e) {
				ErrorsLog.getCurrentInstance().addException(new SEDException(e, "Saving the configuration file \""+configFile.getAbsolutePath()+"\"."));
				System.err.println("### WARNING [ConfigurationFile] : Impossible to open the configuration file \""+configFile.getAbsolutePath()+"\" because: \n\t=> "+e.getMessage());
			}
		}
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	public static UserPreferences getCurrentInstance(){
		return currentInstance;
	}

	public static void setCurrentInstance(UserPreferences up){
		currentInstance = up;
	}

	/*
	 * What must be done when a problem occurs during a synchronization/refresh ?
	 * @return	{@link UserPreferences#ASK}, {@link UserPreferences#APPLY_ALL} or {@link UserPreferences#IGNORE_ALL}<i>(default)</i>.
	 *
	public final int onSynchronizationProblemAction(){
		return synchronizationProblem;
	}

	/*
	 * What must be done when a problem occurs during a synchronization/refresh ?
	 * @param action	{@link UserPreferences#ASK}, {@link UserPreferences#APPLY_ALL} or {@link UserPreferences#IGNORE_ALL}<i>(default)</i>.
	 *
	public final void setOnSynchronizationProblemAction(int action){
		if (action == ASK || action == APPLY_ALL || action == IGNORE_ALL)
			synchronizationProblem = action;
	}

	/*
	 * What must be done when a problem occurs during a synchronization/refresh ?
	 * @param action	{@link UserPreferences#ASK}, {@link UserPreferences#APPLY_ALL} or {@link UserPreferences#IGNORE_ALL}<i>(default)</i>.
	 *
	protected void setOnSynchronizationProblemAction(String action){
		try{
			setOnSynchronizationProblemAction(Integer.parseInt(action));
		}catch(NumberFormatException nfe){
			ErrorsLog.getCurrentInstance().addException(new SEDException(nfe, "Setting the option OnSynchronizationProblem"));
			//System.err.println("### WARNING [ConfigurationFile] : Unknown value for \""+KEY_SYNCHRONIZATION_PROBLEM+"\": "+action+" ! ###");
		}
	}*/

	/*
	 * What must be done when an error occurs during an image loading in Aladin ?
	 * @return	{@link UserPreferences#ASK}, {@link UserPreferences#IGNORE_ALL} or {@link UserPreferences#DELETE_IMAGE}.
	 *
	public final int onImageLoadingError(){
		return errorOnImageLoading;
	}

	/*
	 * What must be done when an error occurs during an image loading in Aladin ?
	 * @param action	{@link UserPreferences#IGNORE_ALL}<i>(default)</i> or {@link UserPreferences#DELETE_IMAGE}.
	 *
	public final void onImageLoadingError(int action){
		if (action == IGNORE_ALL || action == DELETE_IMAGE)
			errorOnImageLoading = action;
	}

	/**
	 * What must be done when an error occurs during an image loading in Aladin ?
	 * @param action	{@link UserPreferences#IGNORE_ALL}<i>(default)</i> or {@link UserPreferences#DELETE_IMAGE}.
	 *
	protected void onImageLoadingError(String action){
		try{
			onImageLoadingError(Integer.parseInt(action));
		}catch(NumberFormatException nfe){
			ErrorsLog.getCurrentInstance().addException(new SEDException(nfe, "Setting the option OnImageLoadingError"));
			//System.err.println("### WARNING [ConfigurationFile] : Unknown value for \""+KEY_ERROR_ON_IMAGE_LOADING+"\": "+action+" ! ###");
		}
	}

	public final boolean displayImageLoadingErrors(){
		return displayImageLoadingError;
	}

	public final void displayImageLoadingError(boolean display){
		displayImageLoadingError = display;
	}

	protected final void displayImageLoadingError(String display){
		if (display.equalsIgnoreCase("true"))
			displayImageLoadingError = true;
		else if (display.equalsIgnoreCase("false"))
			displayImageLoadingError = false;
		else
			System.err.println("### WARNING [ConfigurationFile] : Expected \"true\" or \"false\" for \""+KEY_DISPLAY_LOADING_ERROR+"\": "+display+" forbidden ! ###");

	}*/

	/**
	 * When synchronization with Aladin and the refresh of all spectrums must occur ?
	 * @return {@link UserPreferences#ON_USER_DEMAND} or {@link UserPreferences#ON_ACTIVATE}<i>(default)</i>.
	 */
	public final int whenAllRefreshing(){
		return refreshAllOn;
	}

	/**
	 * When synchronization with Aladin and the refresh of all spectrums must occur ?
	 * @param action	{@link UserPreferences#ON_USER_DEMAND} or {@link UserPreferences#ON_ACTIVATE}<i>(default)</i> or {@link UserPreferences#ON_ANY_ALADIN_MODIFICATION}.
	 */
	public final void setRefreshAllOn(int action){
		if (action == ON_ACTIVATE || action == ON_ANY_ALADIN_MODIFICATION || action == ON_USER_DEMAND)
			refreshAllOn = action;
	}

	/**
	 * When synchronization with Aladin and the refresh of all spectrums must occur ?
	 * @param action	{@link UserPreferences#ON_USER_DEMAND}, {@link UserPreferences#ON_ACTIVATE}<i>(default)</i> or {@link UserPreferences#ON_ANY_ALADIN_MODIFICATION}.
	 */
	protected void setRefreshAllOn(String action){
		try{
			setRefreshAllOn(Integer.parseInt(action));
		}catch(NumberFormatException nfe){
			ErrorsLog.getCurrentInstance().addException(new SEDException(nfe, "Setting the option RefreshAllOn"));
			//System.err.println("### WARNING [ConfigurationFile] : Unknown value for \""+KEY_REFRESH_ALL_ON_KEY+"\": "+action+" ! ###");
		}
	}

	/**
	 * When interactions with Aladin must be deactivated ?
	 * @return	{@link UserPreferences#NEVER}<i>(default)</i> or {@link UserPreferences#AFTER_EACH_ONE}.
	 */
	public final int whenDeactivatingInteractions(){
		return deactivateInteractions;
	}

	/**
	 * When interactions with Aladin must be deactivated ?
	 * @param action	{@link UserPreferences#NEVER}<i>(default)</i> or {@link UserPreferences#AFTER_EACH_ONE}.
	 */
	public final void deactivateInteractionsOn(int action){
		if (action == NEVER || action == AFTER_EACH_ONE)
			deactivateInteractions = action;
	}

	/**
	 * When interactions with Aladin must be deactivated ?
	 * @param action	{@link UserPreferences#NEVER}<i>(default)</i> or {@link UserPreferences#AFTER_EACH_ONE}.
	 */
	protected void deactivateInteractionsOn(String action) {
		try{
			deactivateInteractionsOn(Integer.parseInt(action));
		}catch(NumberFormatException nfe){
			ErrorsLog.getCurrentInstance().addException(new SEDException(nfe, "Setting the option DeactivateInteractionsOn"));
			//System.err.println("### WARNING [ConfigurationFile] : Unknown value for \""+KEY_DEACTIVATE_INTERACTIONS_KEY+"\": "+action+" ! ###");
		}
	}

	/*public final int displayErrorsIfGravityIs(){
		return notifyUserFor;
	}

	public final void displayErrorsIfGravityIs(int gravity){
		/*if (gravity >= SEDException.WARNING && gravity <= SEDException.CRITICAL_ERROR)
			notifyUserFor = gravity;*
	}

	protected void displayErrorsIfGravityIs(String gravity){
		try{
			displayErrorsIfGravityIs(Integer.parseInt(gravity));
		}catch(NumberFormatException nfe){
			ErrorsLog.getCurrentInstance().addException(new SEDException(nfe, "Setting the option DisplayErrorsIfGravityIs"));
			System.err.println("### WARNING [ConfigurationFile] : Unknown value for \""+KEY_NOTIFY_USER_FOR+"\": "+gravity+" ! ###");
		}
	}*/

	public final boolean saveErrorsLogBeforeClosing(){
		return saveErrorsLog;
	}

	public final void saveErrorsLogBeforeClosing(boolean save){
		saveErrorsLog = save;
	}

	protected void saveErrorsLogBeforeClosing(String save){
		if (save.equalsIgnoreCase("true"))
			saveErrorsLog = true;
		else if (save.equalsIgnoreCase("false"))
			saveErrorsLog = false;
		else
			System.err.println("### WARNING [ConfigurationFile] : Expected \"true\" or \"false\" for \""+KEY_SAVE_ERRORS_LOG+"\": "+save+" forbidden ! ###");
	}

	/*
	 * Create a SED when clicking on an image ?
	 * @return <i>true</i> to create a SED on each click on an image<i>false</i> otherwise.
	 *
	public final boolean createSEDOnClick(){
		return createSEDOnClick;
	}

	/*
	 * Create a SED when clicking on an image ?
	 * @param create	<i>true</i> to create a SED on each click on an image <i>false</i> otherwise.
	 *
	public final void setCreateSEDOnClick(boolean create){
		createSEDOnClick = create;
	}

	/*
	 * Create a SED when clicking on an image ?
	 * @param create	"true" to create a SED on each click on an image "false" otherwise.
	 *
	protected void setCreateSEDOnClick(String create){
		if (create.equalsIgnoreCase("true"))
			createSEDOnClick = true;
		else if (create.equalsIgnoreCase("false"))
			createSEDOnClick = false;
		else
			System.err.println("### WARNING [ConfigurationFile] : Expected \"true\" or \"false\" for \""+KEY_CREATE_SED_ON_CLICK+"\": "+create+" forbidden ! ###");
	}*/

	/**
	 * Move the Aladin reticule when clicking on a Spectrum list item ?
	 * @return <i>true</i> to move on each click on a spectrum <i>false</i> else.
	 */
	public final boolean moveToOnSpectrumClick(){
		return moveToOnSpectrumClick;
	}

	/**
	 * Move the Aladin reticule when clicking on a Spectrum list item ?
	 * @param move	<i>true</i> to move on each click on a spectrum <i>false</i> otherwise.
	 */
	public final void setMoveToOnSpectrumClick(boolean move){
		moveToOnSpectrumClick = move;
	}

	/**
	 * Move the Aladin reticule when clicking on a Spectrum list item ?
	 * @param move	"true" to move on each click on a spectrum "false" otherwise.
	 */
	protected void setMoveToOnSpectrumClick(String move){
		if (move.equalsIgnoreCase("true"))
			moveToOnSpectrumClick = true;
		else if (move.equalsIgnoreCase("false"))
			moveToOnSpectrumClick = false;
		else
			System.err.println("### WARNING [ConfigurationFile] : Expected \"true\" or \"false\" for \""+KEY_MOVE_ON_SPECTRUM_CLICK+"\": "+move+" forbidden ! ###");
	}

	/**
	 * Spectrum point values visible when mouse is over in the graph panel ?
	 * @return <i>true</i> if spectrum point values must be visible <i>false</i> else.
	 */
	public final boolean printValuesOnOver(){
		return printPointValuesOnOver;
	}

	/**
	 * Spectrum point values visible when mouse is over in the graph panel ?
	 * @param print	<i>true</i> if spectrum point values must be visible <i>false</i> else.
	 */
	public final void setPrintValuesOnOver(boolean print){
		printPointValuesOnOver = print;
	}

	/**
	 * Spectrum point values visible when mouse is over in the graph panel ?
	 * @param print	"true" if spectrum point values must be visible "false" otherwise.
	 */
	protected void setPrintValuesOnOver(String print){
		if (print.equalsIgnoreCase("true"))
			printPointValuesOnOver = true;
		else if (print.equalsIgnoreCase("false"))
			printPointValuesOnOver = false;
		else
			System.err.println("### WARNING [ConfigurationFile] : Expected \"true\" or \"false\" for \""+KEY_PRINT_VALUES+"\": "+print+" forbidden ! ###");
	}

	/**
	 * Ask ordinate unit confirmation before sending a spectrum via SAMP ?
	 * @return <i>true</i> if an unit confirmation must be prompted <i>false</i> else.
	 */
	public final boolean askConfirmationBeforeSendToSamp(){
		return askConfirmationBeforeSendToSamp;
	}

	/**
	 * Ask ordinate unit confirmation before sending a spectrum via SAMP ?
	 * @param ask	<i>true</i> if an unit confirmation must be prompted <i>false</i> else.
	 */
	public final void setAskConfirmationBeforeSendToSamp(boolean ask){
		askConfirmationBeforeSendToSamp = ask;
	}

	/**
	 * Ask ordinate unit confirmation before sending a spectrum via SAMP ?
	 * @param ask	"true" if an unit confirmation must be prompted "false" otherwise.
	 */
	protected void setAskConfirmationBeforeSendToSamp(String ask){
		if (ask.equalsIgnoreCase("true"))
			askConfirmationBeforeSendToSamp = true;
		else if (ask.equalsIgnoreCase("false"))
			askConfirmationBeforeSendToSamp = false;
		else
			System.err.println("### WARNING [ConfigurationFile] : Expected \"true\" or \"false\" for \""+KEY_ASK_CONFIRMATION_SEND_SAMP+"\": "+ask+" forbidden ! ###");
	}

}
