package sedplugin.sed;

import java.awt.image.BufferedImage;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.PrintStream;

import java.util.Date;

import javax.imageio.ImageIO;

import javax.swing.SwingWorker;

import sedplugin.errors.SEDException;

import sedplugin.graphe.Graph;
import sedplugin.graphe.Serie;

import sedplugin.interfaceGraphique.ExportFileFilter;

import cds.image.EPSGraphics;

/**
 * Exports the figure represented by the given graph
 * <b>or</b> one of the spectrums drawn in the given graph.<BR />
 * The returned value is a boolean telling wether the exportation file exists and so if the exportation is a success.
 * 
 * @author Gregory Mantelet
 * @version 03/2012
 */
public class ExportTask extends SwingWorker<File, String> {

	/** ID of the image formats (jpeg, bmp, png). */
	public static final int IMAGE = 0;

	/** ID of the EPS format. */
	public static final int EPS = 1;

	/** ID of the FITS format. */
	public static final int FITS = 2;

	/** ID of the XML format. */
	public static final int XML = 3;

	/** ID of the VOTable format. */
	public static final int VOTABLE = 4;

	/** ID of the ASCII formats (.tsv, .csv). */
	public static final int ASCII = 5;

	/** The choosen export format. */
	protected final int exportFormat;

	/** The destination file. */
	protected final File exportFile;

	/** The Graph object to use to export the figure or one of the drawn spectrums. */
	protected final Graph myGraph;

	/** The indice of the spectrum to export. */
	protected final int indSpectrum;

	/** A message which gives a little description of the current step. */
	protected String taskMessage;

	/** The exception which has stopped the task. */
	protected SEDException endException;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Intialize the export task.<BR />
	 * In the spectrum exportation case the first drawn spectrum in the given graph will be exported.
	 * 
	 * @param g					The graph which must be used to export either the whole figure or the first drawn spectrum.
	 * @param format			The export format ({@link ExportTask#IMAGE}, {@link ExportTask#EPS} and {@link ExportTask#FITS} to export the figure & {@link ExportTask#XML}, {@link ExportTask#VOTABLE} and {@link ExportTask#ASCII} to export a spectrum).
	 * @param f					The destination file <i>(<u>note</u>: if existing it will be deleted before exportation)</i>.
	 * @throws SEDException		If one of the given parameters is incorrect (<i>format</i> will be IMAGE, EPS, FITS, XML, VOTABLE or ASCII; <i>f</i> and <i>g</i> must be non-null).
	 */
	public ExportTask(Graph g, int format, File f) throws SEDException {
		this(g, 0, format, f);
	}

	/**
	 * Intialize the export task.
	 * 
	 * @param g					The graph which must be used to export either the whole figure or the first drawn spectrum.
	 * @param indSp				Indice of the spectrum to export.
	 * @param format			The export format ({@link ExportTask#IMAGE}, {@link ExportTask#EPS} and {@link ExportTask#FITS} to export the figure & {@link ExportTask#XML}, {@link ExportTask#VOTABLE} and {@link ExportTask#ASCII} to export a spectrum).
	 * @param f					The destination file <i>(<u>note</u>: if existing it will be deleted before exportation)</i>.
	 * @throws SEDException		If one of the given parameters is incorrect (<i>indSp</i> must be a valid indice; <i>format</i> will be IMAGE, EPS, FITS, XML, VOTABLE or ASCII; <i>f</i> and <i>g</i> must be non-null).
	 */
	public ExportTask(Graph g, int indSp, int format, File f) throws SEDException {
		if (g == null)
			throw new SEDException("Exportation is impossible", "ExportTask constructor", "Impossible to export anything because the given Graph object is null !");
		if (f == null)
			throw new SEDException("Exportation is impossible", "ExportTask constructor", "Impossible to export anything without a destination file !");
		if (indSp < 0 || indSp > g.getNbSeries())
			throw new SEDException("Exportation is impossible", "ExportTask constructor", "Impossible to export the "+indSp+"-th drawn spectrum because it doesn't exist !");
		if (format<0 || format>5)
			throw new SEDException("Exportation is impossible", "ExportTask constructor", "No exportation possible without specifying a valid format !");

		exportFormat = format;
		exportFile = f;
		myGraph = g;
		indSpectrum = indSp;

		taskMessage = "";
		endException = null;
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Gets the current message detailing the progression step.
	 * @return	Details about the current progression step.
	 */
	public final String getProgressionMessage(){
		return taskMessage;
	}

	/**
	 * Updates the current step information and fire all register PropertyChangeListener instances.
	 * @param progress	The progression value (in %). <i>To get its value use the {@link SwingWorker#getProgress()} method.</i>
	 * @param msg		Description of the current step. <i>To get this message use the {@link ASpectrumsManagementTask#getProgressionMessage()} method.</i>
	 */
	protected void setProgress(int progress, String msg){
		taskMessage = msg;
		setProgress(progress);
	}

	/**
	 * Gets the exception which causes the end of the task.
	 * @return	The exception which stops the task.
	 */
	public final SEDException getEndException(){
		return endException;
	}

	/**
	 * Sets cancel flag to true, delete destination file and creates a SEDException from the given exception and with the given source and message.
	 * @param ex		The source exception.
	 * @param origin	The exception origin.
	 */
	protected void generateException(Exception ex, String origin){
		cancel(true);
		exportFile.delete();
		endException = new SEDException(ex, origin);
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	@Override
	protected File doInBackground() {
		String extension = ExportFileFilter.getExtension(exportFile);
		BufferedImage image;
		BufferedWriter textOrXml;

		// save it in the wanted format:
		switch(exportFormat){
		case IMAGE:
			setProgress(0, "creating image...");
			// load image:
			image = myGraph.exportAsImage();
			setProgress(50, "writting file...");
			try {
				ImageIO.write(image, extension, exportFile);
			} catch (IOException ex) {
				generateException(ex, "Exporting in "+extension.toUpperCase()+" into the file \""+exportFile.getAbsoluteFile()+"\".");
			}
			break;
		case EPS:
			setProgress(0, "creating image...");
			// load image:
			image = myGraph.exportAsImage();
			setProgress(30, "formatting in EPS...");
			try {
				PrintStream ps = new PrintStream(new FileOutputStream(exportFile));
				EPSGraphics epsg = new EPSGraphics(ps, ((myGraph.getName() != null && myGraph.getName().trim().length()>0)?myGraph.getName():exportFile.getName()), null, 0, 0, myGraph.getWidth(), myGraph.getHeight());
				setProgress(60, "writting file...");
				epsg.drawImage(image, 0, 0, myGraph);
				epsg.end();
				ps.close();
			} catch (FileNotFoundException ex) {
				generateException(ex, "Exporting in EPS into the file \""+exportFile.getAbsoluteFile()+"\".");
			}
			break;
		case XML:
			setProgress(0, "opening file...");
			try{
				textOrXml = new BufferedWriter(new FileWriter(exportFile));
				Serie serie = myGraph.getSerie(indSpectrum);
				if (serie == null){
					cancel(true);
					endException = new SEDException("Exportation failed", "Getting the "+indSpectrum+"-th serie.", "The "+indSpectrum+"-th serie doesn't exist !");
					return null;
				}

				// write header:
				setProgress(25, "writting header...");
				textOrXml.write("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n");
				textOrXml.write("<Spectrum xmlns=\"http://www.ivoa.net/xml/VOTable/v1.2\"\n\t\txmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\"\n\t\txsi:schemaLocation=\"http://www.ivoa.net/xml/Spectrum/v1.01\">\n");
				// dataModel:
				textOrXml.write("\t<DataModel>Spectrum-1.0</DataModel>\n");
				// coordSystem:
				textOrXml.write("\t<CoordSys id=\"ID000001\">\n");
				textOrXml.write("\t\t<SpaceFrame>\n");
				textOrXml.write("\t\t\t<Name ucd=\"pos.frame\">ICRS</Name>\n");
				textOrXml.write("\t\t\t<Equinox ucd=\"time.equinox;pos.eq\">2000.0</Equinox>\n");
				textOrXml.write("\t\t</SpaceFrame>\n");
				textOrXml.write("\t\t<TimeFrame>\n");
				textOrXml.write("\t\t\t<Name ucd=\"time.scale\">UTC</Name>\n");
				textOrXml.write("\t\t\t</TimeFrame>\n");
				textOrXml.write("\t\t<SpectralFrame>\n");
				textOrXml.write("\t\t\t<ReferencePosition>BARYCENTER</ReferencePosition>\n");
				textOrXml.write("\t\t</SpectralFrame>\n");
				textOrXml.write("\t</CoordSys>\n");

				// write the first selected spectrum:
				setProgress(50, "writting SED...");
				SED sp = serie.getSED();
				if (sp != null)
					textOrXml.write(sp.toXMLFormat("\t"));

				// write footer:
				setProgress(75, "writting footer...");
				textOrXml.write("</Spectrum>\n");

				textOrXml.close();
			}catch(IOException ex){
				generateException(ex, "Exporting in XML into the file \""+exportFile.getAbsoluteFile()+"\".");
			}
			break;
		case VOTABLE:
			setProgress(0, "opening file...");
			try{
				textOrXml = new BufferedWriter(new FileWriter(exportFile));
				Serie serie = myGraph.getSerie(indSpectrum);
				if (serie == null){
					cancel(true);
					endException = new SEDException("Exportation failed", "Getting the "+indSpectrum+"-th serie.", "The "+indSpectrum+"-th serie doesn't exist !");
					return null;
				}

				// write header:
				setProgress(25, "writting header...");
				textOrXml.write("<?xml version=\"1.0\" encoding=\"UTF-8\" ?>\n");
				textOrXml.write("<VOTABLE version = \"1.1\"\n\t\txmlns:xsi = \"http://www.w3.org/2001/XMLSchema-instance\"\n\t\txsi:noNamespaceSchemaLocation = \"xmlns:http://www.ivoa.net/xml/VOTable/VOTable-1.2.xsd\"\n\t\txmlns:spec = \"http://www.ivoa.net/xml/Spectrum/v1.01\"\n\t\txmlns = \"http://www.ivoa.net/xml/VOTable/v1.2\">\n");
				textOrXml.write("\t<RESOURCE utype=\"spec:Spectrum\">\n");
				textOrXml.write("\t\t<TABLE utype=\"spec:Spectrum\">\n");
				// dataModel:
				textOrXml.write("\t\t\t<PARAM name=\"DataModel\" utype=\"DataModel\" datatype=\"char\" arraysize=\"*\" value=\"Spectrum-1.01\" />\n");
				// target:
				textOrXml.write("\t\t\t<GROUP utype=\"spec:Target\">\n");
				textOrXml.write("\t\t\t\t<PARAM name=\"Target\" utype=\"spec:Target.Name\" datatype=\"char\" arraysize=\"*\" value=\"null\" />\n");
				textOrXml.write("\t\t\t</GROUP>\n");
				// coordSystem:
				textOrXml.write("\t\t\t<GROUP name=\"CoordSys\" utype=\"spec:CoordSys\">\n");
				textOrXml.write("\t\t\t\t<GROUP utype=\"spec:CoordSys.SpaceFrame\">\n");
				textOrXml.write("\t\t\t\t\t<PARAM name=\"System\" utype=\"spec:CoordSys.SpaceFrame.Name\" ucd=\"pos.frame\" datatype=\"char\" arraysize=\"*\" value=\"ICRS\" />\n");
				textOrXml.write("\t\t\t\t\t<PARAM name=\"Equinox\" utype=\"spec:CoordSys.SpaceFrame.Equinox\" ucd=\"time.equinox;pos.eq\" datatype=\"float\" value=\"2000.0\" />\n");
				textOrXml.write("\t\t\t\t</GROUP>\n");
				textOrXml.write("\t\t\t\t<GROUP utype=\"spec:CoordSys.TimeFrame\">\n");
				textOrXml.write("\t\t\t\t\t<PARAM name=\"TimeFrame\" utype=\"spec:CoordSys.TimeFrame.Name\" ucd=\"time.scale\" datatype=\"char\" arraysize=\"*\" value=\"UTC\" />\n");
				textOrXml.write("\t\t\t\t</GROUP>\n");
				textOrXml.write("\t\t\t\t<GROUP utype=\"spec:CoordSys.SpectralFrame\">\n");
				textOrXml.write("\t\t\t\t\t<PARAM name=\"SpectralFrame\" utype=\"spec:CoordSys.SpectralFrame.RefPos\" datatype=\"char\" arraysize=\"*\" value=\"BARYCENTER\" />\n");
				textOrXml.write("\t\t\t\t</GROUP>\n");
				textOrXml.write("\t\t\t</GROUP>\n");

				// write the first selected spectrum:
				setProgress(50, "writting SED...");
				SED sp = serie.getSED();
				if (sp != null)
					textOrXml.write(sp.toVOTableFormat("\t\t\t"));

				// write footer:
				setProgress(75, "writting footer...");
				textOrXml.write("\t\t</TABLE>\n");
				textOrXml.write("\t</RESOURCE>\n");
				textOrXml.write("</VOTABLE>\n");

				textOrXml.close();
			}catch(IOException ex){
				generateException(ex, "Exporting in VOTable into the file \""+exportFile.getAbsoluteFile()+"\".");
			}
			break;
		case ASCII:
			setProgress(0, "opening file...");
			try{
				textOrXml = new BufferedWriter(new FileWriter(exportFile));
				Serie serie = myGraph.getSerie(indSpectrum);

				// get field separator string:
				String sep = (extension.equalsIgnoreCase("csv"))?Exportable.CSV:Exportable.TSV;

				// write header:
				setProgress(30, "writting header...");
				textOrXml.write("%% By Aladin "+cds.aladin.Aladin.VERSION+"/SEDPlugin "+sedplugin.plugin.SEDPlugin.VERSION+"\n");
				textOrXml.write("%% "+new Date()+"\n");

				// write all drawn (so selected) spectrums:
				setProgress(60, "writting SED...");
				SED sp = serie.getSED();
				if (sp != null)
					textOrXml.write(sp.toASCIIFormat(sep));
				textOrXml.close();
			}catch(IOException ex){
				generateException(ex, "Exporting in ASCII into the file \""+exportFile.getAbsoluteFile()+"\".");
			}
			break;
		}
		return exportFile;
	}

}
