package sedplugin.sed;

import java.util.Comparator;

/**
 * Lets comparing SED points on a given axis, so that ordering points in a SED.
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 03/2011
 * 
 * @see SED
 * @see SEDPoint
 */
public final class PointComparator implements Comparator<SEDPoint> {

	/** Accepted error precision when comparing 2 double values.*/
	public final static double COMP_EPSILON = 1e-5;

	/** Constant for {@link SED#getPointOrder()} and {@link SED#setPointOrder(int)} functions: No order must be applied on spectrum points collection. */
	public final static int NO_ORDER = 0;
	/** Constant for {@link SED#getPointOrder()} and {@link SED#setPointOrder(int)} functions: Spectrum points must be sorted by ascending abscissa value. */
	public final static int ABSCISSA_ORDER = 1;
	/** Constant for {@link SED#getPointOrder()} and {@link SED#setPointOrder(int)} functions: Spectrum points must be sorted by ascending ordinate value. */
	public final static int ORDINATE_ORDER = 2;

	/** Indicates how SED points must be sorted.
	 * @see SED#NO_ORDER
	 * @see SED#ABSCISSA_ORDER
	 * @see SEd#ORDINATE_ORDER */
	protected int pointOrder;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds a SED point comparator on the abscissa axis.
	 */
	public PointComparator(){
		pointOrder = ABSCISSA_ORDER;
	}

	/**
	 * Builds a SED point comparator on the specified axis.
	 * 
	 * @param order	The axis on which points must be ordered ({@link #NO_ORDER}, {@link #ABSCISSA_ORDER} or {@link #ORDINATE_ORDER}).
	 */
	public PointComparator(int order){
		if (order >= 0 && order <= 2)
			pointOrder = order;
		else
			pointOrder = ABSCISSA_ORDER;
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Indicates on which axis points are ordered.
	 * 
	 * @return Its point order ({@link #NO_ORDER}, {@link #ABSCISSA_ORDER} or {@link #ORDINATE_ORDER}).
	 */
	public final int getPointOrder() {
		return pointOrder;
	}

	/**
	 * Lets indicate on which axis points must be ordered.
	 * 
	 * @param pointOrder Its new point order ({@link #NO_ORDER}, {@link #ABSCISSA_ORDER} or {@link #ORDINATE_ORDER}).
	 */
	public final boolean setPointOrder(int order) {
		if (order >= 0 && order <= 2){
			pointOrder = order;
			return true;
		}
		return false;
	}

	/* ***************** */
	/* COMPARISON METHOD */
	/* ***************** */

	/**
	 * Compares 2 double values.
	 * 
	 * @param v1	First value.
	 * @param v2	Second value.
	 * 
	 * @return	0 if both values are equal, 1 if the first value is greater, -1 if the first value is lower.
	 */
	public static final int compare(final double v1, final double v2){
		double diff = v1-v2;
		if (Math.abs(diff) < COMP_EPSILON)
			return 0;
		else if (diff > 0)
			return 1;
		else
			return -1;
	}

	/**
	 * Lets comparing two SED points according to the imposed order.
	 * 
	 * @see java.util.Comparator#compare(Object, Object)
	 * @see SEDPoint
	 * @see SED#getPointOrder()
	 * @see SED#setPointOrder(int)
	 */
	public int compare(SEDPoint obj1, SEDPoint obj2) {
		try{
			// Cast given objects:
			SEDPoint p1 = (SEDPoint)obj1, p2 = (SEDPoint)obj2;
			double p1X, p1Y;
			double p2X, p2Y;

			// If one or both spectrum points are null or if the two points are equals then return immediately:
			if (p1 == null || p2 == null || p1.equals(p2))
				return 0;
			else{	// else get their abscissa and ordinate values:
				p1X = p1.getX(); p1Y = p1.getY();
				p2X = p2.getX(); p2Y = p2.getY();
			}

			// Compare the two objects according to the pointOrder attribute:
			int comp;
			switch(pointOrder){
			case ABSCISSA_ORDER:
				comp = compare(p1X, p2X);
				return (comp == 0) ? compare(p1Y, p2Y) : comp;
			case ORDINATE_ORDER:
				comp = compare(p1Y, p2Y);
				return (comp == 0) ? compare(p1X, p2X) : comp;
			case NO_ORDER:
			default:
				return 1;
			}
		}catch(ClassCastException ex){
			return 0;
		}
	}

}
