package sedplugin.sed;

import java.text.DecimalFormat;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;

import cds.astro.Converter;
import cds.astro.Unit;

public final class ToolBox {

	/** Empty constructor => Singleton class. */
	private ToolBox() { ; }

	/** The format to use when numbers aren't so huge. */
	public final static DecimalFormat classicFormat = new DecimalFormat("##0.##");

	/** The number format to use for big number. */
	public final static DecimalFormat scientificFormat = new DecimalFormat("0.##E0");

	/**
	 * Gets the format to use considering the given value.
	 * @param value	An example of value.
	 * @return		The format to use.
	 */
	public final static DecimalFormat whichFormat(double value){
		DecimalFormat df = classicFormat;
		if (((int)value*100) == 0 || ((int)value/1000) != 0)
			df = scientificFormat;

		return df;
	}

	/**
	 * Formats the given value using the most appropriate decimal format.
	 * 
	 * @param value	The value to format.
	 * 
	 * @return		The formatted value.
	 * 
	 * @see #whichFormat(double)
	 */
	public final static String format(double value){
		return whichFormat(value).format(value);
	}

	/* *********************** */
	/* UNKNOWN UNIT MANAGEMENT */
	/* *********************** */
	/**
	 * Get an object that allows converting a value from <i>fromUnitSymbol</i> to <i>toUnitSymbol</i>.
	 *
	 * @param fromUnitSymbol	The source unit.
	 * @param toUnitSymbol		The destination unit.
	 * @return					A {@link Converter} instance.
	 *
	 * @see ASpectrumsManagementTask#getWavelengthConverter(String)
	 * @see ASpectrumsManagementTask#getFluxConverter(String)
	 */
	public static final Converter getAppropriateConverter(String fromUnitSymbol, String toUnitSymbol){
		final String STERADIAN = "sr";
		boolean compatible = true;
		Converter conv = null;

		// Have they both a "sr" or both nothing ?
		boolean inSR = fromUnitSymbol.indexOf(STERADIAN) > -1;
		compatible = inSR == (toUnitSymbol.indexOf(STERADIAN) > -1);

		try{
			if (compatible){
				// Correspond they to known unit ?
				Unit unit1 = new Unit(fromUnitSymbol), unit2 = new Unit(toUnitSymbol);
				compatible = unit1.isCompatibleWith(unit2);
				// Can we convert from the first unit to the second one ?
				if (compatible)
					conv = new Converter(fromUnitSymbol, toUnitSymbol);
			}
		}catch(Exception e){
			//	ErrorsLog.getCurrentInstance().addException(new SEDException("Conversion impossible", "Can't convert from "+fromUnitSymbol+" to "+toUnitSymbol, e.getMessage(), SEDException.CRITICAL_ERROR));
		}

		return conv;
	}

	/**
	 * Gets an object that allows converting a value from <i>fromUnitSymbol</i> to the reference quantity and unit of a WavelengthUnit.
	 * @param fromUnitSymbol	The source unit.
	 * @return					A {@link Converter} instance.
	 */
	public static final Converter getXConverter(String fromUnitSymbol){
		try{
			XAxis xUnit = new XAxis();
			xUnit.setQuantity(XAxis.WAVELENGTH);
			if (fromUnitSymbol == null || xUnit.getUnit().symbol == null)
				return null;
			Converter conv = getAppropriateConverter(fromUnitSymbol, xUnit.getUnit().symbol);
			if (conv == null){
				xUnit.setQuantity(XAxis.FREQUENCY);
				conv =  getAppropriateConverter(fromUnitSymbol, xUnit.getUnit().symbol);
			}
			return conv;
		}catch(SEDException se){
			ErrorsLog.getCurrentInstance().addException(se);
			return null;
		}
	}

	/**
	 * Gets an object that allows converting a value from <i>fromUnitSymbol</i> to the reference quantity and unit of a FluxUnit.
	 * @param fromUnitSymbol	The source unit.
	 * @return					A {@link Converter} instance.
	 */
	public static final Converter getYConverter(String fromUnitSymbol){
		if (fromUnitSymbol == null)
			return null;

		try{
			int[] types = new int[]{YAxis.INTENSITY, YAxis.NU_Inu, YAxis.FLUX};
			YAxis yUnit = new YAxis(new XAxis());
			Converter conv = null;

			for(int t=0; conv == null && t < types.length; t++){
				yUnit.setQuantity(types[t]);
				if (yUnit.getUnit().symbol != null)
					conv = getAppropriateConverter(fromUnitSymbol, yUnit.getUnit().symbol);
			}

			return conv;
		}catch(SEDException se){
			ErrorsLog.getCurrentInstance().addException(se);
			return null;
		}
	}

	/**
	 * Tests whether the given unit is a valid flux unit, and more particularly if it is a temperature
	 * in K [CMB] or in K [RJ].
	 * @param fluxUnit		Unit to check.
	 * @return				<i>true</i> if the given unit is a flux unit, <i>false</i> otherwise.
	 */
	public static boolean isYUnit(final String fluxUnit) {
		if (fluxUnit == null || fluxUnit.trim().isEmpty())
			return false;

		int indTempType = fluxUnit.lastIndexOf('[');
		if (indTempType > 0){
			String temperatureType = fluxUnit.substring(indTempType).trim();
			if (temperatureType.equalsIgnoreCase("[CMB]") || temperatureType.equalsIgnoreCase("[RJ]"))
				return (getAppropriateConverter(fluxUnit.substring(0, indTempType).trim(), "K") != null);
		}

		return (getYConverter(fluxUnit) != null);
	}

	/**
	 * <p>Normalizes the given flux unit.</p>
	 * <p>This method is useful only for the 2 special cases: "K [CMB]" and "K [RJ]". The last part
	 * (between square brackets) must be without space and in upper case. The 2 parts must be separated only by one space.</p>
	 * @param fluxUnit	Flux unit to normalize.
	 * @return			The same unit but normalized.
	 */
	public static String normalizeYUnit(final String fluxUnit) {
		if (fluxUnit == null)
			return null;

		int indTempType = fluxUnit.lastIndexOf('[');
		if (indTempType > 0){
			String temperatureType = fluxUnit.substring(indTempType).trim();
			if (temperatureType.equalsIgnoreCase("[CMB]") || temperatureType.equalsIgnoreCase("[RJ]"))
				return fluxUnit.substring(0, indTempType).trim()+" "+temperatureType.toUpperCase();
		}
		return fluxUnit.trim();
	}
}
