package sedplugin.sed.source;

import java.beans.PropertyChangeListener;
import java.util.Iterator;
import java.util.Vector;

import sedplugin.sed.SED;

/**
 * <p>Abstract implementation of the interface {@link SEDSource}.</p>
 * 
 * <p>All the behaviors of a SED source are already implemented. Only the way to
 * load the SED is really required by an extension of this class. The {@link SEDLoader} to
 * return will be the only to load a SED. This class is only a link between the source of a SED (i.e. a File)
 * and the generated SED. It uses the loader to re-generate the SED at the creation and afterwards if needed.</p>
 * 
 * <p>A SED source may be in a sources group. For instance, if a file can contain several SED, there will be one
 * SED source of type {@link SourcesGroup} which gathers several single (or grouped) SED sources.</p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 06/2012
 * 
 * @see SEDLoader
 * @see SourcesGroup
 */
public abstract class AbstractSEDSource implements SEDSource {

	protected final SourcesGroup parent;

	protected boolean frozen = false;
	protected SED sed = null;

	private SEDLoader loader = null;

	protected final Vector<SEDSourceListener> listeners = new Vector<SEDSourceListener>();

	public AbstractSEDSource(){
		parent = null;
	}

	public AbstractSEDSource(final SourcesGroup parent){
		this.parent = parent;
	}

	public boolean addSEDSourceListener(SEDSourceListener l) {
		if (listeners.contains(l))
			return false;
		else
			return listeners.add(l);
	}

	public boolean removeSEDSourceListener(SEDSourceListener l) {
		return listeners.remove(l);
	}

	protected void sedAdded(){
		if (parent != null)
			parent.sedAdded(sed, this);
		for(SEDSourceListener l : listeners)
			l.sedAdded(sed, this);
	}

	protected void sedRemoved(final SED removedSed){
		if (parent != null)
			parent.sedRemoved(removedSed, this);
		for(SEDSourceListener l : listeners)
			l.sedRemoved(removedSed, this);
	}

	public void freeze(boolean freezeOrNot) {
		frozen = freezeOrNot;
	}

	public boolean isFrozen() {
		return frozen;
	}

	public SED getSED() {
		if (sed == null)
			update();
		return sed;
	}

	public boolean hasSED() {
		return (sed != null);
	}

	public void removeSED() {
		if (sed != null){
			SED removedSed = sed;
			sed = null;
			removedSed.setSource(null);
			sedRemoved(removedSed);
		}
	}

	/**
	 * Tells whether this SED source has been modified since the last update.
	 * 
	 * @return	<i>true</i> if this source has been recently modified, <i>false</i> otherwise.
	 * 
	 * @see #updateModifiedFields()
	 */
	protected abstract boolean isModified();

	/**
	 * Update any field used to know whether this source needs to be updated.
	 * 
	 * @see #isModified()
	 */
	protected abstract void updateModifiedFields();

	/**
	 * Creates a new SED (to fill) for this SED.
	 * 
	 * @return	The new EMPTY SED.
	 */
	protected abstract SED createEmptySED();

	/**
	 * Creates a new SED loader.
	 * 
	 * @return	The new SED loader.
	 */
	protected abstract SEDLoader createLoader();

	public boolean update() {
		return update(false);
	}

	public boolean update(final boolean force){
		// Initialize the SED:
		if (sed == null){
			sed = createEmptySED();
			if (sed == null) return false;
			sedAdded();
		}else if (!force && !isModified())
			return true;
		else
			sed.clear();

		// Cancel the current loading if any:
		if (loader != null){
			if (loader.isDone() || loader.cancel(true))
				loader = null;
			else
				return false;
		}

		// Update the fields used to detect modifications:
		updateModifiedFields();

		// Initialize the loader:
		loader = createLoader();
		Iterator<PropertyChangeListener> it = sed.getLoadListeners();
		while(it.hasNext())
			loader.addPropertyChangeListener(it.next());

		// Start loading:
		loader.execute();

		return true;
	}

}
