package sedplugin.sed.source;

import sedplugin.sed.Exportable;
import sedplugin.sed.SED;

/**
 * <p>
 * 	A {@link SED} can be built from different kinds of "object": a file or, in Aladin, a position, a drawing (circle or polygon), a catalog, ...
 * 	To be able to synchronize/update easily a {@link SED} with its source, it will be always associated with its source.
 * 	So: 1 {@link SED} <-> 1 {@link SEDSource}.
 * </p>
 * <p>
 * 	However, some sources can generate more than only one {@link SED} (i.e. files and catalogs). These sources will be represented
 * 	by one special kind of SEDSource: {@link SourcesGroup}. A {@link SourcesGroup} may gather several SEDSource.
 * </p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 04/2012
 * 
 * @see SourcesGroup
 */
public interface SEDSource extends Exportable {

	/**
	 * Gets the name of this kind of source (i.e. aladin drawing, catalog, file...).
	 * 
	 * @return	Its kind.
	 */
	public String getType();

	/**
	 * Gets the name of this source.
	 * 
	 * @return Its name.
	 */
	public String getName();

	/**
	 * Updates this source (i.e.reload a drawing, a catalog, a file, ....) ONLY IF THERE IS A MODIFICATION OF THE SOURCE.
	 * 
	 * @return	<i>true</i> If this source has been successfully updated, <i>false</i> otherwise.
	 */
	public boolean update();

	/**
	 * Updates this source (i.e.reload a drawing, a catalog, a file, ....).
	 * 
	 * @param force		<i>true</i> to force the update even if there is no modification of the source, <i>false</i> otherwise.
	 * 
	 * @return	<i>true</i> If this source has been successfully updated, <i>false</i> otherwise.
	 */
	public boolean update(final boolean force);

	/**
	 * Especially for Aladin sources, it lets send a "signal" to highlight
	 * this source in Aladin (i.e. select a drawing, select a catalog item, ...).
	 */
	public void highlight();

	/**
	 * Gets the SED loaded from this source.
	 * 
	 * @return	Its corresponding SED.
	 */
	public SED getSED();

	/**
	 * Indicates whether a SED has been loaded from this source.
	 * 
	 * @return	<i>true</i> if there is a corresponding SED, <i>false</i> otherwise.
	 */
	public boolean hasSED();

	/**
	 * Destroys its corresponding SED.
	 */
	public void removeSED();

	/**
	 * Lets disable/enable the update ability of this source.
	 * 
	 * @param freezeOrNot	<i>true</i> to disable its update ability, <i>false</i> to enable it.
	 */
	public void freeze(boolean freezeOrNot);

	/**
	 * Indicates whether the update ability of this source is disabled/enabled.
	 * 
	 * @return <i>true</i> if its update ability is disabled, <i>false</i> if it's enabled.
	 */
	public boolean isFrozen();

	/**
	 * Adds an object to notify when a SED is added into or removed from this SED source.
	 * 
	 * @param l	The object to notify.
	 * @return	<i>true</i> if the listener has been successfully added, <i>false</i> otherwise (even if the given object is already a listener of this source).
	 */
	public boolean addSEDSourceListener(final SEDSourceListener l);

	/**
	 * Removes the given object from the list of listeners of this source.
	 * 
	 * @param l	The object to remove.
	 * @return	<i>true</i> if the given listener has been successfully removed from this source, <i>false</i> otherwise.
	 */
	public boolean removeSEDSourceListener(final SEDSourceListener l);

}
