package sedplugin.sed.source.aladin;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;


import cds.aladin.AladinData;
import cds.aladin.AladinException;
import cds.aladin.Obj;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.plugin.AladinImage;
import sedplugin.plugin.AladinProxy;
import sedplugin.plugin.SEDPlugin;
import sedplugin.sed.source.SourcesGroup;

public class Drawing extends SourcesGroup {

	public final static String DRAWINGS_TYPE = "DRAWINGS";

	/** Obj type of a circle (drawn by the user) intern object of a measures AladinData. To check if an Obj is a circle: <pre>obj.getObjType().equals(CIRCLE_OBJECT) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String CIRCLE_OBJECT = "Tag";

	/** Obj type of a photometry object (drawn by the user) intern object of a measures AladinData. To check if an Obj is a photometry: <pre>obj.getObjType().equals(PHOT_OBJECT) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String PHOT_OBJECT = "Phot";

	/** Obj type of a circle (drawn by a command) intern object of a measures AladinData. To check if an Obj is a circle: <pre>obj.getObjType().equals(CIRCLE_OBJECTBis) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String CIRCLE_OBJECTBis = "Circle";

	/** Obj type of a polygon point (begin or middle) intern object of a measures AladinData. To check if an Obj is a polygon point: <pre>obj.getObjType().equals(POLYGON_OBJECT) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String POLYGON_OBJECT = "Line+";

	/** Obj type of a polygon point (end only) intern object of a measures AladinData. To check if an Obj is a polygon point: <pre>obj.getObjType().equals(POLYGON_END) <i>(where </i>obj<i> is an Obj instance)</i></pre> */
	public static final String POLYGON_END = "Line";

	private final AladinProxy aladin;
	private final SEDPlugin plugin;

	protected final AladinData data;
	protected ArrayList<AladinImage> images;
	protected HashMap<Obj, PhotometrySource> drawings;

	private int nbObjs = -1;

	private boolean disconnectedFromAladin = false;

	public Drawing(AladinData plane, AladinProxy aladin, SEDPlugin plugin) {
		this.data = plane;
		this.aladin = aladin;
		this.plugin = plugin;
		this.images = new ArrayList<AladinImage>();
		this.drawings = new HashMap<Obj, PhotometrySource>(5);
	}

	public final Iterator<AladinImage> getImages(){
		return images.iterator();
	}

	public final Iterator<AladinImage> getCorrectImages(){
		return new CorrectImagesIterator(images);
	}

	public final int getNbImages(){
		return images.size();
	}

	public final int getNbCorrectImages(){
		Iterator<AladinImage> it = new CorrectImagesIterator(images);
		int nbImages = 0;
		while(it.hasNext()){
			it.next();
			nbImages++;
		}
		return nbImages;
	}

	public final AladinImage getImage(final int index){
		return images.get(index);
	}

	public final String getPolygonCoords(Obj obj){
		if (obj != null && isPolygon(obj.getObjType())){
			try{
				final Obj[] objs = data.seeObj();
				int ind = data.getNbObj()-1;
				while(ind > 0 && objs[ind] != obj)
					ind--;
				if (ind >= 0){
					StringBuffer buf = new StringBuffer();
					do{
						obj = objs[ind--];
						buf.append(obj.raj).append(' ').append(obj.dej).append(';');
					} while(ind >= 0 && objs[ind].getObjType().equals(POLYGON_OBJECT));
					return buf.toString();
				}
			}catch(AladinException ae){
				ErrorsLog.getCurrentInstance().addException(new SEDException(ae, "Drawing.getPolygonCoords(Obj)"));
			}
		}
		return null;
	}

	public double[] getPolygonCenter(Obj obj) {
		if (obj != null && isPolygon(obj.getObjType())){
			try{
				final Obj[] objs = data.seeObj();
				int ind = data.getNbObj()-1;
				while(ind > 0 && objs[ind] != obj)
					ind--;
				if (ind >= 0){
					double[] center = new double[]{0,0};
					int nbPoints = 0;
					do{
						obj = objs[ind--];
						center[0] += obj.raj; center[1] += obj.dej;
						nbPoints++;
					} while(ind >= 0 && objs[ind].getObjType().equals(POLYGON_OBJECT));
					center[0] /= nbPoints; center[1] /= nbPoints;
					return center;
				}
			}catch(AladinException ae){
				ErrorsLog.getCurrentInstance().addException(new SEDException(ae, "Drawing.getPolygonCenter(Obj)"));
			}
		}
		return null;
	}

	protected final boolean isModified(){
		try{
			return (data.getNbObj() != nbObjs);
		}catch(AladinException ae){
			ErrorsLog.getCurrentInstance().addException(new SEDException(ae, "Drawing.isModified()"));
			return true;
		}
	}

	@Override
	public void freeze(boolean freezeOrNot) {
		if (isDisconnectedFromAladin())
			frozen = true;
		else
			super.freeze(freezeOrNot);
	}

	public final boolean isDisconnectedFromAladin(){
		return disconnectedFromAladin;
	}

	public final void disconnectFromAladin(){
		disconnectedFromAladin = true;
		freeze(true);
		images.clear();
		nbObjs = -1;
		Collection<PhotometrySource> sources = drawings.values();
		for(PhotometrySource src : sources){
			src.disconnectFromAladin();
			if (src.hasSED())
				plugin.addSEDSource(src);
		}
		drawings.clear();
	}

	@Override
	public boolean update(boolean force) {
		images = aladin.getImages(data);
		return super.update(force);
	}

	@Override
	protected boolean primaryUpdate(final boolean force) {
		if (isDisconnectedFromAladin())
			return false;

		if (!force && !isModified())
			return true;

		try{
			final int objCnt = data.getNbObj();
			Obj[] objs = data.seeObj();
			List<Obj> toRemove = new ArrayList<Obj>(drawings.keySet());

			// Add/Update Obj of the plane into this drawing:
			for(int j=0; j<objCnt; j++){
				if (objs[j].hasSurface()){
					PhotometrySource sedSource = null;
					// CIRCLE:
					if (isCircle(objs[j].getObjType())){
						if (!drawings.containsKey(objs[j]))
							sedSource = new Circle(objs[j], this, plugin);

					}// POLYGON
					else if (isPolygon(objs[j].getObjType())){
						if (!drawings.containsKey(objs[j]))
							sedSource = new Polygon(objs[j], this, plugin);
					}
					if (sedSource != null){
						drawings.put(objs[j], sedSource);
						vSources.add(sedSource);
					}
					toRemove.remove(objs[j]);
				}
				nbObjs++;
			}

			// Remove drawings:
			for(Obj o : toRemove){
				PhotometrySource src = drawings.remove(o);
				if (src != null){
					src.disconnectFromAladin();
					plugin.addSEDSource(src);
				}

				vSources.remove(src);
			}

		}catch(AladinException ae){
			ErrorsLog.getCurrentInstance().addException(new SEDException(ae, "Drawing.primaryUpdate(boolean)"));
			return false;
		}

		return true;
	}

	protected final boolean isCircle(final String objType){
		return objType != null && (objType.equals(PHOT_OBJECT) || objType.equals(CIRCLE_OBJECT) || objType.equals(CIRCLE_OBJECTBis));
	}

	protected final boolean isPolygon(final String objType){
		return objType != null && objType.equals(POLYGON_END);
	}

	public String getName() {
		return data.getLabel();
	}

	public String getType() {
		return DRAWINGS_TYPE;
	}

	public void highlight() {
		;
	}

	protected final static class CorrectImagesIterator implements Iterator<AladinImage> {

		private final Iterator<AladinImage> it;

		private AladinImage next = null;

		public CorrectImagesIterator(final Iterable<AladinImage> images){
			it = images.iterator();
			searchNext();
		}

		private void searchNext(){
			next = null;
			while(it.hasNext() && next == null){
				next = it.next();
				if (!next.hasCharacterization() || !next.getCharacterization().isCorrect())
					next = null;
			}
		}

		public boolean hasNext() {
			return (next != null);
		}

		public AladinImage next() {
			if (next == null)
				throw new NoSuchElementException();
			else{
				AladinImage current = next;
				searchNext();
				return current;
			}
		}

		public void remove() {
			it.remove();
		}

	}

	/* **************** */
	/* EXPORT FUNCTIONS */
	/* **************** */
	public final String toASCIIFormat(String sep) { return null; }
	public final String toVOTableFormat(String linePrefix) { return null; }
	public final String toXMLFormat(String linePrefix) { return null; }

}
