package sedplugin.sed.source.aladin;

import java.awt.Color;
import java.util.Iterator;

import cds.aladin.AladinException;
import cds.aladin.Obj;

import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.plugin.AladinImage;
import sedplugin.plugin.SEDPlugin;
import sedplugin.sed.SED;
import sedplugin.sed.SEDPoint;
import sedplugin.sed.source.AbstractSEDSource;
import sedplugin.sed.source.SEDLoader;

public abstract class PhotometrySource extends AbstractSEDSource {

	private final SEDPlugin plugin;
	protected Drawing drawing;
	protected Obj object;

	protected String name = null;
	protected Color color = null;

	public PhotometrySource(final Obj object, final Drawing drawing, final SEDPlugin plugin){
		super(drawing);
		this.plugin = plugin;
		this.object = object;
		this.drawing = drawing;
	}

	public final boolean isDisconnectedFromAladin(){
		return object == null;
	}

	public final void disconnectFromAladin(){
		name = drawing.getName();
		color = object.getColor();
		drawing = null;
		object = null;
		freeze(true);
	}

	@Override
	public void freeze(boolean freezeOrNot){
		if (isDisconnectedFromAladin())
			frozen = true;
		else
			frozen = freezeOrNot;
	}

	public final Color getColor(){
		return isDisconnectedFromAladin() ? color : object.getColor();
	}

	public final void setColor(final Color color){
		if (isDisconnectedFromAladin())
			this.color = color;
		else
			object.setColor(color);
	}

	@Override
	public void removeSED() {
		super.removeSED();
		if (isDisconnectedFromAladin() && sed == null)
			plugin.removeSEDSource(this);
	}

	public String getName() {
		return isDisconnectedFromAladin() ? name : drawing.getName();
	}

	public void highlight() {
		if (!isDisconnectedFromAladin()){
			//drawing.setSelected(true);
			object.setSelected(true);
		}
	}

	@Override
	protected SED createEmptySED() {
		return new SED(drawing.getName(), plugin.getXAxis(), plugin.getYAxis(), this);
	}

	@Override
	protected SEDLoader createLoader() {
		return new PhotometryLoader();
	}

	@Override
	public boolean update(boolean force) {
		if (isDisconnectedFromAladin())
			return false;

		return super.update(force);
	}

	protected class PhotometryLoader extends SEDLoader {
		protected final int imgCount = drawing.getNbCorrectImages();

		@Override
		protected SED doInBackground() throws Exception {
			Iterator<AladinImage> it = drawing.getCorrectImages();
			AladinImage image;
			double[] stats;
			int nbImg = 0;
			setProgress(0);
			while(it.hasNext()){
				if (isCancelled()) return null;
				image = it.next();
				try{
					if (isCancelled()) return null;
					/* "JAVADOC": getStatistics(AladinData): double[] =
					 *   - le nombre de pixels [0]
					 *   - la somme [1]
					 *   - le sigma (racine carrée de la variance) [2]
					 *   - la surface total (en deg carrés) [3]
					 */
					image.getPlane().getPixel(0, 0);	// TODO Hack to avoid a "pretty" collection of error messages !
					stats = object.getStatistics(image.getPlane());
					if (isCancelled()) return null;
					if (stats[0] == 0)
						;//System.out.println("INFO: no selected pixels for \""+image.getName()+"\" !");
					else if (Double.isNaN(stats[0]) || Double.isInfinite(stats[0]) || stats[0] < 0)
						ErrorsLog.getCurrentInstance().addRefreshException(new SEDException("Incorrect number of pixels", "Computing statistics about \""+image.getName()+"\"", "AladinData.getStatistics(...) error: incorrect number of pixels (="+stats[0]+") !"));
					else if (Double.isNaN(stats[1]) || Double.isInfinite(stats[1]) || stats[1] < 0)
						ErrorsLog.getCurrentInstance().addRefreshException(new SEDException("Incorrect pixels sum", "Computing statistics about \""+image.getName()+"\"", "AladinData.getStatistics(...) error: incorrect pixels sum (="+stats[1]+") !"));
					else if (Double.isNaN(stats[3]) || Double.isInfinite(stats[3]) || stats[3] <= 0)
						ErrorsLog.getCurrentInstance().addRefreshException(new SEDException("Incorrect surface", "Computing statistics about \""+image.getName()+"\"", "AladinData.getStatistics(...) error: incorrect surface (="+stats[3]+") !"));
					else{
						if (image != null && image.hasCharacterization() && image.getCharacterization().isCorrect())
							sed.add(new SEDPoint(image, stats[1]/stats[0], stats[3]));
					}
				}catch(AladinException ae){
					System.out.println("WARNING: "+image.getName()+" => No pixel value !");	// TODO Hack to avoid a "pretty" collection of error messages !
				}catch(Exception ex){
					ErrorsLog.getCurrentInstance().addRefreshException(new SEDException(ex, "Computing statistics about \""+image.getName()+"\""));
				}
				if (isCancelled()) return null;
				nbImg++;
				setProgress((int)(((double)nbImg/imgCount)*100));
			}

			if (sed != null && sed.isEmpty()){
				boolean oldFrozen = frozen;
				removeSED();
				frozen = oldFrozen;
			}else
				sed.notifySEDListeners();

			return sed;
		}

	}

}
