package sedplugin.sed.source.operation;

import sedplugin.errors.SEDException;

import sedplugin.sed.PointComparator;
import sedplugin.sed.SED;
import sedplugin.sed.SEDPoint;

import sedplugin.sed.unit.UnitDef;
import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;

public class FunctionToken extends OperationToken {

	public static enum OperationType {
		ADD, SUB, MULT, DIV, POW, LEFT_PARENTHESIS;

		public static OperationType parse(char op){
			switch(op){
			case '+': return ADD;
			case '-': return SUB;
			case '*': return MULT;
			case '/': return DIV;
			case '^': return POW;
			case '(': return LEFT_PARENTHESIS;
			default : return null;
			}
		}

		public boolean hasPriority(){
			return (this == MULT || this == DIV || this == POW);
		}

		@Override
		public String toString(){
			switch(this){
			case ADD: return "+";
			case SUB: return "-";
			case MULT: return "*";
			case DIV: return "/";
			case POW: return "^";
			case LEFT_PARENTHESIS: return "(";
			default: return null;
			}
		}
	}

	public final OperationType operator;
	public final Operation source;

	public FunctionToken(char op, int begin, int end, Operation source) throws SEDException {
		super(op+"", TokenType.FUNCTION, begin, end);
		operator = OperationType.parse(op);
		this.source = source;
		if (operator == null)
			throw new SEDException("Unknown operator", "Parsing an operation", "\""+op+"\" is not an allowed operator (+,-,*,/,^). Position: between col."+begin+" and col."+end);
	}

	public static final boolean isFunction(char op){
		return (OperationType.parse(op) != null);
	}

	public final boolean hasPriority(){
		return operator.hasPriority();
	}

	@Override
	public int getNbPoints(){
		return 0;
	}

	public OperationToken compute(final OperationToken leftToken, final OperationToken rightToken) throws SEDException {
		if (operator == OperationType.LEFT_PARENTHESIS)
			throw new SEDException("Incorrect operation", "Computing an operation", "The operator "+operator+" is not a computing function !");
		else if (leftToken == null)
			throw new SEDException("Missing operand", "Computing an operation", "The left operand of "+operator+" is missing !");
		else if (rightToken == null)
			throw new SEDException("Missing operand", "Computing an operation", "The right operand of "+operator+" is missing !");


		// Apply the operation on 2 constants:
		if (leftToken.type == TokenType.CONSTANT && rightToken.type == TokenType.CONSTANT){
			double result = compute(((ConstantToken)leftToken).value, ((ConstantToken)rightToken).value);
			return new ConstantToken(result+"", leftToken.beginPosition, rightToken.endPosition+1);

		}// Apply the operation on at least 1 SED:
		else{
			XAxis xUnit = new XAxis();
			YAxis yUnit = new YAxis(xUnit);

			// Get 2 SEDs (if needed convert a constant into a SED):
			SED leftSED, rightSED;
			if (leftToken.type == TokenType.CONSTANT){
				rightSED = ((SEDToken)rightToken).sed;
				leftSED = ((ConstantToken)leftToken).toSED(rightSED);
			}else if (rightToken.type == TokenType.CONSTANT){
				leftSED = ((SEDToken)leftToken).sed;
				rightSED = ((ConstantToken)rightToken).toSED(leftSED);
			}else{
				leftSED = ((SEDToken)leftToken).sed;
				rightSED = ((SEDToken)rightToken).sed;
			}

			// Check there respective size:
			if (leftSED.size() != rightSED.size())
				throw new SEDException("Incorrect operation", "Computing an operation", "Impossible to apply the operation "+operator+" on the 2 SEDs because they do not have the same number of points.");

			// Build an empty SED:
			String destUnit = compute(leftSED.getYAxis().getUnit().symbol);
			XAxis sedXAxis = leftSED.getXAxis();
			YAxis sedYAxis = leftSED.getYAxis().unitExists(destUnit) ? leftSED.getYAxis() : new YAxis(sedXAxis, destUnit);
			SED result = new SED(leftToken.expression+operator+rightToken.expression, sedXAxis, sedYAxis, source);

			// Fill it:
			SEDPoint leftPoint, rightPoint;
			double leftX, rightX, leftY, rightY;
			for(int i=0; i<leftSED.size(); i++){
				leftPoint = leftSED.get(i);
				rightPoint = rightSED.get(i);

				leftX = leftPoint.getOriginalX();
				if (xUnit.unitExists(leftPoint.getOriginalXUnit()))
					xUnit.setUnit(leftPoint.getOriginalXUnit());
				else
					xUnit.setCustomUnit(new UnitDef(leftPoint.getOriginalXUnit()));

				rightX = xUnit.convert(rightPoint);

				if (PointComparator.compare(leftX, rightX) != 0)
					throw new SEDException("Incompatible wavelength", "Computing an operation", "The "+i+"-th points of the 2 SEDs have not the same wavelength: "+leftPoint.getOriginalX()+leftPoint.getOriginalXUnit()+" VS "+rightPoint.getOriginalX()+rightPoint.getOriginalXUnit());

				leftY = leftPoint.getOriginalY();
				if (yUnit.unitExists(leftPoint.getOriginalYUnit()))
					yUnit.setUnit(leftPoint.getOriginalYUnit());
				else
					yUnit.setCustomUnit(new UnitDef(leftPoint.getOriginalYUnit()));

				rightY = yUnit.convert(rightPoint);

				result.add(new SEDPoint(leftX, leftPoint.getOriginalXUnit(), compute(leftY, rightY), destUnit));
			}

			return new SEDToken(leftToken.expression+operator+rightToken.expression, result, leftToken.beginPosition, rightToken.endPosition+1);

		}

	}

	protected String compute(final String unit){
		switch(operator){
		case ADD:
		case SUB:
		case POW:
			return unit;
		case MULT:
			return "("+unit+")*("+unit+")";
		case DIV:
			return "("+unit+")/("+unit+")";
		default:
			return unit;
		}
	}

	protected double compute(final double leftValue, final double rightValue) {
		switch(operator){
		case ADD:
			return leftValue + rightValue;
		case SUB:
			return leftValue - rightValue;
		case MULT:
			return leftValue * rightValue;
		case DIV:
			return leftValue / rightValue;
		case POW:
			return Math.pow(leftValue, rightValue);
		default:
			return Double.NaN;
		}
	}

}
