package sedplugin.sed.source.operation;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.Stack;




import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.plugin.SEDPlugin;
import sedplugin.sed.SED;
import sedplugin.sed.SEDListener;
import sedplugin.sed.SEDPoint;
import sedplugin.sed.source.AbstractSEDSource;
import sedplugin.sed.source.SEDLoader;
import sedplugin.sed.source.SEDSource;
import sedplugin.sed.source.operation.OperationToken.TokenType;
import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;

public class Operation extends AbstractSEDSource implements SEDListener {

	public final static String OPERATION_TYPE = "OPERATION";

	private final SEDPlugin plugin;
	protected OperationToken result = null;

	protected final String expression;
	protected final ArrayList<OperationToken> tokens;

	private XAxis xAxis = null;
	private YAxis yAxis = null;


	public Operation(final String expr, final SEDPlugin plugin) throws SEDException {
		expression = expr;
		SED[] seds;
		if (plugin != null){
			seds = new SED[plugin.getNbSEDs()];
			Iterator<SED> it = plugin.getSEDs();
			int i=0;
			while(it.hasNext())
				seds[i++] = it.next();
		}else
			seds = new SED[0];
		OperationParser parser = new OperationParser(seds);
		this.tokens = parser.parse(expr, this);
		watchSEDs();
		this.plugin = plugin;
		init();
	}

	protected final void init() throws SEDException {
		SED newSED = null;

		Stack<OperationToken> results = new Stack<OperationToken>();

		for(OperationToken token : tokens){
			if (token.type == TokenType.FUNCTION){
				FunctionToken fct = (FunctionToken)token;
				results.push(fct.compute(results.pop(), results.pop()));
			}else
				results.push(token);
		}

		result = results.pop();
		newSED = (result.type == TokenType.SED) ? ((SEDToken)result).sed : null;
		xAxis = newSED.getXAxis();
		yAxis = newSED.getYAxis();

		sed = createEmptySED();

		if (newSED != null && !newSED.isEmpty()){
			for(SEDPoint pt : newSED)
				sed.add(new SEDPoint(pt));
		}
	}

	protected void watchSEDs(){
		for(OperationToken token : tokens){
			if (token.type == TokenType.SED)
				((SEDToken)token).sed.addSEDListener(this);
		}
	}

	public void unwatchSEDs(){
		for(OperationToken token : tokens){
			if (token.type == TokenType.SED)
				((SEDToken)token).sed.removeSEDListener(this);
		}
	}

	public String getName() {
		return expression;
	}

	public String getType() {
		return OPERATION_TYPE;
	}

	/**
	 * Gets the result of the operation. This token can be of 2 types: {@link SEDToken} or {@link ConstantToken}.
	 * 
	 * @return	The resulting token.
	 * 
	 * @see OperationToken#type
	 * @see SEDToken
	 * @see ConstantToken
	 */
	public OperationToken getResult(){
		return result;
	}

	public void sedUpdated(SED sed, SEDSource src) {
		if (!isFrozen())
			update(true);
	}

	@Override
	public boolean isModified() { return true; }

	@Override
	protected SED createEmptySED() { return new SED(expression, (xAxis==null)?plugin.getXAxis():xAxis, (yAxis==null)?plugin.getYAxis():yAxis, this); }

	@Override
	protected SEDLoader createLoader() { return new OperationLoader(); }

	@Override
	protected void updateModifiedFields() { ; }

	public void highlight() { ; }

	@Override
	public String toString(){
		return expression;
	}

	protected class OperationLoader extends SEDLoader {

		@Override
		protected SED doInBackground() throws Exception {
			SED newSED = null;

			try{
				if (isCancelled()) return null;
				Stack<OperationToken> results = new Stack<OperationToken>();
				int nbTokens = tokens.size();
				int count = 0;

				setProgress(0);
				for(OperationToken token : tokens){
					if (isCancelled()) return null;
					if (token.type == TokenType.FUNCTION){
						FunctionToken fct = (FunctionToken)token;
						results.push(fct.compute(results.pop(), results.pop()));
					}else
						results.push(token);
					count++;
					setProgress((int)(((double)count/nbTokens)*100));
				}
				if (isCancelled()) return null;
				setProgress(100);

				result = results.pop();
				newSED = (result.type == TokenType.SED) ? ((SEDToken)result).sed : null;
			}catch(SEDException se){
				ErrorsLog.getCurrentInstance().addException(se);
				newSED = null;
			}

			if (newSED != null && !newSED.isEmpty()){
				for(SEDPoint pt : newSED)
					sed.add(new SEDPoint(pt));
			}

			sed.notifySEDListeners();
			return sed;
		}

	}

	/* **************** */
	/* EXPORT FUNCTIONS */
	/* **************** */
	public final String toASCIIFormat(String sep) {
		return null;
	}

	public String toVOTableFormat(String linePrefix) {
		return null;
	}

	public String toXMLFormat(String linePrefix) {
		return null;
	}
}
