package sedplugin.sed.source.operation;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Stack;


import sedplugin.errors.SEDException;
import sedplugin.sed.SED;
import sedplugin.sed.source.operation.FunctionToken.OperationType;
import sedplugin.sed.unit.XAxis;

public class OperationParser {
	/** Type of an opened parenthesis token. */
	public static final int ParOpen = 0;

	/** Type of a closed parenthesis token. */
	public static final int ParClose = 1;

	/** Type of a spectrum token. */
	public static final int Spectrum_Token = 2;

	/** Type of a constant (numeric or ref) token. */
	public static final int Constant_Token = 3;

	/** Type of the operator + token. */
	public static final int ADD = 4;

	/** Type of the operator - token. */
	public static final int SUB = 5;

	/** Type of the operator * token. */
	public static final int MULT = 6;

	/** Type of the operator / token. */
	public static final int DIV = 7;

	/** Type of the operator ^ token. */
	public static final int POW = 8;

	/** Character which must prefix a spectrum reference */
	public static final char SpectrumPrefix = '$';

	/** Character which must prefix a predefined constant reference. (ex: "@PI" and "@SPEED_OF_LIGHT") */
	public static final char PredefinedConstantPrefix = '@';

	public static final String[] PredefinedConstantsString = new String[]{"PI", "SPEED_OF_LIGHT"};
	public static final double[] PredefinedConstants = new double[]{Math.PI, XAxis.SPEED_OF_LIGHT};

	/** List of all implied spectrums. */
	protected SED[] lstSEDs = new SED[0];

	/** List of all predefined constants. */
	protected HashMap<String, Double> mapConstants = null;

	protected Stack<FunctionToken> operators;

	protected Stack<ArrayList<OperationToken>> operands;

	protected static class LeftParenthesisToken extends OperationToken {

		public LeftParenthesisToken() {
			super("(", null, -1, -1);
		}

		@Override
		public int getNbPoints() {
			return 0;
		}

	}

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds a parser of SED operations.
	 * 
	 * @param seds	List of all available SEDs.
	 */
	public OperationParser(SED[] seds){
		lstSEDs = seds;
		operators = new Stack<FunctionToken>();
		operands = new Stack<ArrayList<OperationToken>>();
	}

	/**
	 * Builds a parser of SED operations.
	 * 
	 * @param seds List of all available SEDs.
	 */
	public OperationParser(Collection<SED> seds) {
		lstSEDs = new SED[seds.size()];
		int i=0;
		for(SED s : seds)
			lstSEDs[i++] = s;

		operators = new Stack<FunctionToken>();
		operands = new Stack<ArrayList<OperationToken>>();
	}

	/* **************** */
	/* TOKENIZE METHODS */
	/* **************** */
	/**
	 * Lets split the given expression in tokens.<BR />
	 * <i><u>Note</u>: The spectrums list is the last one used by a tokenizeExpr methods. It can be obsolete.</i>
	 * 
	 * @param expr				The string expression to tokenize.
	 * @return					The corresponding tokens list.
	 * @throws SEDException	If there is an error in a token identification.
	 */
	public ArrayList<OperationToken> parse(String expr, Operation source) throws SEDException {
		if (mapConstants == null)
			refreshConstantsList();

		operators.clear();
		operands.clear();

		char[] exprChar = expr.toCharArray();
		StringBuffer capture = new StringBuffer();
		int indCapture = 0;
		FunctionToken operator;

		for(int i=0; i<exprChar.length; i++){
			if (exprChar[i] == ' ' || exprChar[i] == '\t' || exprChar[i] == '\n')
				continue;

			if (capture.length() == 0) indCapture = i;

			if (FunctionToken.isFunction(exprChar[i])){
				addOperand(capture, indCapture);
				operator = new FunctionToken(exprChar[i], i, i, source);
				if (!operators.isEmpty() && operator.operator != OperationType.LEFT_PARENTHESIS && !operator.hasPriority() && operators.peek().operator != OperationType.LEFT_PARENTHESIS && operators.peek().hasPriority())
					unstack();
				operators.push(operator);
			}else if (exprChar[i] == ')'){
				addOperand(capture, indCapture);
				unstackAll();
				if (!operators.isEmpty() && operators.peek().operator != OperationType.LEFT_PARENTHESIS && operators.peek().hasPriority())
					unstack();
			}else
				capture.append(exprChar[i]);
		}
		addOperand(capture, indCapture);
		unstackAll();

		return operands.pop();
	}

	/**
	 * Try to make the association between the last captured string with a spectrum or constant reference.
	 * 
	 * @param capture			The last captured characters concatenated in a single string.
	 * @param indCapture		The last read position in the whole expression to tokenize.
	 * 
	 * @throws SEDException
	 */
	protected void addOperand(StringBuffer capture, int indCapture) throws SEDException {
		// If the captured string is a reference it must begin by a prefix:
		if (capture != null && capture.length() > 0){
			ArrayList<OperationToken> tokens = new ArrayList<OperationToken>(1);
			int endCapture = indCapture+capture.length();

			// CONSTANT CASE:
			if (capture.charAt(0) == PredefinedConstantPrefix){
				String cstName = capture.substring(1);
				Double value = (Double)mapConstants.get(cstName);
				if (value == null)
					throw new SEDException("Invalid token", "Identification of \""+cstName+"\" at ["+indCapture+";"+(indCapture+capture.length())+"]", "The constant \""+cstName+"\" doesn't exist !");
				else
					tokens.add(new ConstantToken(value, indCapture, endCapture));

			}// SPECTRUM CASE:
			else if (capture.charAt(0) == SpectrumPrefix){
				String sedID = capture.substring(1).trim();
				try{
					int sedIndex = Integer.parseInt(sedID)-1;
					if (sedIndex < 0 || sedIndex >= lstSEDs.length)
						throw new SEDException("Invalid token", "Identification of \""+capture+"\" at ["+indCapture+";"+(indCapture+capture.length()+2)+"]", "There is no "+sedIndex+"-th SED !");
					else
						tokens.add(new SEDToken(capture.toString(), lstSEDs[sedIndex], indCapture, endCapture));
				}catch(NumberFormatException nfe){
					throw new SEDException("Invalid SED index", "Identification of \""+capture+"\" at ["+indCapture+";"+endCapture+"]", sedID+" is not a SED index !");
				}

			}// If it's not a reference it MUST be a numeric value:
			else{
				try{
					double value = Double.parseDouble(capture.toString());
					tokens.add(new ConstantToken(value, indCapture, endCapture));
				}catch(NumberFormatException nfe){
					throw new SEDException("Invalid token", "Identification of \""+capture+"\" at ["+indCapture+";"+(indCapture+capture.length())+"]", "A constant must be a numeric value !");
				}
			}

			operands.push(tokens);
			capture.delete(0, capture.length());
		}
	}

	private void unstack() throws SEDException {
		ArrayList<OperationToken> tokens = new ArrayList<OperationToken>();
		tokens.addAll(operands.pop());
		tokens.addAll(operands.pop());
		tokens.add(operators.pop());

		operands.push(tokens);

		if (operands.isEmpty() && !operators.isEmpty())
			throw new SEDException("Too much operators", "Parsing an operation", "Too much operators !");
	}

	private void unstackAll() throws SEDException {
		ArrayList<OperationToken> tokens;

		FunctionToken operator;
		while(!operators.isEmpty()){
			operator = operators.pop();
			if (operator.operator == OperationType.LEFT_PARENTHESIS)
				return;
			else{
				tokens = new ArrayList<OperationToken>();
				tokens.addAll(operands.pop());
				tokens.addAll(operands.pop());
				tokens.add(operator);
				operands.push(tokens);
			}
		}

		if (operands.isEmpty() && !operators.isEmpty())
			throw new SEDException("Too much operators", "Parsing an operation", "Too much operators !");
	}

	protected void refreshConstantsList(){
		mapConstants = new HashMap<String, Double>();
		if (PredefinedConstantsString.length == PredefinedConstants.length){
			for(int i=0; i<PredefinedConstantsString.length; i++)
				mapConstants.put(PredefinedConstantsString[i], new Double(PredefinedConstants[i]));
		}
	}
}
