package sedplugin.sed.unit;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Iterator;
import java.util.Vector;

import sedplugin.errors.SEDException;
import sedplugin.errors.unit.UnknownUnitException;

import sedplugin.sed.SED;

/**
 * <p>
 * 	Representation of a physical quantity. A physical quantity gathers several units which measure the same quantity.
 * 	Conversion between these units are done only by multiplying by a given factor.
 * </p>
 * <p>
 * 	The first unit is always considered as the reference unit.
 * 	All conversion factors must always be computed in function of this reference unit.
 * 	If the first unit is removed, all conversion factors are immediately re-computed (see {@link #updateMatrix()} for more details).
 * </p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 03/2011
 * 
 * @see UnitDef
 * @see AxisUnit
 */
public class PhysicalQuantity implements Iterable<UnitDef> {

	/**
	 * Association between a unit and its conversion factor inside a physical quantity.
	 * 
	 * @author Gr&eacute;gory Mantelet (CDS)
	 * @version 03/2011
	 */
	public static final class UnitEntry {

		protected UnitDef unit;

		protected double convFactor;

		/**
		 * Builds a unit entry.
		 * 
		 * @param unit			A unit.
		 * @param convFactor	Its conversion factor to the reference unit of a physical quantity.
		 * 
		 * @throws SEDException	If the given unit is <i>null</i> or if the conversion factory is 0 or Infinite or NaN.
		 */
		public UnitEntry(UnitDef unit, double convFactor) throws SEDException {
			if (unit == null || convFactor == 0 || Double.isInfinite(convFactor) || Double.isNaN(convFactor))
				throw new SEDException("Undefined unit", "Creating a unit entry", "Impossible to create a unit entry without a valid unit (!= null) and a valid conversion factor (!= 0, NaN and +/-Infinity) !");

			this.unit = unit;
			this.convFactor = convFactor;
		}

		public final UnitDef unit(){ return unit; }
		public final String unitSymbol(){ return unit.symbol; }
		public final double convFactor(){ return convFactor; }
		public final boolean convFactor(double newFactor){
			if (newFactor == 0 || Double.isInfinite(newFactor) || Double.isNaN(newFactor))
				return false;
			else{
				convFactor = newFactor;
				return true;
			}
		}
		@Override
		public boolean equals(Object obj) {
			return (obj instanceof UnitEntry && ((UnitEntry)obj).unit.equals(unit));
		}
		@Override
		public String toString(){
			return "{ "+unit.toString()+" -> "+convFactor+" }";
		}
	}

	/**
	 * <p>
	 * 	Iterator of unit on a physical quantity.
	 * 	It lets iterating on all unit entries contained in a physical quantity.
	 * 	It returns only the UnitDef object of each UnitEntry.
	 * </p>
	 * <p>
	 * 	The {@link UnitIterator#remove()} function removes the whole current UnitEntry.
	 * </p>
	 * 
	 * @author Gr&eacute;gory Mantelet (CDS)
	 * @version 03/2011
	 */
	public static final class UnitIterator implements Iterator<UnitDef> {

		private final Iterator<UnitEntry> itUnitEntry;

		public UnitIterator(PhysicalQuantity quantity){
			itUnitEntry = quantity.units.iterator();
		}

		public boolean hasNext() {
			return itUnitEntry.hasNext();
		}

		public UnitDef next() {
			return itUnitEntry.next().unit;
		}

		public void remove() {
			itUnitEntry.remove();
		}

	}

	/* ********** */
	/* ATTRIBUTES */
	/* ********** */
	/** Name of the physical quantity. It is used to distinguish two physical quantities. */
	private final String name;

	/** The quantity symbol. */
	private String symbol;

	/** The unicode version of the quantity symbol. */
	private String unicodeSymbol;

	/** The HTML version of the quantity symbol. */
	private String htmlSymbol;

	/** List of all managed units. Each item is a {@link UnitEntry} object which associates a {@link UnitDef} to a conversion factor (to the reference unit of the quantity). */
	protected final ArrayList<UnitEntry> units;

	/**
	 * <p>Matrix (2D) which lets convert a unit to another one.</p>
	 * <p>There are as many lines as columns. The number of line is equal to the number of units.</p>
	 * <p>To convert x from the i-th unit to the j-th unit: x*conversionMatrix[j][i].</p> */
	protected double[][] conversionMatrix;

	/** The index of the current unit. (By default, the reference unit: 0) */
	protected int currentUnit = 0;

	/** The index which indicates there is no current unit (generally used when there is no more unit in this physical quantity). */
	public static final int NONE = -1;

	/** List of axis to notify when the current unit is changed. */
	protected final Vector<AxisUnit> vAxis;

	/* ************ */
	/* CONSTRUCTORS */
	/* ************ */
	/**
	 * Builds a physical quantity with its name and its reference unit.
	 * 
	 * @param quantityName	Name of this quantity.
	 * @param refUnit		Its reference unit (this unit will be the first one of this quantity).
	 * 
	 * @see #PhysicalQuantity(String, String, UnitDef)
	 * 
	 * @throws SEDException 	If the reference unit is incorrect or <i>null</i>.
	 */
	public PhysicalQuantity(String quantityName, UnitDef refUnit) throws SEDException {
		this(quantityName, "", refUnit);
	}

	/**
	 * Builds a physical quantity with its name, its symbol and its reference unit.
	 * 
	 * @param quantityName		Name of this quantity.
	 * @param quantitySymbol	Symbol of this quantity.
	 * @param refUnit			Its reference unit (this unit will be the first one of this quantity).
	 * 
	 * @see #PhysicalQuantity(String, String, String, UnitDef)
	 * 
	 * @throws SEDException 	If the reference unit is incorrect or <i>null</i>.
	 */
	public PhysicalQuantity(String quantityName, String quantitySymbol, UnitDef refUnit) throws SEDException {
		this(quantityName, quantitySymbol, quantitySymbol, refUnit);
	}

	/**
	 * Builds a physical quantity with its name, its symbol, the unicode version of its symbol and its reference unit.
	 * 
	 * @param quantityName		Name of this quantity.
	 * @param quantitySymbol	Symbol of this quantity.
	 * @param unicodeSymbol		Unicode version of the given symbol.
	 * @param refUnit			Its reference unit (this unit will be the first one of this quantity).
	 * 
	 * @throws SEDException 	If the reference unit is incorrect or <i>null</i>.
	 */
	public PhysicalQuantity(String quantityName, String quantitySymbol, String unicodeSymbol, UnitDef refUnit) throws SEDException {
		name = quantityName;
		symbol = quantitySymbol;
		this.unicodeSymbol = unicodeSymbol;
		units = new ArrayList<UnitEntry>();
		units.add(new UnitEntry(refUnit, 1));
		conversionMatrix = new double[1][1];
		conversionMatrix[0][0] = 1;
		vAxis = new Vector<AxisUnit>(2, 2);
	}

	/* ************************* */
	/* MATRIX MANAGEMENT METHODS */
	/* ************************* */
	/**
	 * If the conversion factors are (in the same order than the contained units)
	 * 	|  1  |  x  |  y  |  z  |
	 * 
	 * Then conversionMatrix =
	 * 	|  1  |  x  |  y  |  z  |
	 * 	| 1/x |  1  | y/x | z/x |
	 * 	| 1/y | x/y |  1  | z/y |
	 * 	| 1/z | x/z | y/z |  1  |
	 */
	protected void updateMatrix(){
		int nbUnits = units.size();

		conversionMatrix = new double[nbUnits][nbUnits];

		// Initialize the first line:
		for(int j=0; j<nbUnits; j++)
			conversionMatrix[0][j] = getConvFactor(j);

		// Computes the others:
		for(int i=1; i<nbUnits; i++){
			for(int j=0; j<i; j++)
				conversionMatrix[i][j] = (conversionMatrix[j][i]==0)?Double.NaN:(1/conversionMatrix[j][i]);
			conversionMatrix[i][i] = 1;
			for(int j=i+1; j<nbUnits; j++)
				conversionMatrix[i][j] = conversionMatrix[0][j]/conversionMatrix[0][i];
		}
	}

	/* ***************** */
	/* GETTERS & SETTERS */
	/* ***************** */
	/**
	 * Gets the name of this physical quantity.
	 * 
	 * @return	Its name.
	 */
	public final String getName(){
		return name;
	}

	/**
	 * Gets the index of the selected unit.
	 * 
	 * @return	The index of the selected unit.
	 */
	public final int getIndUnit() {
		return currentUnit;
	}

	/**
	 * Gets the symbol of this physical quantity.
	 * 
	 * @return	Its symbol.
	 */
	public final String getSymbol(){
		return symbol;
	}

	/**
	 * Sets the symbol of this physical quantity.
	 * 
	 * @param quantitySymbol	Its new symbol.
	 */
	public final void setSymbol(String quantitySymbol){
		symbol = quantitySymbol;
	}

	/**
	 * Gets the unicode version of the symbol of this physical quantity.
	 * 
	 * @return	The unicode version of its symbol.
	 */
	public final String getUnicodeSymbol(){
		return unicodeSymbol;
	}

	/**
	 * Sets the unicode version of the symbol of this physical quantity.
	 * 
	 * @param quantitySymbol	The new unicode version of its symbol.
	 */
	public final void setUnicodeSymbol(String quantitySymbol){
		unicodeSymbol = quantitySymbol;
	}

	/**
	 * Gets the HTML version of the symbol of this physical quantity.
	 * 
	 * @return The HTML version of its symbol.
	 */
	public final String getHtmlSymbol() {
		return htmlSymbol;
	}

	/**
	 * Sets the HTML version of the symbol of this physical quantity.
	 * 
	 * @param htmlSymbol The new HTML version of its symbol.
	 */
	public final void setHtmlSymbol(String htmlSymbol) {
		this.htmlSymbol = htmlSymbol;
	}

	/**
	 * Gets the indUnit-th units of this physical quantity.
	 * 
	 * @param indUnit	The index of the unit to return.
	 * 
	 * @return	The corresponding unit.
	 */
	public final UnitDef getUnit(int indUnit){
		return ((UnitEntry)units.get(indUnit)).unit();
	}

	/**
	 * Gets the conversion factor of the indUnit-th units of this physical quantity.
	 * 
	 * @param indUnit	The index of the unit whose the conversion factor must be returned.
	 * 
	 * @return	The corresponding conversion factor.
	 */
	public final double getConvFactor(int indUnit){
		return ((UnitEntry)units.get(indUnit)).convFactor();
	}

	/**
	 * Gets the conversion factor of the given unit of this physical quantity.
	 * 
	 * @param unit		A unit of this physical quantity.
	 * 
	 * @return			The corresponding conversion factor.
	 */
	public final double getConvFactor(final String unit){
		int index = indexOf(unit);
		if (index >= 0)
			return units.get(index).convFactor;
		else
			return Double.NaN;
	}

	/**
	 * Sets the conversion factor of the indUnit-th units of this physical quantity.
	 * 
	 * @param indUnit	The index of the unit whose the conversion factor must be returned.
	 * @param newFactor	The new conversion factor.
	 * 
	 * @return	<i>true</i> if the conversion factor has been updated, <i>false</i> otherwise.
	 */
	protected final boolean setConvFactor(int indUnit, double newFactor){
		return ((UnitEntry)units.get(indUnit)).convFactor(newFactor);
	}

	/**
	 * <p>Gets the reference unit of this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> The reference unit corresponds to the first unit.</i></p>
	 * 
	 * @return	Its reference unit.
	 */
	public final UnitDef getReferenceUnit(){
		return getUnit(0);
	}

	/**
	 * Gets the current unit of this physical quantity.
	 * 
	 * @return	Its current unit.
	 */
	public final UnitDef getCurrentUnit(){
		if (currentUnit == NONE)
			return null;
		else
			return getUnit(currentUnit);
	}

	/**
	 * <p>Sets the current unit considering the given symbol.</p>
	 * 
	 * <p><i><u>Note:</u> All associated axis are notified of the modification.</i></p>
	 * 
	 * @param unitSymbol	The symbol of the unit which must be selected as current unit.
	 * 
	 * @return	<i>true</i> if the current unit has been changed, <i>false</i> otherwise.
	 * 
	 * @see #setCurrentUnit(int)
	 */
	public final boolean setCurrentUnit(String unitSymbol){
		return setCurrentUnit(indexOf(unitSymbol));
	}

	/**
	 * <p>Sets the current unit of this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> All associated axis are notified of the modification.</i></p>
	 * 
	 * @param indUnit	The index of the unit which must be considered as the current unit.
	 * 
	 * @return	<i>true</i> if the current unit has been changed, <i>false</i> otherwise.
	 * 
	 * @see #notifySEDs()
	 */
	public final boolean setCurrentUnit(int indUnit){
		if (indUnit >= 0 && indUnit < getNbUnits()){
			currentUnit = indUnit;
			notifySEDs();
			return true;
		}
		return false;
	}

	/**
	 * Gets all contained units.
	 * 
	 * @return	An iterator of unit.
	 */
	public final Iterator<UnitDef> getUnits(){
		return new UnitIterator(this);
	}

	/**
	 * Gets all contained units.
	 * 
	 * @return	An iterator of unit.
	 * 
	 * @see #getUnits()
	 */
	public final Iterator<UnitDef> iterator(){
		return getUnits();
	}

	/**
	 * Gets the number of units contained in this physical quantity.
	 * 
	 * @return	The number of its units.
	 */
	public final int getNbUnits(){
		return units.size();
	}

	/* ******************* */
	/* UNIT SEARCH METHODS */
	/* ******************* */
	/**
	 * Gets the unit whose the symbol is equals to the given one.
	 * 
	 * @param unitSymbol	The symbol (not unicode) of the unit to get.
	 * 
	 * @return	The corresponding unit if found, otherwise <i>null</i>.
	 */
	public final UnitDef getUnit(String unitSymbol){
		if (unitSymbol == null)
			return null;

		Iterator<UnitDef> it = getUnits();
		UnitDef foundUnit = null;
		while(foundUnit == null && it.hasNext()){
			UnitDef temp = (UnitDef)it.next();
			if (temp.symbol.equals(unitSymbol))
				foundUnit = temp;
		}
		return foundUnit;
	}

	/**
	 * Gets the index of the given unit if it is inside this physical quantity.
	 * 
	 * @param unit	The unit whose the index is asked.
	 * 
	 * @return	The index of the given unit if found, {@link #NONE} otherwise.
	 */
	public final int indexOf(UnitDef unit){
		if (unit == null)
			return NONE;

		int indUnit = NONE;
		for(int i=0; indUnit == NONE && i<units.size(); i++)
			if (getUnit(i).equals(unit))
				indUnit = i;
		return indUnit;
	}

	/**
	 * Gets the index of the unit whose the symbol is given in parameter.
	 * 
	 * @param unitSymbol	The symbol of the unit whose the index is asked.
	 * 
	 * @return	The index of the unit whose the symbol is given in parameter, {@link #NONE} otherwise.
	 */
	public final int indexOf(String unitSymbol){
		if (unitSymbol == null)
			return NONE;

		int indUnit = NONE;
		for(int i=0; indUnit == NONE && i<units.size(); i++)
			if (((UnitEntry)units.get(i)).unitSymbol().equals(unitSymbol))
				indUnit = i;
		return indUnit;
	}

	/**
	 * Tells whether the given unit is inside this physical quantity.
	 * 
	 * @param unit	A unit.
	 * 
	 * @return	<i>true</i> if the given unit is inside this physical quantity, <i>false</i> otherwise.
	 * 
	 * @see #indexOf(UnitDef)
	 */
	public final boolean exists(UnitDef unit){
		return indexOf(unit) != NONE;
	}

	/**
	 * Tells whether there is a unit with the given symbol inside this physical quantity.
	 * 
	 * @param unitSymbol	A unit symbol.
	 * 
	 * @return	<i>true</i> if there is a unit with the given symbol inside this physical quantity, <i>false</i> otherwise.
	 * 
	 * @see #indexOf(String)
	 */
	public final boolean exists(String unitSymbol){
		return indexOf(unitSymbol) != NONE;
	}

	/* *********************** */
	/* UNIT MANAGEMENT METHODS */
	/* *********************** */
	/**
	 * <p>Adds the given unit (with its conversion factor to the reference unit) to this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> The conversion matrix is automatically updated if the given unit is added.</i></p>
	 * 
	 * @param newUnit			The unit to add.
	 * @param refConversion		The conversion factor from this unit to the reference unit.
	 * 
	 * @return	<i>true</i> if the given unit has been added, <i>false</i> otherwise (for instance, if the unit already exists).
	 */
	public boolean add(UnitDef newUnit, double refConversion){
		// If the given unit already exists, only its conversion factor is updated:
		int indUnit = indexOf(newUnit);
		if (indUnit != NONE){
			if (!setConvFactor(indUnit, refConversion))
				return false;

			// Otherwise, add the given entry:
		}else{
			try{
				units.add(new UnitEntry(newUnit, refConversion));
			}catch(SEDException ex){
				return false;
			}
		}

		// Update the conversion matrix:
		updateMatrix();

		return true;
	}

	/**
	 * <p>Adds all the given unit entries to this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> The conversion matrix is automatically updated if at least one unit is added.</i></p>
	 * 
	 * @param newUnits	All the units to add.
	 * 
	 * @return	The number of all added units.
	 */
	public int add(UnitEntry[] newUnits){
		if (newUnits == null || newUnits.length == 0)
			return 0;

		int nbAdded = 0;
		for(int i=0; i<newUnits.length; i++){
			// Ignore NULL entries:
			if (newUnits[i] == null)
				continue;

			// If the unit already exists, only update its conversion factor:
			int indUnit = indexOf(newUnits[i].unit());
			if (indUnit != NONE)
				setConvFactor(indUnit, newUnits[i].convFactor());

			// Otherwise, add the entry:
			else
				units.add(newUnits[i]);

			nbAdded++;
		}

		// Update the conversion matrix:
		if (nbAdded > 0)
			updateMatrix();

		return nbAdded;
	}

	/**
	 * <p>Adds all the given units to this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> The conversion matrix is automatically updated if at least one unit is added.</i></p>
	 * 
	 * @param newUnits			The units to add.
	 * @param refConversions	Their conversion factor to the reference unit.
	 * 
	 * @return	The number of added units.
	 */
	public int add(UnitDef[] newUnits, double[] refConversions){
		if (newUnits.length != refConversions.length)
			return 0;

		int nbAdded = 0;
		for(int i=0; i<newUnits.length; i++){
			UnitDef newUnit = newUnits[i];
			double refConversion = refConversions[i];

			// If the unit already exists, only update its conversion factor:
			int indUnit = indexOf(newUnit);
			if (indUnit != NONE)
				setConvFactor(indUnit, refConversion);

			// Otherwise, add the entry:
			else{
				try{
					units.add(new UnitEntry(newUnit, refConversion));
				}catch(SEDException ex){
					continue;
				}
			}

			nbAdded++;
		}

		// Update the conversion matrix:
		if (nbAdded > 0)
			updateMatrix();

		return nbAdded;
	}

	/**
	 * <p>Removes the given unit from this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> The conversion matrix is automatically updated.</i></p>
	 * 
	 * @param unitToRemove	The unit to remove.
	 * 
	 * @return	<i>true</i> if the given unit has been removed, <i>false</i> otherwise.
	 * 
	 * @see #remove(int)
	 */
	public boolean remove(UnitDef unitToRemove){
		return remove(indexOf(unitToRemove));
	}

	/**
	 * <p>Removes the unit whose the symbol is given in parameter from this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> The conversion matrix is automatically updated.</i></p>
	 * 
	 * @param unitToRemove	The symbol of the unit to remove.
	 * 
	 * @return	<i>true</i> if the corresponding unit has been removed, <i>false</i> otherwise.
	 * 
	 * @see #remove(int)
	 */
	public boolean remove(String unitToRemove){
		return remove(indexOf(unitToRemove));
	}

	/**
	 * <p>Removes the indUnit-th unit from this physical quantity.</p>
	 * 
	 * <p><i><u>Note:</u> The conversion matrix is automatically updated.</i></p>
	 * 
	 * @param indUnit	The index of the unit to remove.
	 * 
	 * @return	<i>true</i> if the specified unit has been removed, <i>false</i> otherwise (generally it the index is incorrect).
	 * 
	 * @see #updateMatrix()
	 */
	public boolean remove(int indUnit){
		// Returns if index out of bounds:
		if (indUnit < 0 || indUnit > units.size())
			return false;

		// Returns if there is only the reference unit:
		if (indUnit == 0 && units.size() == 1)
			return false;

		// Remove the specified unit:
		units.remove(indUnit);

		// Update the current unit if needed:
		if (currentUnit >= units.size())
			setCurrentUnit(currentUnit-1);

		// Update the conversion matrix:
		updateMatrix();

		return true;
	}

	/**
	 * <p>Removes the units whose the indexes are given in parameter.</p>
	 * 
	 * <p><i><u>Note:</u> The conversion matrix is automatically updated if at least one unit is removed.</i></p>
	 * 
	 * @param indUnits	The indexes of the units to remove.
	 * 
	 * @return	The number of removed units.
	 * 
	 * @see #updateMatrix()
	 */
	public int remove(int[] indUnits){
		Arrays.sort(indUnits);
		int indRemove = 0;

		// If the first unit (the reference unit) must be removed:
		while (units.size()>1 && indUnits[indRemove]-indRemove == 0){
			indRemove++;
			// remove the first unit => the second unit becomes the reference unit:
			units.remove(0);
			// update the conversion factors = they are replaced by the second line from the second column of the conversion matrix:
			for(int j=indRemove; j<conversionMatrix[indRemove].length; j++)
				setConvFactor(j-indRemove, conversionMatrix[indRemove][j]);
		}

		// Remove the next specified units:
		int i=1;
		while(i<units.size()){
			if (indRemove < indUnits.length && i == indUnits[indRemove]-indRemove){
				indRemove++;
				units.remove(i);
			}else
				i++;
		}

		// Update the current unit if needed:
		if (currentUnit >= units.size())
			setCurrentUnit(currentUnit-1);

		// Update the conversion matrix:
		if (indRemove > 0)
			updateMatrix();

		return indRemove;
	}

	/* *************************************** */
	/* CONVERSION FUNCTIONS: ? -> CURRENT_UNIT */
	/* *************************************** */
	/**
	 * Lets convert the given value to the current unit.
	 * 
	 * @param value			The value to convert.
	 * @param fromUnit		The unit of the given value.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If the source unit is unknown.
	 * 
	 * @see #convert(double, int, String)
	 */
	public double convert(double value, String fromUnit) throws UnknownUnitException {
		return convert(value, indexOf(fromUnit), fromUnit);
	}

	/**
	 * Lets convert the given value to the current unit.
	 * 
	 * @param value			The value to convert.
	 * @param indFromUnit	The unit index of the given value.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If the source unit is unknown.
	 * 
	 * @see #convert(double, int, String)
	 */
	public double convert(double value, int indFromUnit) throws UnknownUnitException {
		return convert(value, indFromUnit, null);
	}

	/**
	 * Lets convert the given value to the current unit.
	 * 
	 * @param value				The value to convert.
	 * @param fromUnit			The index of the unit of the given value.
	 * @param fromUnitSymbol	The symbol of the unit of the given value.
	 * 
	 * @return					The converted value.
	 * 
	 * @throws UnknownUnitException		If the source unit is unknown.
	 */
	protected double convert(double value, int fromUnit, String fromUnitSymbol) throws UnknownUnitException {
		// Check the source unit index:
		if (fromUnit < 0 || fromUnit >= units.size()){
			if (fromUnitSymbol == null)
				throw new UnknownUnitException(fromUnit, "Conversion into the physical quantity\""+getName()+"\"");
			else
				throw new UnknownUnitException(fromUnitSymbol, "Conversion into the physical quantity\""+getName()+"\"");
		}

		// If all is right, make the conversion:
		double result = value*conversionMatrix[currentUnit][fromUnit];
		// TODO DEBUG Trace !
		if (SED.trace) System.out.println("** Conversion to CurrentUnit: "+value+getUnit(fromUnit).symbol+"["+fromUnit+"] -> "+result+getUnit(currentUnit).symbol+"["+currentUnit+"]");
		return result;
	}

	/* **************************** */
	/* CONVERSION FUNCTIONS: ? -> ? */
	/* **************************** */
	/**
	 * Lets convert the give unit in the given unit.
	 * 
	 * @param value			The value to convert.
	 * @param fromUnit		The unit of the given value.
	 * @param toUnit		The destination unit.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If one or both given units are unknown.
	 * 
	 * @see #convert(double, int, String, int, String)
	 */
	public double convert(double value, String fromUnit, String toUnit) throws UnknownUnitException {
		return convert(value, indexOf(fromUnit), fromUnit, indexOf(toUnit), toUnit);
	}

	/**
	 * Lets convert the give unit in the given unit.
	 * 
	 * @param value			The value to convert.
	 * @param indFromUnit	The index of the unit of the given value.
	 * @param indToUnit		The index of the destination unit.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If one or both given units are unknown.
	 * 
	 * @see #convert(double, int, String, int, String)
	 */
	public double convert(double value, int indFromUnit, int indToUnit) throws UnknownUnitException {
		return convert(value, indFromUnit, null, indToUnit, null);
	}

	/**
	 * Lets convert the give unit in the given unit.
	 * 
	 * @param value				The value to convert.
	 * @param fromUnit			The index of the unit of the given value.
	 * @param fromUnitSymbol	The symbol of the source unit.
	 * @param toUnit			The destination unit.
	 * @param toUnitSymbol		The symbol of the destination unit.
	 * 
	 * @return					The converted value.
	 * 
	 * @throws UnknownUnitException		If one or both given units are unknown.
	 */
	protected double convert(double value, int fromUnit, String fromUnitSymbol, int toUnit, String toUnitSymbol) throws UnknownUnitException {
		// Check the source unit index:
		if (fromUnit < 0 || fromUnit >= units.size()){
			if (fromUnitSymbol == null)
				throw new UnknownUnitException(fromUnit, "Conversion into the physical quantity \""+getName()+"\"");
			else
				throw new UnknownUnitException(fromUnitSymbol, "Conversion into the physical quantity \""+getName()+"\"");
		}

		// Check the destination unit index:
		if (toUnit < 0 || toUnit >= units.size()){
			if (toUnitSymbol == null)
				throw new UnknownUnitException(toUnit, "Conversion into the physical quantity \""+getName()+"\"");
			else
				throw new UnknownUnitException(toUnitSymbol, "Conversion into the physical quantity \""+getName()+"\"");
		}

		// If all is right, make the conversion:
		double result = value*conversionMatrix[toUnit][fromUnit];
		// TODO DEBUG Trace !
		if (SED.trace) System.out.println("** Conversion: "+value+getUnit(fromUnit).symbol+"["+fromUnit+"] -> "+result+getUnit(toUnit).symbol+"["+toUnit+"]");
		return result;
	}

	/* *********************************** */
	/* CONVERSION FUNCTIONS: ? -> REF_UNIT */
	/* *********************************** */
	/**
	 * Lets convert the given value to the reference unit.
	 * 
	 * @param value		The value to convert
	 * @param fromUnit	The unit of the given value.
	 * 
	 * @return			The converted value.
	 * 
	 * @throws UnknownUnitException	If the source unit is unknown.
	 * 
	 * @see #convertToRef(double, int, String)
	 */
	public double convertToRef(double value, String fromUnit) throws UnknownUnitException {
		return convertToRef(value, indexOf(fromUnit), fromUnit);
	}

	/**
	 * Lets convert the given value to the reference unit.
	 * 
	 * @param value			The value to convert
	 * @param indFromUnit	The unit of the given value.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If the source unit is unknown.
	 * 
	 * @see #convertToRef(double, int, String)
	 */
	public double convertToRef(double value, int indFromUnit) throws UnknownUnitException {
		return convertToRef(value, indFromUnit, null);
	}

	/**
	 * Lets convert the given value to the reference unit.
	 * 
	 * @param value				The value to convert
	 * @param fromUnit			The index of the source unit.
	 * @param fromUnitSymbol	The symbol of the source unit.
	 * 
	 * @return					The converted value.
	 * 
	 * @throws UnknownUnitException		If the source unit is unknown.
	 */
	protected double convertToRef(double value, int fromUnit, String fromUnitSymbol) throws UnknownUnitException {
		// Check the source unit index:
		if (fromUnit == NONE){
			if (fromUnitSymbol == null)
				throw new UnknownUnitException(fromUnit, "Conversion into the physical quantity \""+getName()+"\"");
			else
				throw new UnknownUnitException(fromUnitSymbol, "Conversion into the physical quantity \""+getName()+"\"");
		}

		// If all is right, make the conversion:
		double result = value*conversionMatrix[0][fromUnit];
		// TODO DEBUG Trace !
		if (SED.trace) System.out.println("** Conversion to RefUnit: "+value+getUnit(fromUnit).symbol+"["+fromUnit+"] -> "+result+getUnit(0).symbol+"[0]");
		return result;
	}

	/* *********************************** */
	/* CONVERSION FUNCTIONS: REF_UNIT -> ? */
	/* *********************************** */
	/**
	 * Lets convert from the reference unit to the given unit.
	 * 
	 * @param value			The value to convert (expressed in the reference unit).
	 * @param toUnit		The unit in which the given value must be converted.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If the destination unit is unknown.
	 * 
	 * @see #convertFromRef(double, int, String)
	 */
	public double convertFromRef(double value, String toUnit) throws UnknownUnitException {
		return convertFromRef(value, indexOf(toUnit), toUnit);
	}

	/**
	 * Lets convert from the reference unit to the given unit.
	 * 
	 * @param value			The value to convert (expressed in the reference unit).
	 * @param indToUnit		The index of the unit in which the given value must be converted.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If the destination unit is unknown.
	 * 
	 * @see #convertFromRef(double, int, String)
	 */
	public double convertFromRef(double value, int indToUnit) throws UnknownUnitException {
		return convertFromRef(value, indToUnit, null);
	}

	/**
	 * Lets convert from the reference unit to the given unit.
	 * 
	 * @param value			The value to convert (expressed in the reference unit).
	 * @param toUnit		The index of the unit in which the given value must be converted.
	 * @param toUnitSymbol	The symbol of the unit in which the given value must be converted.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws UnknownUnitException	If the destination unit is unknown.
	 */
	protected double convertFromRef(double value, int toUnit, String toUnitSymbol) throws UnknownUnitException {
		// Check the destination unit index:
		if (toUnit == NONE){
			if (toUnitSymbol == null)
				throw new UnknownUnitException(toUnit, "Conversion into the physical quantity \""+getName()+"\"");
			else
				throw new UnknownUnitException(toUnitSymbol, "Conversion into the physical quantity \""+getName()+"\"");
		}

		// If all is right, make the conversion:
		double result = value*conversionMatrix[toUnit][0];
		// TODO DEBUG Trace !
		if (SED.trace) System.out.println("** Conversion from RefUnit: "+value+getUnit(0).symbol+"[0] -> "+result+getUnit(toUnit).symbol+"["+toUnit+"]");
		return result;
	}

	/* ************************ */
	/* SED NOTIFICATION METHODS */
	/* ************************ */
	/**
	 * Adds the given axis to this physical quantity to notify it at each current unit modification.
	 * 
	 * @param newAxis	A axis to notify when the current unit changes.
	 */
	public final void addAxis(AxisUnit newAxis){
		vAxis.add(newAxis);
	}

	/**
	 * Removes the given axis from the list of axis to notify when the current unit changes.
	 * 
	 * @param axisToRemove	Axis to remove.
	 */
	public final void removeAxis(AxisUnit axisToRemove){
		vAxis.remove(axisToRemove);
	}

	/**
	 * Notifies all associated axis.
	 */
	protected void notifySEDs(){
		Iterator<AxisUnit> it = vAxis.iterator();
		while(it.hasNext())
			((AxisUnit)it.next()).notifySEDs();
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	@Override
	public boolean equals(Object obj){
		return (obj instanceof PhysicalQuantity) && ((PhysicalQuantity)obj).getName().equals(getName());
	}

	@Override
	public String toString(){
		return getName()+" ("+getCurrentUnit()+")";
	}

	/* ********* */
	/* SELF TEST */
	/* ********* */
	/**
	 * Displays this quantity on the standard output.
	 */
	public final void display(){
		System.out.println("\tQUANTITY: "+name+" ("+unicodeSymbol+")"+"\n\tUNITS:");
		for(int i=0; i<getNbUnits(); i++){
			UnitDef unit = getUnit(i);
			System.out.println("\t\t- "+unit.symbol+" ; "+unit.unicodeSymbol);
		}

		System.out.println("\tCONVERSION MATRIX:");
		for(int i=0; i<conversionMatrix.length; i++){
			System.out.print("\t\t| ");
			for(int j=0; j<conversionMatrix[i].length; j++)
				System.out.print(conversionMatrix[i][j]+" | ");
			System.out.println();
		}
		System.out.println();
	}

	public static void main(String[] args) throws Exception {
		PhysicalQuantity q = new PhysicalQuantity("Wavelength", "", "\u03BB", new UnitDef("um", "\u00B5m"));

		UnitDef[] units = new UnitDef[]{new UnitDef("Angstrom", "\u212B"), new UnitDef("km"), new UnitDef("m")};
		double[] convFactors = new double[]{1e-4, 1e9, 1e6};
		q.add(units, convFactors);

		System.out.println("### BEFORE DELETION ###");
		q.display();

		double value = 2;
		String fromUnit = "Angstrom", toUnit = "m";
		System.out.println("\n### CONVERTING "+value+fromUnit.toUpperCase()+" INTO "+toUnit.toUpperCase()+"... ###");
		value = q.convert(value, fromUnit, toUnit);
		System.out.println("Result = "+value+toUnit+"\n");

		q.remove(new int[]{3,0});

		System.out.println("### AFTER DELETION ###");
		q.display();
	}

}
