package sedplugin.sed.unit;

import java.util.Iterator;

import cds.astro.Converter;


import sedplugin.errors.ErrorsLog;
import sedplugin.errors.SEDException;
import sedplugin.errors.unit.ConversionException;
import sedplugin.errors.unit.UnknownUnitException;
import sedplugin.sed.SED;
import sedplugin.sed.SEDPoint;
import sedplugin.sed.ToolBox;
import sedplugin.sed.unit.PhysicalQuantity.UnitEntry;

/**
 * <p>Abscissa axis of a SED.</p>
 * 
 * <p>There are, by default, only two physical quantities: Wavelength (reference unit: um) and Frequency (reference unit: Hz).</p>
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 03/2012
 */
public class XAxis extends AxisUnit {

	/** Index of the wavelength physical quantity. */
	public static final int WAVELENGTH = 0;

	/** Index of the frequency physical quantity. */
	public static final int FREQUENCY = 1;

	/** Speed of light constant (in um/s). */
	public static final double SPEED_OF_LIGHT = 299792458e6;

	/** Current instance of this class. */
	private static XAxis currentXAxis;

	/* *********** */
	/* CONSTRUCTOR */
	/* *********** */
	/**
	 * <p>Initializes the physical quantities of this axis:</p>
	 * <ul>
	 * 	<li>Wavelength (name={@link #WAVELENGTH} and reference unit=um):
	 * 		<ul>
	 * 			<li>um <i>(default)</i></li>
	 * 			<li>Angstrom</li>
	 * 		</ul>
	 * 	</li>
	 * 	<li>Frequency (name={@link #FREQUENCY} and reference unit=Hz):
	 * 		<ul>
	 * 			<li>Hz</li>
	 * 			<li>MHz</li>
	 * 			<li>GHz <i>(default)</i></li>
	 * 			<li>THz</li>
	 * 		</ul>
	 * 	</li>
	 * </ul>
	 */
	public XAxis() throws SEDException {
		super("Abscissa");

		// Create the wavelength units:
		PhysicalQuantity wQuantity = new PhysicalQuantity("Wavelength", "", "\u03BB", new UnitDef("um", "\u00B5m"));
		UnitEntry[] wUnits = new UnitEntry[]{
				new UnitEntry(new UnitDef("Angstrom", "\u212B"), 1e-4)
		};
		wQuantity.add(wUnits);
		add(wQuantity);

		// Create the frequency units:
		PhysicalQuantity fQuantity = new PhysicalQuantity("Frequency", "", "\u03BD", new UnitDef("Hz"));
		UnitEntry[] fUnits = new UnitEntry[]{
				new UnitEntry(new UnitDef("MHz"), 1e6),
				new UnitEntry(new UnitDef("GHz"), 1e9),
				new UnitEntry(new UnitDef("THz"), 1e12)
		};
		fQuantity.add(fUnits);
		fQuantity.setCurrentUnit(2);
		add(fQuantity);
	}

	/**
	 * Initializes the physical quantities of this axis and set the current unit to the given value.
	 * 
	 * @param unitName	Names of the unit to set as the current one in this axis.
	 */
	public XAxis(final String unitName) throws SEDException {
		this();
		if (!setUnit(unitName))
			setCustomUnit(new UnitDef(unitName));
	}

	/* ****** */
	/* GETTER */
	/* ****** */
	/**
	 * Gets the current instance of the {@link XAxis} class.
	 * 
	 * @return	The current instance of {@link XAxis}.
	 */
	public static final synchronized XAxis getCurrentInstance(){
		if (currentXAxis == null){
			try{
				currentXAxis = new XAxis();
			}catch(SEDException ex){
				ErrorsLog.getCurrentInstance().addException(ex);
			}
		}
		return currentXAxis;
	}

	/* ***************** */
	/* INHERITED METHODS */
	/* ***************** */
	@Override
	protected void notifySEDs() {
		Iterator<SED> it = seds.iterator();
		while(it.hasNext())
			((SED)it.next()).updateX();
	}

	@Override
	protected double extractValueToConvert(SEDPoint point){
		return point.getOriginalX();
	}

	@Override
	protected String extractSourceUnit(SEDPoint point){
		return point.getOriginalXUnit();
	}

	@Override
	protected double convertQuantity(double value, int indFromQuantity, int indToQuantity, SEDPoint point) throws ConversionException, UnknownUnitException {
		if (indFromQuantity == indToQuantity)
			return value;

		if (indFromQuantity < 0 || indFromQuantity > 1)
			throw new ConversionException("Source quantity incorrect: there is no "+indFromQuantity+"th physical quantity !", "Quantity conversion in the axis \""+getName()+"\"");
		if (indToQuantity < 0 || indToQuantity > 1)
			throw new ConversionException("Destination quantity incorrect: there is no "+indToQuantity+"th physical quantity !", "Quantity conversion in the axis \""+getName()+"\"");

		// TODO DEBUG Trace !
		if (SED.trace) System.out.println("** Convert quantity: "+getQuantity(indFromQuantity).getUnicodeSymbol()+" -> "+getQuantity(indToQuantity).getUnicodeSymbol());

		return SPEED_OF_LIGHT/value;
	}

	/* ************************ */
	/* OTHER CONVERSION METHODS */
	/* ************************ */
	/**
	 * Converts the given value from the given unit to the current unit of this axis.
	 * 
	 * @param value			The value to convert.
	 * @param fromUnit		The unit of the given value.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws ConversionException	If one conversion is impossible.
	 * @throws UnknownUnitException	If one or both original units of the given SED point are unknown.
	 * 
	 * @see #searchUnit(String)
	 * 
	 * @see PhysicalQuantity#convertToRef(double, int)
	 * @see #convertQuantity(double, int, int, SEDPoint)
	 * @see PhysicalQuantity#convertFromRef(double, int)
	 */
	public double convert(double value, String fromUnit) throws ConversionException, UnknownUnitException {
		boolean knownSrcUnit = true;
		int[] indexes = new int[0];

		try{
			indexes = searchUnit(fromUnit);
		}catch(UnknownUnitException uue){
			knownSrcUnit = false;
		}

		double result = value;

		if (!knownSrcUnit || isCustomQuantity(indexes[0]) || isCustomUnit()){
			Converter converter = ToolBox.getAppropriateConverter(fromUnit, getUnit().symbol);
			if (converter == null)
				throw new ConversionException("Destination quantity incorrect: the unit \""+getUnit().symbol+"\" is not compatible with \""+fromUnit+"\" !", "Quantity conversion in the axis \""+getName()+"\"");
			else
				result = converter.convert(result);
		}else{
			result = getQuantity(indexes[0]).convertToRef(result, indexes[1]);
			result = convertQuantity(result, indexes[0], currentQuantity, null);
			result = getQuantity().convertFromRef(result, getQuantity().currentUnit);
		}

		if (Double.isNaN(result) || Double.isInfinite(result))
			throw new ConversionException(value, fromUnit, getUnit().symbol, "converted value is "+result);

		return result;
	}

	/**
	 * Converts the given value from the given unit to the given unit.
	 * 
	 * @param value			The value to convert.
	 * @param fromUnit		The source unit.
	 * @param toUnit		The destination unit.
	 * 
	 * @return				The converted value.
	 * 
	 * @throws ConversionException	If one conversion is impossible.
	 * @throws UnknownUnitException	If one or both original units of the given SED point are unknown.
	 * 
	 * @see #searchUnit(String)
	 * 
	 * @see PhysicalQuantity#convertToRef(double, int)
	 * @see #convertQuantity(double, int, int, SEDPoint)
	 * @see PhysicalQuantity#convertFromRef(double, int)
	 */
	public double convert(double value, String fromUnit, String toUnit) throws ConversionException, UnknownUnitException {
		boolean knownSrcUnit = true;
		int[] srcIndexes = new int[0];

		try{
			srcIndexes = searchUnit(fromUnit);
		}catch(UnknownUnitException uue){
			knownSrcUnit = false;
		}

		int[] destIndexes = searchUnit(toUnit);

		double result = value;

		if (!knownSrcUnit || isCustomQuantity(srcIndexes[0]) || isCustomQuantity(destIndexes[0])){
			Converter converter = ToolBox.getAppropriateConverter(fromUnit, toUnit);
			if (converter == null)
				throw new ConversionException("Destination quantity incorrect: the unit \""+toUnit+"\" is not compatible with \""+fromUnit+"\" !", "Quantity conversion in the axis \""+getName()+"\"");
			else
				result = converter.convert(result);
		}else{
			result = getQuantity(srcIndexes[0]).convertToRef(result, srcIndexes[1]);
			result = convertQuantity(result, srcIndexes[0], destIndexes[0], null);
			result = getQuantity(destIndexes[0]).convertFromRef(result, destIndexes[1]);
		}

		if (Double.isNaN(result) || Double.isInfinite(result))
			throw new ConversionException(value, fromUnit, toUnit, "converted value is "+result);

		return result;
	}

	/* ************** */
	/* TODO SELF TEST */
	/* ************** */
	public static void main(String[] args) throws Exception {
		AxisUnit axis = new XAxis();
		System.out.println("AXIS["+axis.getName()+"]: "+axis);
		Iterator<PhysicalQuantity> itQuantities = axis.getQuantities();
		while(itQuantities.hasNext())
			((PhysicalQuantity)itQuantities.next()).display();
		axis.setCustomUnit(new UnitDef("pc"));
		System.out.println("12.0um => frequency = "+axis.convertQuantity(12.0, 0, 1, null));
		System.out.println("2.4982704833333332E13 frequency => um = "+axis.convertQuantity(2.4982704833333332E13, 1, 0, null));
		System.out.println("CHANGED ? "+axis.setQuantity(0)+" => UNIT = "+axis.getUnit());
		System.out.println("3.89E-22 => um = "+axis.convert(new SEDPoint(3.89E-22, "pc", 5.23, "Jy/sr")));
	}

}
