package sedplugin.votable;

import java.io.File;
import java.io.InputStream;

import java.net.URL;

import java.util.HashMap;
import java.util.Iterator;
import java.util.Vector;


import sedplugin.sed.SED;
import sedplugin.sed.SEDPoint;
import sedplugin.sed.unit.XAxis;
import sedplugin.sed.unit.YAxis;

/**
 * Parser of VOTable files following the data-model Photometry
 * (cf Sebastien Derriere for more details about this data-model).
 * 
 * @author Gr&eacute;gory Mantelet (CDS)
 * @version 06/2012
 */
public class SEDParser extends VOTableParser {

	public final static String PDM_UTYPE = "PhotometryPoint";

	protected final XAxis xUnit;
	protected final YAxis yUnit;

	protected Vector<SED> seds = new Vector<SED>();

	protected SEDMeta sedMetadata = null;
	protected SEDPointMeta currentSEDPoint = null;

	protected HashMap<String, Meta> mapReferences = new HashMap<String, Meta>();

	protected int indField = -1;
	protected int nbFields = 0;
	protected Meta[] row;

	protected String parentInSed = "";

	public SEDParser(XAxis wUnit, YAxis fUnit) {
		super();
		xUnit = wUnit;
		yUnit = fUnit;
	}

	public SEDParser(XAxis wUnit, YAxis fUnit, InputStream voTableInput) throws Exception {
		super(voTableInput);
		xUnit = wUnit;
		yUnit = fUnit;
	}

	public SEDParser(XAxis wUnit, YAxis fUnit, URL voTableUrl) throws Exception {
		super(voTableUrl);
		xUnit = wUnit;
		yUnit = fUnit;
	}

	public SEDParser(XAxis wUnit, YAxis fUnit, File voTableFile) throws Exception {
		super(voTableFile);
		xUnit = wUnit;
		yUnit = fUnit;
	}

	public Iterator<SED> getSEDs(){
		return seds.iterator();
	}

	protected static boolean isMeanWavelength(String utype){
		return (utype.equalsIgnoreCase("phfdm:PhotometryFilter.SpectralAxis.Coverage.Location.Value") || utype.equalsIgnoreCase("PhotometryFilter.SpectralAxis.Coverage.Location.Value")
				|| utype.equalsIgnoreCase("phfdm:PhotometryFilter.meanWavelength") || utype.equalsIgnoreCase("PhotometryFilter.meanWavelength")
				|| utype.equalsIgnoreCase("phfdm:PhotometryFilter.meanFrequency") || utype.equalsIgnoreCase("PhotometryFilter.meanFrequency")
				|| utype.equalsIgnoreCase("spec:Spectrum.Data.SpectralAxis.Value") || utype.equalsIgnoreCase("Spectrum.Data.SpectralAxis.Value")
				|| utype.equalsIgnoreCase("spec:Spectrum.Char.SpectralAxis.Coverage.Location.Value"));
	}

	protected static boolean isMinWavelength(String utype){
		return (utype.equalsIgnoreCase("phfdm:PhotometryFilter.SpectralAxis.Coverage.Location.Start") || utype.equalsIgnoreCase("PhotometryFilter.SpectralAxis.Coverage.Location.Start")
				|| utype.equalsIgnoreCase("phfdm:PhotometryFilter.minWavelength") || utype.equalsIgnoreCase("PhotometryFilter.minWavelength")
				|| utype.equalsIgnoreCase("spec:Spectrum.Char.SpectralAxis.Coverage.Bounds.Start") || utype.equalsIgnoreCase("Spectrum.Char.SpectralAxis.Coverage.Bounds.Start"));
	}

	protected static boolean isMaxWavelength(String utype){
		return (utype.equalsIgnoreCase("phfdm:PhotometryFilter.SpectralAxis.Coverage.Location.Stop") || utype.equalsIgnoreCase("PhotometryFilter.SpectralAxis.Coverage.Location.Stop")
				|| utype.equalsIgnoreCase("phfdm:PhotometryFilter.maxWavelength") || utype.equalsIgnoreCase("PhotometryFilter.maxWavelength")
				|| utype.equalsIgnoreCase("spec:Spectrum.Char.SpectralAxis.Coverage.Bounds.Stop") || utype.equalsIgnoreCase("Spectrum.Char.SpectralAxis.Coverage.Bounds.Stop"));
	}

	protected static boolean isZeroPoint(String utype){
		return (utype.equalsIgnoreCase("phfdm:ZeroPoint.zeroPointFlux") || utype.equalsIgnoreCase("ZeroPoint.zeroPointFlux")
				|| utype.equalsIgnoreCase("phfdm:PhotometryFilter.zeroPoint") || utype.equalsIgnoreCase("PhotometryFilter.zeroPoint"));
	}

	protected static boolean isFluxValue(String utype){
		return (utype.equalsIgnoreCase("phfdm:PhotometryPoint.Value.value") || utype.equalsIgnoreCase("PhotometryPoint.Value.value")
				|| utype.equalsIgnoreCase("spec:Spectrum.Data.FluxAxis.Value") || utype.equalsIgnoreCase("Spectrum.Data.FluxAxis.Value"));
	}

	protected static boolean isFluxError(String utype){
		return (utype.equalsIgnoreCase("phfdm:PhotometryPoint.Value.error") || utype.equalsIgnoreCase("PhotometryPoint.Value.error")
				|| utype.equalsIgnoreCase("spec:Spectrum.Data.FluxAxis.Accuracy.StatError") || utype.equalsIgnoreCase("Spectrum.Data.FluxAxis.Accuracy.StatError"));
	}

	@Override
	protected void reset(){
		super.reset();
		seds = new Vector<SED>();
		sedMetadata = null;
		currentSEDPoint = null;
		mapReferences = new HashMap<String, Meta>();
		indField = -1;
		nbFields = 0;
		row = null;
		parentInSed = "";
	}

	@Override
	protected void startDocument() throws Exception { ; }

	@Override
	protected void endDocument() throws Exception {
		//		System.out.println("***** METADATA *****\n"+sedMetadata);
		for(int i=0; i<seds.size(); i++){
			/*SED sed = */seds.get(i);
			//System.out.println("\n"+(i+1)+"°- SED "+sed);
		}
	}

	@Override
	protected void startVoTable(HashMap<String, String> attributes) throws Exception { ; }

	@Override
	protected void endVoTable() throws Exception { ; }

	@Override
	protected void startResource(HashMap<String, String> attributes) throws Exception { ; }

	@Override
	protected void endResource() throws Exception { ; }

	@Override
	protected void startTable(HashMap<String, String> attributes) throws Exception {
		sedMetadata = new SEDMeta();
		sedMetadata.name = attributes.get("name").toString();
		parentInSed = "table";
	}

	@Override
	protected void endTable() throws Exception {
		//		sedMetadata = null;
		parentInSed = "";
	}

	@Override
	protected void startGroup(HashMap<String, String> attributes) throws Exception {
		currentSEDPoint = null;
		if (attributes.size() > 0){
			String utype = attributes.get("utype");
			if (utype != null && (utype.equalsIgnoreCase("phfdm:"+PDM_UTYPE) || utype.equalsIgnoreCase(PDM_UTYPE))){
				currentSEDPoint = new SEDPointMeta();
				currentSEDPoint.name = attributes.get("name");
				parentInSed = "group";
			}
		}
	}

	@Override
	protected void endGroup() throws Exception {
		if (currentSEDPoint != null){
			sedMetadata.points.add(currentSEDPoint);
			currentSEDPoint = null;
			parentInSed = "table";
		}
	}

	@Override
	protected void startParam(HashMap<String, String> attributes) throws Exception {
		if (currentSEDPoint != null){
			String utype=null, unit=null, value=null;
			utype = (attributes.containsKey("utype")?attributes.get("utype").toString():null);
			unit = (attributes.containsKey("unit")?attributes.get("unit").toString():"");
			value = (attributes.containsKey("value")?attributes.get("value").toString():"");

			if (value == null)
				value = "NaN";

			if (utype != null){
				if (isMeanWavelength(utype)){
					currentSEDPoint.x.value = value;
					currentSEDPoint.x.unit = unit;
				}else if (isMinWavelength(utype)){
					currentSEDPoint.xMin.value = value;
					currentSEDPoint.xMin.unit = unit;
				}else if (isMaxWavelength(utype)){
					currentSEDPoint.xMax.value = value;
					currentSEDPoint.xMax.unit = unit;
				}else if (isZeroPoint(utype)){
					currentSEDPoint.zp.value = value;
					currentSEDPoint.zp.unit = unit;
				}else if (isFluxValue(utype)){
					currentSEDPoint.y.value = value;
					currentSEDPoint.y.unit = unit;
				}else if (isFluxError(utype)){
					currentSEDPoint.yError.value = value;
					currentSEDPoint.yError.unit = unit;
				}
			}
		}
	}

	@Override
	protected void endParam() throws Exception { ; }

	@Override
	protected void startField(HashMap<String, String> attributes) throws Exception {
		indField++;
		nbFields++;

		String id = null;
		if (attributes.containsKey(ID))
			id = attributes.get(ID).toString().trim();
		else if (attributes.containsKey("id"))
			id = attributes.get("id").toString().trim();

		if (id != null){
			String unit = "";
			if (attributes.containsKey(UNIT))
				unit = attributes.get(UNIT).toString().trim();
			if (mapReferences.containsKey(id)){
				Meta meta = mapReferences.get(id);
				meta.indField = indField;
				meta.unit = unit;
			}
		}
	}

	@Override
	protected void endField() throws Exception { ; }

	@Override
	protected void startFieldRef(HashMap<String, String> attributes) throws Exception {
		if (currentSEDPoint != null){
			String utype = null, reference = null;

			utype = attributes.get("utype").toString();
			reference = attributes.get("ref").toString();

			if (reference != null){
				if (isMeanWavelength(utype)){
					currentSEDPoint.x.reference = reference;
					mapReferences.put(reference, currentSEDPoint.x);
				}else if (isMinWavelength(utype)){
					currentSEDPoint.xMin.reference = reference;
					mapReferences.put(reference, currentSEDPoint.xMin);
				}else if (isMaxWavelength(utype)){
					currentSEDPoint.xMax.reference = reference;
					mapReferences.put(reference, currentSEDPoint.xMax);
				}else if (isZeroPoint(utype)){
					currentSEDPoint.zp.reference = reference;
					mapReferences.put(reference, currentSEDPoint.zp);
				}else if (isFluxValue(utype)){
					currentSEDPoint.y.reference = reference;
					mapReferences.put(reference, currentSEDPoint.y);
				}else if (isFluxError(utype)){
					currentSEDPoint.yError.reference = reference;
					mapReferences.put(reference, currentSEDPoint.yError);
				}
			}
		}
	}

	@Override
	protected void endFieldRef() throws Exception { ; }

	@Override
	protected void startTableData() throws Exception {
		row = new Meta[nbFields];
		Iterator<Meta> it = mapReferences.values().iterator();
		while(it.hasNext()){
			Meta item = it.next();
			row[item.indField] = item;
		}
	}

	@Override
	protected void endTableData() throws Exception {
		row = null;
	}

	@Override
	protected void startTR(HashMap<String, String> attributes) throws Exception {
		indField = -1;
	}

	@Override
	protected void endTR() throws Exception {
		SED sed = sedMetadata.toSED("", xUnit, yUnit);
		if (sed.size() > 0){
			sed.setName(sed.getName()+"["+seds.size()+"]");
			seds.add(sed);
		}
	}

	@Override
	protected void startTD(HashMap<String, String> attributes) throws Exception {
		indField++;
	}

	@Override
	protected void textTD(String text) throws Exception {
		text = text.trim();
		if (row[indField] != null)
			row[indField].value = text;
	}

	@Override
	protected void endTD() throws Exception { ; }

	@Override
	protected void startDescription(HashMap<String, String> attributes) throws Exception { ; }

	@Override
	protected void textDescription(String text) throws Exception {
		if (parentInSed.equalsIgnoreCase("table") && sedMetadata.description.length() == 0)
			sedMetadata.description = text;
		else if (parentInSed.equalsIgnoreCase("group") && currentSEDPoint != null && currentSEDPoint.description.length() == 0)
			currentSEDPoint.description = text;
	}

	@Override
	protected void endDescription() throws Exception { ; }


	/**
	 * @param args
	 */
	public static void main(String[] args) throws Exception {
		XAxis xAxis = new XAxis();
		YAxis yAxis = new YAxis(xAxis);
		yAxis.setQuantity(YAxis.FLUX);
		SEDParser parser = new SEDParser(xAxis, yAxis);
		try{
			parser.parse(new File("photometry_examples/Examples Photometry/03-12-2010/UGC2826_IRAS_phot.xml"));
			Iterator<SED> it = parser.getSEDs();
			while(it.hasNext()){
				SED sed = it.next();
				System.out.println("* "+sed+":");
				for(SEDPoint pt : sed)
					System.out.println("\t-> "+pt);
			}
		}catch(Exception ex){
			System.err.println("((X)) ERROR AT LINE "+parser.getLineNumber()+", COLUMN "+parser.getColumnNumber()+" ! ((X))");
			ex.printStackTrace();
		}
	}

	public static class SEDMeta {
		public String name = "Anonymous";
		public String description = "";
		public Vector<SEDPointMeta> points = new Vector<SEDPointMeta>();

		public SEDMeta(){ ; }

		public SEDMeta(SEDMeta toCopy){
			name = toCopy.name;
			description = toCopy.description;
			points = new Vector<SEDPointMeta>(toCopy.points.size());
			for(int i=0; i<toCopy.points.size(); i++)
				points.add(new SEDPointMeta(toCopy.points.get(i)));
		}

		public SED toSED(String catalogName, XAxis xUnit, YAxis yUnit){
			SED sp = new SED(xUnit, yUnit);
			sp.setName(name);

			for(int i=0; i<points.size(); i++){
				SEDPointMeta point = points.get(i);
				if (!point.y.unit.equalsIgnoreCase("mag"))
					sp.add(point.toSpectrumPoint());
			}

			return sp;
		}

		@Override
		public String toString(){
			String sedSerial = "SED["+name+" ; "+description+"]:\n";
			for(int i=0; i<points.size(); i++){
				SEDPointMeta point = points.get(i);
				sedSerial += "\t* "+point+"\n";
			}
			return sedSerial;
		}
	}

	public static class SEDPointMeta {
		public String name = "Anonymous";
		public String description = "";

		public Meta x = new Meta();
		public Meta xMin = new Meta();
		public Meta xMax = new Meta();

		public Meta zp = new Meta();

		public Meta y = new Meta();
		public Meta yError = new Meta();

		public SEDPointMeta(){ ; }

		public SEDPointMeta(SEDPointMeta toCopy){
			name = toCopy.name;
			description = toCopy.description;

			x = new Meta(toCopy.x);
			xMin = new Meta(toCopy.xMin);
			xMax = new Meta(toCopy.xMax);

			zp = new Meta(toCopy.zp);

			y = new Meta(toCopy.y);
			yError = new Meta(toCopy.yError);
		}

		public SEDPoint toSpectrumPoint(){
			if (x.value.length() == 0)
				x.value = "NaN";
			if (y.value.length() == 0)
				y.value = "NaN";
			return new SEDPoint(Double.parseDouble(x.value), x.unit, Double.parseDouble(y.value), y.unit, Double.NaN);
		}

		@Override
		public String toString(){
			return name+" ("+description+") {x: "+x+", min: "+xMin+", max: "+xMax+", zp: "+zp+" ; y: "+y+", error: "+yError+"}";
		}
	}

	public static class Meta {
		public String reference = null;
		public int indField = -1;
		public String unit = "";
		public String value = "";

		public Meta(){ ; }

		public Meta(Meta toCopy){
			unit = toCopy.unit;
			value = toCopy.value;
		}

		@Override
		public String toString(){
			return ((reference != null)?("[!"+reference+"]<->"+indField):value)+" "+unit;
		}
	}

}
